#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# ratio_pars_vs_nextpars_msa_case_III.pl:
#	This Perl script was created by modifying 'ratio_pars_vs_nextpars_pwa_case_iii.pl'.
#	It examines the local MSAs in case (III),
#	that is, each of them has a run of sites in sequence 1 aligned with runs of gaps in sequences 2 and 3,
#	in between a pair of PASs.
#	It compares the contribution by the 3rd-type non-parsimonious history (consisting of deletions along branches 2 and 3)
#	with that contributed by the parsimonious history.
#
#	[CREATED on May 10 (Sun), 2015.]
#
# ratio_pars_vs_nextpars_pwa_case_iii.pl:
#	This Perl script does nearly the same jobs as its predecessor, 'ratio_pars_vs_nextpars_pwa_case_ii.pl' does.
#	The only difference is that it examines the local PWAs in case (iii),
#	that is, each of them has no ancestral sites and some descendant sites
#	in between two preserved ancestral sites (PASs) that are next to each other.
#
#	[CREATED on May 4 (Mon), 2015.]
#
# ratio_pars_vs_nextpars_pwa_case_ii.pl:
#	This Perl script calculates the portion of the multiplication factor for a local PWA
#	contributed by the parsimonious (i.e., fewest-indel) local histories,
#	that contributed by the next-parsimonious (i.e., next-fewest-indel) local histories,
#	and the ratio of the former to the latter,
#	under fixed model parameters and under several different values of the time lapse (t_F - t_I).
#	The local PWAs examined in this script are in case (ii),
#	that is, each of them has some ancestral sites and no descendant sites
#	in between two preserved ancestral sites (PASs) that are next to each other.
#
#	[CREATED on May 2 (Sat), 2015.]
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyArrayManipu qw( diff_array );

use MyPertProb_locPWA_dawg qw(
	mfac_agaps_vs_dsites_1event
	mfac_agaps_vs_dsites_2events
	);
	##mfac_asites_vs_dgaps_1event
	##mfac_asites_vs_dgaps_2events

use MyPertProb_locMSA_dawg_smpl qw(
	mfac_lone_dsites_vs_ddgaps_2events_3rd
	);


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 3;


	### MODEL PARAMETERS ###

our $TOTRATE_INS = 0.1; # Total insertion rate (per unit time). (lambda_I in Dawg's indel model.)
our $TOTRATE_DEL = 0.1; # Total deletion rate (per unit time). (lambda_D in Dawg's indel model.)

our $EXPONENT_INS = 1.6; # Exponent for the power-law distribution of insertion lengths: f_I (l) is proportional to l^{-$EXPONENT_INS}.
our $EXPONENT_DEL = 1.6; # Exponent for the power-law distribution of deletion lengths: f_D (l) is proportional to l^{-$EXPONENT_DEL}.

=for the full-scale analysis (1)
our $CUTOFF_INS_LEN = 5000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 5000; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (2)
our $CUTOFF_INS_LEN = 1000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 1000; # Cutoff value for the deletion length.
=cut
##=for the full-scale analysis (3)
our $CUTOFF_INS_LEN = 500; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 500; # Cutoff value for the deletion length.
##=cut
=for TESTING
our $CUTOFF_INS_LEN = 200; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 200; # Cutoff value for the deletion length.
=cut

my $nfac_ins = my $nfac_del = 0;
for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $nfac_ins += exp (-$EXPONENT_INS * log ($l)); }
for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $nfac_del += exp (-$EXPONENT_DEL * log ($l)); }


our @FREQS_INS = (0); # Relative frequencies of the insertion lengths.
our @FREQS_DEL = (0); # Relative frequencies of the deletion lengths.

for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $FREQS_INS[$l] = exp (-$EXPONENT_INS * log ($l)) /$nfac_ins; }
for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $FREQS_DEL[$l] = exp (-$EXPONENT_DEL * log ($l)) /$nfac_del; }

	### Other Important Parameters ###

our @TIME_LAPSES = (0.05, 0.2, 0.5, 1.0, 2.0); # Values of the time lapse (= t_F - t_I) to be examined.

=for the full-scale analysis (1,2)
our $UB_CT_DECSITES1 = 1000; # Upper-bound of the number of descendant sites in between the PASs.
=cut
##=for the full-scale analysis (3)
our $UB_CT_DECSITES1 = 500; # Upper-bound of the number of descendant sites in between the PASs.
##=cut
=for TESTING
our $UB_CT_DECSITES1 = 100; # Upper-bound of the number of descendant sites in between the PASs.
=cut

	### Templates of parameters along branches. ###
	### (ADDED in creation of 'ratio_pars_vs_nextpars_msa_case_III.pl' on May 10 (Sun), 2015.)

our @SET_PARAMS_TMPL = (1, $TOTRATE_INS, $TOTRATE_DEL, \@FREQS_INS, \@FREQS_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN);


	### Output files. ###

	# $outfile = "$outdir/$out_pref$tlapse$out_suff"

my $outdir = 'Fixed_Time_Lapse'; ## For the full-scale analysis.
##my $outdir = 'TESTs/Test1'; ## For TESTING the script.
my $out_pref = 'out_mfacs_ratios_3rd.';
my $out_suff = '.txt';

my $out_summ = "$outdir/out_ratios_3rd_for_varied_tlapse.txt";


	### Create the output directory. ###

unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

print STDOUT "# It compares the contribution by the 3rd-type next-parsimonious history with that by the parsimonious history, both to the local MSA probability in case (III).\n\n";

print STDOUT "Summary output file: $out_summ .\n\n";

print STDOUT "Detailed output files: {out_dir}/$out_pref", '{time_lapse}', "$out_suff,\n";
print STDOUT "       where {out_dir} = $outdir ,\n";
print STDOUT "             {time_lapse} = ", join (', ', @TIME_LAPSES), " .\n\n";


print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";


print STDOUT "\n<< Model Parameters (shared by all branches) >>\n\n";

print STDOUT '$TOTRATE_INS = ', $TOTRATE_INS, " ,\n";
print STDOUT '$TOTRATE_DEL = ', $TOTRATE_DEL, " ,\n\n";

print STDOUT '@FREQS_INS[l] prpt l^{', -$EXPONENT_INS, "}   (l=1,...,$CUTOFF_INS_LEN) ,\n";
print STDOUT '@FREQS_DEL[l] prpt l^{', -$EXPONENT_DEL, "}   (l=1,...,$CUTOFF_DEL_LEN) .\n\n";


if ($TEST_LEVEL>2) {

   print STDOUT "\n< Indel length frequencies >\n\n";

   print STDOUT join ("\t", 'Length', 'Ins', 'Del'), "\n\n";

   my $cutoff = ($CUTOFF_INS_LEN > $CUTOFF_DEL_LEN) ? $CUTOFF_INS_LEN : $CUTOFF_DEL_LEN ;

   for (my $l=1; $l<= $cutoff; $l++) {
	my $f_ins = ($l > $CUTOFF_INS_LEN) ? 'undef' : $FREQS_INS[$l] ;
	my $f_del = ($l > $CUTOFF_DEL_LEN) ? 'undef' : $FREQS_DEL[$l] ;

	print STDOUT join ("\t", $l, $f_ins, $f_del), "\n";
   }
   print STDOUT "\n\n";

}  # END of "if ($TEST_LEVEL>2) {...}". #

print STDOUT "\n<<< Main Analysis >>>\n\n\n";


my @ratios = (); # $ratios[$n_D1]->[$indx_tlapse] is the ratio of the pars to the next-pars for the time lapse of $TIME_LAPSES[$indx_tlapse] and $n_D1 descendant sites.

for (my $n_D1=0; $n_D1 <= $UB_CT_DECSITES1; $n_D1++) { $ratios[$n_D1] = []; }


for (my $indx_tlapse = 0; $indx_tlapse < @TIME_LAPSES; $indx_tlapse++) { # Main outer for-loop (over time-lapses). #

    my $tlapse = $TIME_LAPSES[$indx_tlapse];

		# (ADDED in creation of 'ratio_pars_vs_nextpars_msa_case_III.pl' on May 10 (Sun), 2015.)

		# Indel model parameters along the 3 branches. (Assume that all models share the same parameters.) #
    my @set_params1 = my @set_params2 = my @set_params3 = @SET_PARAMS_TMPL;

		# Assume that all branches have the same length (= $tlapse).
    $set_params1[0] = $set_params2[0] = $set_params3[0] = $tlapse;

		# (END of "ADDED in creation of 'ratio_pars_vs_nextpars_msa_case_III.pl' on May 10 (Sun), 2015.")


    my $outfile = "$outdir/$out_pref$tlapse$out_suff";

    my $fh_out = open_output_file ($outfile, 0);
    unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED for the time-lapse = $tlapse: $! !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    print $fh_out join ("\t", '#{anc_sites}', 'A=Mfac(pars)', 'B=Mfac(next-pars(3rd))', 'Ratio(=B/A)'), "\n\n";
    ##print $fh_out join ("\t", '#{anc_sites}', 'A=Mfac(pars)', 'B=Mfac(next-pars)', 'Ratio(=B/A)'), "\n\n";

    for (my $n_D1=1; $n_D1 <= $UB_CT_DECSITES1; $n_D1++) { # Main inner for-loop (over the descendant site counts in sequence 1). #

		# 1st MAJOR MODIFICATION in creation of 'ratio_pars_vs_nextpars_msa_case_III.pl' (on May 10 (Sun), 2015). #
	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @set_params1;
	my $mfac_pars = mfac_agaps_vs_dsites_1event ($n_D1, $b1, $lambda_I1, $lambda_D1, @{$freqs_I1}, $cutoff_I1);

=for the predecessor 'ratio_pars_vs_nextpars_pwa_case_iii.pl'.
		# A MAJOR MODIFICATION in creation of  'ratio_pars_vs_nextpars_pwa_case_iii.pl' (on May 4 (Mon), 2015).
	my $mfac_pars = mfac_agaps_vs_dsites_1event ($n_D, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, $CUTOFF_INS_LEN);
	##my $mfac_pars = mfac_asites_vs_dgaps_1event ($n_A, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_DEL, $CUTOFF_DEL_LEN);
=cut
	unless (defined $mfac_pars) {
	    print STDOUT "[ERROR '$program_name'] 'mfac_agaps_vs_dsites_1event' FAILED for #{dec_sites}_1 = $n_D1 and {branch_length}_1 = $b1 !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

		# 2nd MAJOR MODIFICATION in creation of 'ratio_pars_vs_nextpars_msa_case_III.pl' (on May 10 (Sun), 2015). #

	my $mfac_nextpars_3rd = mfac_lone_dsites_vs_ddgaps_2events_3rd ($n_D1, @set_params1, @set_params2, @set_params3);
	unless (defined $mfac_nextpars_3rd) {
	    print STDOUT "[ERROR '$program_name'] 'mfac_lone_dsites_vs_ddgaps_2events_3rd' FAILED for #{dec_sites}_1 = $n_D1 and {time-lapse} = $tlapse !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
=for the predecessor 'ratio_pars_vs_nextpars_pwa_case_iii.pl'.
		# A MAJOR MODIFICATION in creation of  'ratio_pars_vs_nextpars_pwa_case_iii.pl' (on May 4 (Mon), 2015).
	my ($mfac_nextpars, $components) = mfac_agaps_vs_dsites_2events ($n_D, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN);
	##my ($mfac_nextpars, $components) = mfac_asites_vs_dgaps_2events ($n_A, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN);
	unless (defined $mfac_nextpars) {
	    print STDOUT "[ERROR '$program_name'] 'mfac_agaps_vs_dsites_2events' FAILED for #{dec_sites} = $n_D and {time-lapse} = $tlapse !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
=cut

	my $ratio_mfac_3rd = $mfac_nextpars_3rd / $mfac_pars ;

	print $fh_out join ("\t", $n_D1, $mfac_pars, $mfac_nextpars_3rd, $ratio_mfac_3rd), "\n";

	$ratios[$n_D1]->[$indx_tlapse] = $ratio_mfac_3rd ;
	##$ratios[$n_A]->[$indx_tlapse] = $ratio_mfac ;

    }					# End of the main inner for-loop (over the descendant site counts in sequence 1). #

    close ($fh_out);

}							# End of the main outer for-loop (over time-lapses). #

print STDOUT "\n--- Finished the Analysis. Now, print out the summary. ---\n\n";

my $fh_osm = open_output_file ($out_summ, 0);
unless (defined $fh_osm) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the summary file, '$out_summ'!\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

print $fh_osm join ("\t", '#{dec_sites}\time-lapse', @TIME_LAPSES), "\n\n";

for (my $n_D1=1; $n_D1 <= $UB_CT_DECSITES1; $n_D1++) { print $fh_osm join ("\t", $n_D1, @{$ratios[$n_D1]}), "\n"; }  

close ($fh_osm);


print STDOUT "\n\n=== The Program FINISHED !! ===\n\n";


