#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# ratio_linear_vs_quadratic_pwa_case_iii.pl:
#	This script does almost the same job as its predecessor 'ratio_linear_vs_quadratic_pwa_case_ii.pl'.
#	The only difference is that, instead of case (ii), this script examines local PWAs in case (iii),
#	that is, the case where each of them has no ancestral sites and some descendant sites
#	in between two preserved ancestral sites (PASs) that are next to each other.
#
#	[CREATED on May 12 (Tue), 2015.]
#
# ratio_linear_vs_quadratic_pwa_case_ii.pl:
#	This Perl script calculates the coefficients of the linear and the quadratic terms
#	of the expansion of the local PWA probability in {total indel rate} * {time lapse},
#	and compares the latter to the former, under fixed model parameters.
#
#	The local PWAs examined in this script are in case (ii),
#	that is, the case where each of them has some ancestral sites and no descendant sites
#	in between two preserved ancestral sites (PASs) that are next to each other.
#
#	The script was created by taking advantage of the framework of the script 'ratio_pars_vs_nextpars_pwa_case_ii.pl'.
#
#	[CREATED on May 12 (Tue), 2015.]
#
# ratio_pars_vs_nextpars_pwa_case_ii.pl:
#	This Perl script calculates the portion of the multiplication factor for a local PWA
#	contributed by the parsimonious (i.e., fewest-indel) local histories,
#	that contributed by the next-parsimonious (i.e., next-fewest-indel) local histories,
#	and the ratio of the former to the latter,
#	under fixed model parameters and under several different values of the time lapse (t_F - t_I).
#	The local PWAs examined in this script are in case (ii),
#	that is, each of them has some ancestral sites and no descendant sites
#	in between two preserved ancestral sites (PASs) that are next to each other.
#
#	[CREATED on May 2 (Sat), 2015.]
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyArrayManipu qw( diff_array );

use MyPertProb_locPWA_dawg qw(
	coeff_quad_agaps_vs_dsites
	);
	##coeff_quad_asites_vs_dgaps
	##mfac_asites_vs_dgaps_1event
	##mfac_asites_vs_dgaps_2events


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 3;


	### MODEL PARAMETERS ###

our $TOTRATE_INS = 0.1; # Total insertion rate (per unit time). (lambda_I in Dawg's indel model.)
our $TOTRATE_DEL = 0.1; # Total deletion rate (per unit time). (lambda_D in Dawg's indel model.)

our $EXPONENT_INS = 1.6; # Exponent for the power-law distribution of insertion lengths: f_I (l) is proportional to l^{-$EXPONENT_INS}.
our $EXPONENT_DEL = 1.6; # Exponent for the power-law distribution of deletion lengths: f_D (l) is proportional to l^{-$EXPONENT_DEL}.

=for the full-scale analysis (1)
our $CUTOFF_INS_LEN = 5000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 5000; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (2)
our $CUTOFF_INS_LEN = 1000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 1000; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (3)
our $CUTOFF_INS_LEN = 500; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 500; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (4)
our $CUTOFF_INS_LEN = 3000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 3000; # Cutoff value for the deletion length.
=cut
##=for the full-scale analysis (5)
our $CUTOFF_INS_LEN = 10000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 10000; # Cutoff value for the deletion length.
##=cut
=for TESTING
our $CUTOFF_INS_LEN = 200; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 200; # Cutoff value for the deletion length.
=cut

my $nfac_ins = my $nfac_del = 0;
for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $nfac_ins += exp (-$EXPONENT_INS * log ($l)); }
for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $nfac_del += exp (-$EXPONENT_DEL * log ($l)); }


our @FREQS_INS = (0); # Relative frequencies of the insertion lengths.
our @FREQS_DEL = (0); # Relative frequencies of the deletion lengths.

for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $FREQS_INS[$l] = exp (-$EXPONENT_INS * log ($l)) /$nfac_ins; }
for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $FREQS_DEL[$l] = exp (-$EXPONENT_DEL * log ($l)) /$nfac_del; }

	### Other Important Parameters ###

=for the full-scale analysis (1,2)
our $UB_CT_DECSITES = 1000; # Upper-bound of the number of descendant sites in between the PASs.
=cut
##=for the full-scale analysis (3)
our $UB_CT_DECSITES = 500; # Upper-bound of the number of descendant sites in between the PASs.
##=cut
=for TESTING
our $UB_CT_DECSITES = 100; # Upper-bound of the number of descendant sites in between the PASs.
=cut


	### Output files. ###

	# $outfile = "$outdir/$out_pref$tlapse$out_suff"

my $outdir = "Cutoff$CUTOFF_INS_LEN"; ## For the full-scale analysis.

##my $outdir = 'TESTs/Test1'; ## For TESTING the script.

my $out_summ = "$outdir/out_coeffs_ratios.txt";


	### Create the output directory. ###

unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

print STDOUT "Output file: $out_summ .\n\n";

=for the predecessor 'ratio_pars_vs_nextpars_pwa_case_ii.pl'.
print STDOUT "Detailed output files: {out_dir}/$out_pref", '{time_lapse}', "$out_suff,\n";
print STDOUT "       where {out_dir} = $outdir ,\n";
print STDOUT "             {time_lapse} = ", join (', ', @TIME_LAPSES), " .\n\n";
=cut


print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";


print STDOUT "\n<< Model Parameters >>\n\n";

print STDOUT '$TOTRATE_INS = ', $TOTRATE_INS, " ,\n";
print STDOUT '$TOTRATE_DEL = ', $TOTRATE_DEL, " ,\n\n";

print STDOUT '@FREQS_INS[l] prpt l^{', -$EXPONENT_INS, "}   (l=1,...,$CUTOFF_INS_LEN) ,\n";
print STDOUT '@FREQS_DEL[l] prpt l^{', -$EXPONENT_DEL, "}   (l=1,...,$CUTOFF_DEL_LEN) .\n\n";


if ($TEST_LEVEL>2) {

   print STDOUT "\n< Indel length frequencies >\n\n";

   print STDOUT join ("\t", 'Length', 'Ins', 'Del'), "\n\n";

   my $cutoff = ($CUTOFF_INS_LEN > $CUTOFF_DEL_LEN) ? $CUTOFF_INS_LEN : $CUTOFF_DEL_LEN ;

   for (my $l=1; $l<= $cutoff; $l++) {
	my $f_ins = ($l > $CUTOFF_INS_LEN) ? 'undef' : $FREQS_INS[$l] ;
	my $f_del = ($l > $CUTOFF_DEL_LEN) ? 'undef' : $FREQS_DEL[$l] ;

	print STDOUT join ("\t", $l, $f_ins, $f_del), "\n";
   }
   print STDOUT "\n\n";

}  # END of "if ($TEST_LEVEL>2) {...}". #

print STDOUT "\n<<< Main Analysis >>>\n\n\n";

my $fh_out = open_output_file ($out_summ, 0);
##my $fh_out = open_output_file ($outfile, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED !\n\n";
	##print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED for the time-lapse = $tlapse: $! !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

print $fh_out join ("\t", '#{dec_sites}', 'A=Coeff_lin', 'B=Coeff_quad', 'Ratio(=B/A)'), "\n\n";
##print $fh_out join ("\t", '#{anc_sites}', 'A=Coeff_lin', 'B=Coeff_quad', 'Ratio(=B/A)'), "\n\n";
####print $fh_out join ("\t", '#{anc_sites}', 'A=Mfac(pars)', 'B=Mfac(next-pars)', 'Ratio(=B/A)'), "\n\n";

for (my $n_D=1; $n_D <= $UB_CT_DECSITES; $n_D++) { # Main inner for-loop (over the descendant site counts). #

	($n_D > $CUTOFF_INS_LEN) and next;

	my $coeff_lin = $FREQS_INS[$n_D];

	my $coeff_quad = coeff_quad_agaps_vs_dsites ($n_D, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN);
	unless (defined $coeff_quad) {
	    print STDOUT "[ERROR '$program_name'] 'coeff_quad_asites_vs_dgaps' FAILED for #{dec_sites} = $n_D !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my $ratio_coeff = $coeff_quad / $coeff_lin ;
	##my $ratio_mfac = $mfac_nextpars / $mfac_pars ;

	print $fh_out join ("\t", $n_D, $coeff_lin, $coeff_quad, $ratio_coeff), "\n";
	##print $fh_out join ("\t", $n_A, $coeff_lin, $coeff_quad, $ratio_coeff), "\n";
	####print $fh_out join ("\t", $n_A, $mfac_pars, $mfac_nextpars, $ratio_mfac), "\n";

	##$ratios[$n_A]->[$indx_tlapse] = $ratio_mfac ;

}					# End of the main inner for-loop (over the ancestral site counts). #

close ($fh_out);



print STDOUT "\n\n=== The Program FINISHED !! ===\n\n";


