#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# ratio_KS_vs_ref2nd_pwa_case_iv.pl:
#	This Perl script calculates the probabilities (normalized by that in case (i)) of local PWAs in case (iv),
#	that is, the case where a pair of PAWs flank some ancestral sites and some descendant sites that are NOT homologous to each other.
#	The probabilities are calculated by the method of Kim and Sinha (2007),
#	and they are compared to the "reference" probabilities calculated up to the 2nd-order (i.e., the lowest-order) in {total indel rate} * {time lapse}.
#
#	The script was created by taking advantage of the framework of 'ratio_pars_vs_nextpars_pwa_case_iv.pl'.
#
#	[CREATED on May 12 (Tue), 2015.]
#
# ratio_pars_vs_nextpars_pwa_case_iv.pl:
#	This Perl script does basically the same jobs as its predecessor, 'ratio_pars_vs_nextpars_pwa_case_iii.pl,' does.
#	There are, howver, two major differences:
#	(i) It examines the local PWAs in case (iv), instead of those in case (iii);
#	(ii) It only examines a sparse set of grid-points, instead of the exhaustive set of combinations, ($n_A, $n_D).
#
#	[CREATED on May 4 (Mon), 2015.]
#
# ratio_pars_vs_nextpars_pwa_case_iii.pl:
#	This Perl script does nearly the same jobs as its predecessor, 'ratio_pars_vs_nextpars_pwa_case_ii.pl' does.
#	The only difference is that it examines the local PWAs in case (iii),
#	that is, each of them has no ancestral sites and some descendant sites
#	in between two preserved ancestral sites (PASs) that are next to each other.
#
#	[CREATED on May 4 (Mon), 2015.]
#
# ratio_pars_vs_nextpars_pwa_case_ii.pl:
#	This Perl script calculates the portion of the multiplication factor for a local PWA
#	contributed by the parsimonious (i.e., fewest-indel) local histories,
#	that contributed by the next-parsimonious (i.e., next-fewest-indel) local histories,
#	and the ratio of the former to the latter,
#	under fixed model parameters and under several different values of the time lapse (t_F - t_I).
#	The local PWAs examined in this script are in case (ii),
#	that is, each of them has some ancestral sites and no descendant sites
#	in between two preserved ancestral sites (PASs) that are next to each other.
#
#	[CREATED on May 2 (Sat), 2015.]
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyArrayManipu qw( diff_array );

use MyPertProb_locPWA_dawg qw(
	coeff_quad_asites_nhom_dsites
	);
	##mfac_asites_nhom_dsites_2events
	##mfac_asites_nhom_dsites_3events
	##mfac_asites_vs_dgaps_1event
	##mfac_asites_vs_dgaps_2events
	##mfac_agaps_vs_dsites_1event
	##mfac_agaps_vs_dsites_2events


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 3;


	### MODEL PARAMETERS ###

our $TOTRATE_INS = 0.1; # Total insertion rate (per unit time). (lambda_I in Dawg's indel model.)
our $TOTRATE_DEL = 0.1; # Total deletion rate (per unit time). (lambda_D in Dawg's indel model.)

our $EXPONENT_INS = 1.6; # Exponent for the power-law distribution of insertion lengths: f_I (l) is proportional to l^{-$EXPONENT_INS}.
our $EXPONENT_DEL = 1.6; # Exponent for the power-law distribution of deletion lengths: f_D (l) is proportional to l^{-$EXPONENT_DEL}.

=for the full-scale analysis (1)
our $CUTOFF_INS_LEN = 5000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 5000; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (2)
our $CUTOFF_INS_LEN = 1000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 1000; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (3)
our $CUTOFF_INS_LEN = 500; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 500; # Cutoff value for the deletion length.
=cut
#=for the full-scale analysis (5)
our $CUTOFF_INS_LEN = 10000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 10000; # Cutoff value for the deletion length.
#=cut
=for TESTING
our $CUTOFF_INS_LEN = 200; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 200; # Cutoff value for the deletion length.
=cut

my $nfac_ins = my $nfac_del = 0;
for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $nfac_ins += exp (-$EXPONENT_INS * log ($l)); }
for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $nfac_del += exp (-$EXPONENT_DEL * log ($l)); }


our @FREQS_INS = (0); # Relative frequencies of the insertion lengths.
our @FREQS_DEL = (0); # Relative frequencies of the deletion lengths.

for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $FREQS_INS[$l] = exp (-$EXPONENT_INS * log ($l)) /$nfac_ins; }
for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $FREQS_DEL[$l] = exp (-$EXPONENT_DEL * log ($l)) /$nfac_del; }


	### Other Important Parameters ###

	# Pairs of the number of ancestral sites and the number of descendant sites.
	# (ADDED in creation of 'ratio_pars_vs_nextpars_pwa_case_iv.pl' on May 4 (Mon), 2015.)
	#
##=for the full-scale analysis
our @SET_PAIRS_ADSITENUMS = ( # element = [$n_A, $n_D].
	[1, 1], [3, 1], [1, 3], [3, 3], [10, 1], [1, 10],
	[25, 1], [5, 5], [1, 25], [100, 1], [25, 4], [10, 10], [4, 25], [1, 100],
	[300, 1], [100, 3], [30, 10], [10, 30], [3, 100], [1, 300]
	);
##=cut
=For TESTING
our @SET_PAIRS_ADSITENUMS = ( # element = [$n_A, $n_D].
	[1, 1], [10, 1], [1, 10], [100, 1], [10, 10], [1, 100]
	);
=cut

our $CT_PAIRS = @SET_PAIRS_ADSITENUMS;

	### Output files. ###

	# $outfile = "$outdir/$out_pref$tlapse$out_suff"

my $outdir = "Cutoff$CUTOFF_INS_LEN"; ## For the full-scale analysis.
##my $outdir = 'Fixed_Time_Lapse'; ## For the full-scale analysis.
##my $outdir = 'TESTs/Test1'; ## For TESTING the script.

my $out_summ = "$outdir/out_coeffs_ratios.txt";
##my $out_summ = "$outdir/out_ratios_for_varied_tlapse.txt";


	### Create the output directory. ###

unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

print STDOUT "Output file: $out_summ .\n\n";
##print STDOUT "Summary output file: $out_summ .\n\n";


print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";


print STDOUT "\n<< Model Parameters >>\n\n";

print STDOUT '$TOTRATE_INS = ', $TOTRATE_INS, " ,\n";
print STDOUT '$TOTRATE_DEL = ', $TOTRATE_DEL, " ,\n\n";

print STDOUT '@FREQS_INS[l] prpt l^{', -$EXPONENT_INS, "}   (l=1,...,$CUTOFF_INS_LEN) ,\n";
print STDOUT '@FREQS_DEL[l] prpt l^{', -$EXPONENT_DEL, "}   (l=1,...,$CUTOFF_DEL_LEN) .\n\n";


	# ADDED in creation of 'ratio_pars_vs_nextpars_pwa_case_iv.pl' on May 4 (Mon), 2015. #
	#
print STDOUT "\n<< Examined grid points (#{points} = $CT_PAIRS) >>\n\n";

print STDOUT '(#{ancestral_sites}, #{descendant_sites}) = ';

foreach my $pair (@SET_PAIRS_ADSITENUMS) { printf STDOUT "  (%d, %d), ", @{$pair}; } 
print STDOUT " .\n\n";


if ($TEST_LEVEL>2) {

   print STDOUT "\n< Indel length frequencies >\n\n";

   print STDOUT join ("\t", 'Length', 'Ins', 'Del'), "\n\n";

   my $cutoff = ($CUTOFF_INS_LEN > $CUTOFF_DEL_LEN) ? $CUTOFF_INS_LEN : $CUTOFF_DEL_LEN ;

   for (my $l=1; $l<= $cutoff; $l++) {
	my $f_ins = ($l > $CUTOFF_INS_LEN) ? 'undef' : $FREQS_INS[$l] ;
	my $f_del = ($l > $CUTOFF_DEL_LEN) ? 'undef' : $FREQS_DEL[$l] ;

	print STDOUT join ("\t", $l, $f_ins, $f_del), "\n";
   }
   print STDOUT "\n\n";

}  # END of "if ($TEST_LEVEL>2) {...}". #

print STDOUT "\n<<< Main Analysis >>>\n\n\n";


my $fh_out = open_output_file ($out_summ, 0);
##my $fh_out = open_output_file ($outfile, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED !\n\n";
	##print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED for the time-lapse = $tlapse: $! !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

print $fh_out join ("\t", '#{anc_sites}', '#{dec_sites}', 'A=Coeff_KS', 'B=Coeff_ref2nd', 'Ratio(=B/A)'), "\n\n";
##print $fh_out join ("\t", '#{anc_sites}', '#{dec_sites}', 'A=Mfac(pars)', 'B=Mfac(next-pars)', 'Ratio(=B/A)'), "\n\n";

		# MODIFIED in creation of 'ratio_pars_vs_nextpars_pwa_case_iv.pl' (on May 4 (Mon), 2015). #
for (my $i = 0; $i < $CT_PAIRS; $i++) { # Main inner for-loop (over the pairs of #{ancsites} and #{decsites}). #
##for (my $n_A=1; $n_A <= $UB_CT_ANCSITES; $n_A++) { # Main inner for-loop (over the ancestral site counts). #

	my ($n_A, $n_D) = @{$SET_PAIRS_ADSITENUMS[$i]};

	(($n_A <= $CUTOFF_DEL_LEN) and ($n_D <= $CUTOFF_INS_LEN)) or next; 


	print STDOUT '(#{ancsites}, #{decsites}) = ', "($n_A, $n_D) .\n\n";

	my $coeff_KS = $FREQS_DEL[$n_A] * $FREQS_INS[$n_D] ; 


	my $coeff_ref2nd = coeff_quad_asites_nhom_dsites ($n_A, $n_D, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) ;

	unless (defined $coeff_ref2nd) {
	    print STDOUT "[ERROR '$program_name'] 'coeff_quad_asites_nhom_dsites' FAILED for (#{ancsites}, #{decsites}) = ($n_A, $n_D) !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}


	my $ratio_coeff = $coeff_ref2nd / $coeff_KS ;
	##my $ratio_mfac = $mfac_nextpars / $mfac_pars ;

	print $fh_out join ("\t", $n_A, $n_D, $coeff_KS, $coeff_ref2nd, $ratio_coeff), "\n";
	##print $fh_out join ("\t", $n_A, $n_D, $mfac_pars, $mfac_nextpars, $ratio_mfac), "\n";

	##$ratios[$i]->[$indx_tlapse] = $ratio_mfac ;
	####$ratios[$n_A]->[$indx_tlapse] = $ratio_mfac ;

}				# End of the main inner for-loop (over the pairs of #{ancsites} and #{decsites}). #

close ($fh_out);



print STDOUT "\n\n=== The Program FINISHED !! ===\n\n";


