#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# calc_fa_log_mfacs_pars_pwa_dawg.pl:
#	This Perl script calculates the multiplication factors for the local PWAs
#	under the first-order approximation, where only the parsimonious local indel histories are taken into account.
#	The factors are calculated under Dawg's indel model, whose rate parameters are space- and time-homogeneous.
#
#	The script was created by taking advantage of some elements in the scripts, 'ratio_pars_vs_nextpars_pwa_case_ii/iii/iv.pl'.
#
#	[CREATION STARTED on May 16 (Sat), 2015, ENDED on May xx (yyy), 2015.]
#
# ratio_pars_vs_nextpars_pwa_case_iv.pl:
#	This Perl script does basically the same jobs as its predecessor, 'ratio_pars_vs_nextpars_pwa_case_iii.pl,' does.
#	There are, howver, two major differences:
#	(i) It examines the local PWAs in case (iv), instead of those in case (iii);
#	(ii) It only examines a sparse set of grid-points, instead of the exhaustive set of combinations, ($n_A, $n_D).
#
#	[CREATED on May 4 (Mon), 2015.]
#
# ratio_pars_vs_nextpars_pwa_case_iii.pl:
#	This Perl script does nearly the same jobs as its predecessor, 'ratio_pars_vs_nextpars_pwa_case_ii.pl' does.
#	The only difference is that it examines the local PWAs in case (iii),
#	that is, each of them has no ancestral sites and some descendant sites
#	in between two preserved ancestral sites (PASs) that are next to each other.
#
#	[CREATED on May 4 (Mon), 2015.]
#
# ratio_pars_vs_nextpars_pwa_case_ii.pl:
#	This Perl script calculates the portion of the multiplication factor for a local PWA
#	contributed by the parsimonious (i.e., fewest-indel) local histories,
#	that contributed by the next-parsimonious (i.e., next-fewest-indel) local histories,
#	and the ratio of the former to the latter,
#	under fixed model parameters and under several different values of the time lapse (t_F - t_I).
#	The local PWAs examined in this script are in case (ii),
#	that is, each of them has some ancestral sites and no descendant sites
#	in between two preserved ancestral sites (PASs) that are next to each other.
#
#	[CREATED on May 2 (Sat), 2015.]
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyArrayManipu qw( diff_array );

use MyPertProb_locPWA_dawg qw(
	mfac_asites_vs_dgaps_1event
	mfac_agaps_vs_dsites_1event
	tld_mu_P_pwa_e tld_mu_P_pwa_fg
	mfac_asites_nhom_dsites_2events
	);
	##mfac_asites_vs_dgaps_2events
	##mfac_agaps_vs_dsites_2events
	##mfac_asites_nhom_dsites_3events

use MyTreeMap_indels_ML_hs qw(
	cal_delta_exp_del_5p2
	);


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 3;


	### MODEL PARAMETERS ###

our $DISTRI = 'PL'; # Power-law distributions.
##our $DISTRI = 'GM'; # Geometric distributions. 

our $TOTRATE_INS = 0.1; # Total insertion rate (per unit time). (lambda_I in Dawg's indel model.)
our $TOTRATE_DEL = 0.1; # Total deletion rate (per unit time). (lambda_D in Dawg's indel model.)

our $EXPONENT_INS = 1.6; # Exponent for the power-law distribution of insertion lengths: f_I (l) is proportional to l^{-$EXPONENT_INS}.
our $EXPONENT_DEL = 1.6; # Exponent for the power-law distribution of deletion lengths: f_D (l) is proportional to l^{-$EXPONENT_DEL}.

our $PROB_EXT_INS = 0.5; # Extention probability for the geometric distribution of insertion lengths: f_I (l) is proportional to $PROB_EXT_INS^{l}.
our $PROB_EXT_DEL = 0.5; # Extention probability for the geometric distribution of deletion lengths: f_D (l) is proportional to $PROB_EXT_DEL^{l}.

=for the full-scale analysis (1)
our $CUTOFF_INS_LEN = 5000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 5000; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (2)
our $CUTOFF_INS_LEN = 1000; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 1000; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (3)
our $CUTOFF_INS_LEN = 500; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 500; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (5)
our $CUTOFF_INS_LEN = 200; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 200; # Cutoff value for the deletion length.
=cut
##=for the full-scale analysis (10)
our $CUTOFF_INS_LEN = 100; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 100; # Cutoff value for the deletion length.
##=cut
=for TESTING the script.
our $CUTOFF_INS_LEN = 20; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 20; # Cutoff value for the deletion length.
=cut


	# Calculate the insertion/deletion length frequencies. #
my $nfac_ins = my $nfac_del = 0;

if ($DISTRI eq 'PL') {
   for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $nfac_ins += exp (-$EXPONENT_INS * log ($l)); }
   for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $nfac_del += exp (-$EXPONENT_DEL * log ($l)); }
} elsif ($DISTRI eq 'GM') {
   for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $nfac_ins += exp ( $l * log($PROB_EXT_INS) ); }
   for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $nfac_ins += exp ( $l * log($PROB_EXT_DEL) ); }
} else {
    print STDOUT "[ERROR '$program_name'] ILLEGAL PARAMETER: ", '$DISTRI = ', "$DISTRI !\n\n";
    print STDERR "[ERROR '$program_name'] $!\n";
    die ;
}


our @FREQS_INS = (0); # Relative frequencies of the insertion lengths.
our @FREQS_DEL = (0); # Relative frequencies of the deletion lengths.

if ($DISTRI eq 'PL') {
   for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $FREQS_INS[$l] = exp (-$EXPONENT_INS * log ($l)) /$nfac_ins; }
   for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $FREQS_DEL[$l] = exp (-$EXPONENT_DEL * log ($l)) /$nfac_del; }
} elsif ($DISTRI eq 'GM') {
   for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $FREQS_INS[$l] = exp ( $l * log ($PROB_EXT_INS) ) /$nfac_ins; }
   for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $FREQS_DEL[$l] = exp ( $l * log ($PROB_EXT_DEL) ) /$nfac_del; }
}

	### Other Important Parameters ###

our @TIME_LAPSES = (0.02, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0); # Values of the time lapse (= t_F - t_I) to be examined. # FOr the full-scale calculation.
##our @TIME_LAPSES = (0.02, 0.05, 0.1, 0.2, 0.3, 0.5, 0.75, 1.0); # Values of the time lapse (= t_F - t_I) to be examined. # FOr the full-scale calculation.
##our @TIME_LAPSES = (0.05, 0.2, 0.5, 1.0); # Values of the time lapse (= t_F - t_I) to be examined. # For TESTING the script. #

our @E_INDEL_CTS = (); # Expected counts of indels. #
foreach my $tlapse (@TIME_LAPSES) { push @E_INDEL_CTS, ($TOTRATE_INS + $TOTRATE_DEL) * $tlapse; }

=for the full-scale analysis (1,2,3)
our $UB_CT_ANCSITES = 500; # Upper-bound of the number of ancestral sites in between the PASs.
our $UB_CT_DECSITES = 500; # Upper-bound of the number of descendant sites in between the PASs.
=cut
##=for the full-scale analysis (3,5,10)
our $UB_CT_ANCSITES = 100; # Upper-bound of the number of ancestral sites in between the PASs.
our $UB_CT_DECSITES = 100; # Upper-bound of the number of descendant sites in between the PASs.
##=cut
=for TESTING the script.
our $UB_CT_ANCSITES = 20; # Upper-bound of the number of ancestral sites in between the PASs.
our $UB_CT_DECSITES = 20; # Upper-bound of the number of descendant sites in between the PASs.
=cut



	### Output files. ###

my $outdir = ($DISTRI eq 'PL') ? "$DISTRI${EXPONENT_INS}_co$CUTOFF_INS_LEN" : "$DISTRI${PROB_EXT_INS}_co$CUTOFF_INS_LEN" ; ## For the full-scale analysis.
##my $outdir = 'TESTs/Test1'; ## For TESTING the script.

my $out_deldel = "$outdir/delta_del.txt";

	# For the bulk multiplication factors. #
my $out_as_dg = "$outdir/lmfacs_asites_vs_dgaps.txt";
my $out_ag_ds = "$outdir/lmfacs_agaps_vs_dsites.txt";
my $out_as_ds = "$outdir/lmfacs_asites_nhom_dsites.txt";

	# For the boundary multiplication factors. #
my $out_bd_as_dg = "$outdir/bd_lmfacs_asites_vs_dgaps.txt";
my $out_bd_ag_ds = "$outdir/bd_lmfacs_agaps_vs_dsites.txt";
my $out_bd_as_ds = "$outdir/bd_lmfacs_asites_nhom_dsites.txt";


	### Create the output directory. ###

unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

print STDOUT "Output directory: $outdir .\n\n";

print STDOUT "Output files:\n\n";

print STDOUT "   (Delta_del) $out_deldel ,\n\n";

print STDOUT "   (log-multiplication factors in the bulk)\n";
print STDOUT "     (for ancestral sites vs descendant gaps) $out_as_dg ,\n";
print STDOUT "     (for ancestral gaps vs descendant sites) $out_ag_ds ,\n";
print STDOUT "     (for ancestral sites & non-homologous descendant sites) $out_as_ds ,\n\n";

print STDOUT "   (log-multiplication factors on the PWA boundaries)\n";
print STDOUT "     (for ancestral sites vs descendant gaps) $out_bd_as_dg ,\n";
print STDOUT "     (for ancestral gaps vs descendant sites) $out_bd_ag_ds ,\n";
print STDOUT "     (for ancestral sites & non-homologous descendant sites) $out_bd_as_ds .\n\n\n";


print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " .\n\n";

print STDOUT "\n<< Domains within which the multiplication factors will be calculated >>\n\n";

print STDOUT '#{Ancestral sites}  = (0,) 1, ..., ', $UB_CT_ANCSITES, " ,\n";
print STDOUT '#{Descendant sites} = (0,) 1, ..., ', $UB_CT_DECSITES, " .\n\n";

print STDOUT '@TIME_LAPSES = (', join (', ', @TIME_LAPSES), ") .\n\n";

print STDOUT '@E_INDEL_CTS = (', join (', ', @E_INDEL_CTS), ") .\n\n";


print STDOUT "\n<< Model Parameters >>\n\n";

print STDOUT '$TOTRATE_INS = ', $TOTRATE_INS, " ,\n";
print STDOUT '$TOTRATE_DEL = ', $TOTRATE_DEL, " ,\n\n";

print STDOUT '$DISTRI = ', $DISTRI, " .\n\n";

if ($DISTRI eq 'PL') {
   print STDOUT '@FREQS_INS[l] prpt l^{', -$EXPONENT_INS, "}   (l=1,...,$CUTOFF_INS_LEN) ,\n";
   print STDOUT '@FREQS_DEL[l] prpt l^{', -$EXPONENT_DEL, "}   (l=1,...,$CUTOFF_DEL_LEN) .\n\n";
} elsif ($DISTRI eq 'GM') {
   print STDOUT '@FREQS_INS[l] prpt ', $PROB_EXT_INS, "^{l}   (l=1,...,$CUTOFF_INS_LEN) ,\n";
   print STDOUT '@FREQS_DEL[l] prpt ', $PROB_EXT_DEL, "^{l}   (l=1,...,$CUTOFF_DEL_LEN) .\n\n";
}


print STDOUT "\n<<< Preliminary Process: Calculating the Indel length frequencies on the boundaries >>>\n\n";

our @FREQS_INS_BD = @FREQS_INS; # Under Dawg's model, the insertion length frequency remains the same whether the insertion is in the bulk or on either boundary.

our @FREQS_DEL_BD = ();

for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { # Preliminary for-loop (over deletion lenghts). #

   my $freq_bd = 0;
   for (my $lp= $l; $lp <= $CUTOFF_DEL_LEN; $lp++) { $freq_bd += $FREQS_DEL[$lp]; }

   $FREQS_DEL_BD[$l] = $freq_bd;
}				# End of the preliminary for-loop (over deletion lengths). #


if ($TEST_LEVEL>2) {

   print STDOUT "\n< Indel length frequencies >\n\n";

   print STDOUT join ("\t", 'Length', 'Ins', 'Del', 'Ins(bd)', 'Del(bd)'), "\n\n";

   my $cutoff = ($CUTOFF_INS_LEN > $CUTOFF_DEL_LEN) ? $CUTOFF_INS_LEN : $CUTOFF_DEL_LEN ;

   for (my $l=1; $l<= $cutoff; $l++) {
	my $f_ins = ($l > $CUTOFF_INS_LEN) ? 'undef' : $FREQS_INS[$l] ;
	my $f_del = ($l > $CUTOFF_DEL_LEN) ? 'undef' : $FREQS_DEL[$l] ;
	my $f_ins_bd = ($l > $CUTOFF_INS_LEN) ? 'undef' : $FREQS_INS_BD[$l] ;
	my $f_del_bd = ($l > $CUTOFF_DEL_LEN) ? 'undef' : $FREQS_DEL_BD[$l] ;

	print STDOUT join ("\t", $l, $f_ins, $f_del, $f_ins_bd, $f_del_bd), "\n";
   }
   print STDOUT "\n\n";

}  # END of "if ($TEST_LEVEL>2) {...}". #


print STDOUT "\n<<< Main Process 0: Calculating the 'universal' exponent factor >>>\n\n\n";

my $delta_del = cal_delta_exp_del_5p2 (@FREQS_DEL, $CUTOFF_DEL_LEN);
unless (defined $delta_del) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'cal_delta_exp_dep_5p2' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $fh_out0 = open_output_file ($out_deldel, 0);
unless (defined $fh_out0) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file, '$out_deldel' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out0 "* Delta_del := Sum_{l=1}^{$CUTOFF_DEL_LEN} ", '(l-1) * $FREQ_DEL_LEN[l]', " .\n";
print $fh_out0 "\n";

print $fh_out0 "Delta_del = $delta_del\n\n";

close ($fh_out0);


print STDOUT "\n<<< Main Process 1: Calculating the log-multiplication factors (ancestral sites vs descendant gaps; bulk) >>>\n\n";

my $fh_out1 = open_output_file ($out_as_dg, 0);
unless (defined $fh_out1) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file, '$out_as_dg' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out1 join ("\t", '#{Anc_sites}\E[#{indels}]', @E_INDEL_CTS), "\n\n";

for (my $n_A = 1; $n_A <= $UB_CT_ANCSITES; $n_A++) { # 1st main for-loop (over #{ancestral sites}). #

	my @lmfacs = ();

	foreach my $tlapse (@TIME_LAPSES) { # Inner foreach-loop (over time-lapses). #

	    my $mfac = mfac_asites_vs_dgaps_1event ($n_A, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_DEL, $CUTOFF_DEL_LEN) ;
	    unless (defined $mfac) {
		print STDOUT "[ERROR '$program_name'] 'mfac_asites_vs_dgaps_1event' FAILED on ", '($n_A, $tlapse) = ', "($n_A, $tlapse) !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    push @lmfacs, ($mfac > 0) ? log ($mfac) : 'log(zero)' ;

	}			# End of the inner foreach-loop (over time-lapses). #

	print $fh_out1 join ("\t", $n_A, @lmfacs), "\n";

}	# End of the 1st main for-loop (over #{ancestral sites}). #

close ($fh_out1);


print STDOUT "\n<<< Main Process 2: Calculating the log-multiplication factors (ancestral gaps vs descendant sites; bulk) >>>\n\n";

my $fh_out2 = open_output_file ($out_ag_ds, 0);
unless (defined $fh_out2) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file, '$out_ag_ds' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out2 join ("\t", '#{Dec_sites}\E[#{indels}]', @E_INDEL_CTS), "\n\n";

for (my $n_D = 1; $n_D <= $UB_CT_DECSITES; $n_D++) { # 2nd main for-loop (over #{descendant sites}). #

	my @lmfacs = ();

	foreach my $tlapse (@TIME_LAPSES) { # Inner foreach-loop (over time-lapses). #

	    my $mfac = mfac_agaps_vs_dsites_1event ($n_D, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, $CUTOFF_INS_LEN) ;
	    unless (defined $mfac) {
		print STDOUT "[ERROR '$program_name'] 'mfac_agaps_vs_dsites_1event' FAILED on ", '($n_D, $tlapse) = ', "($n_D, $tlapse) !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    push @lmfacs, ($mfac > 0) ? log ($mfac) : 'log(zero)' ;

	}			# End of the inner foreach-loop (over time-lapses). #

	print $fh_out2 join ("\t", $n_D, @lmfacs), "\n";

}	# End of the 2nd main for-loop (over #{descendant sites}). #

close ($fh_out2);



print STDOUT "\n<<< Main Process 3: Calculating the log-multiplication factors (ancestral sites & non-homologous descendant sites; bulk) >>>\n\n";

my $fh_out3 = open_output_file ($out_as_ds, 0);
unless (defined $fh_out3) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file, '$out_as_ds' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out3 join ("\t", '#{Anc_sites}:#{dec_sites}\E[#{indels}]', @E_INDEL_CTS), "\n\n";

for (my $n_A = 1; $n_A <= $UB_CT_ANCSITES; $n_A++) { # 3rd main outer for-loop (over #{ancestral sites}). #
   for (my $n_D = 1; $n_D <= $UB_CT_DECSITES; $n_D++) { # main middle for-loop (over #{descendant sites}). #

	my @lmfacs = ();

	foreach my $tlapse (@TIME_LAPSES) { # Inner foreach-loop (over time-lapses). #

	    my ($mfac, $components) = mfac_asites_nhom_dsites_2events ($n_A, $n_D, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) ;
	    unless (defined $mfac) {
		print STDOUT "[ERROR '$program_name'] 'mfac_asites_nhom_dsites_2events' FAILED on ", '($n_A, $n_D, $tlapse) = ', "($n_A, $n_D, $tlapse) !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    push @lmfacs, ($mfac > 0) ? log ($mfac) : 'log(zero)' ;

	}			# End of the inner foreach-loop (over time-lapses). #

	print $fh_out3 join ("\t", "$n_A:$n_D", @lmfacs), "\n";

    }		# End of the main middle for-loop (over #{descendant sites}). #
}	# End of the 3rd main outer for-loop (over #{ancestral sites}). #

close ($fh_out3);


print STDOUT "\n<<< Main Process 4: Calculating the log-multiplication factors (ancestral sites vs descendant gaps; on either boundary) >>>\n\n";

my $fh_out4 = open_output_file ($out_bd_as_dg, 0);
unless (defined $fh_out4) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file, '$out_bd_as_dg' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out4 "* This file lists the logarithmic multiplication factors on either PWA boundary.\n\n";

print $fh_out4 join ("\t", '#{Anc_sites}\E[#{indels}]', @E_INDEL_CTS), "\n\n";

for (my $n_A = 1; $n_A <= $UB_CT_ANCSITES; $n_A++) { # 4th main for-loop (over #{ancestral sites}). #

	my @lmfacs = ();

	foreach my $tlapse (@TIME_LAPSES) { # Inner foreach-loop (over time-lapses). #

	    my $mfac = mfac_asites_vs_dgaps_1event ($n_A, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_DEL_BD, $CUTOFF_DEL_LEN) ;
	    unless (defined $mfac) {
		print STDOUT "[ERROR '$program_name'] 'mfac_asites_vs_dgaps_1event' (with the deletion on a boundary) FAILED on ", '($n_A, $tlapse) = ', "($n_A, $tlapse) !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    push @lmfacs, ($mfac > 0) ? log ($mfac) : 'log(zero)' ;

	}			# End of the inner foreach-loop (over time-lapses). #

	print $fh_out4 join ("\t", $n_A, @lmfacs), "\n";

}	# End of the 4th main for-loop (over #{ancestral sites}). #

close ($fh_out4);


print STDOUT "\n<<< Main Process 5: Calculating the log-multiplication factors (ancestral gaps vs descendant sites; on either boundary) >>>\n\n";

my $fh_out5 = open_output_file ($out_bd_ag_ds, 0);
unless (defined $fh_out5) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file, '$out_bd_ag_ds' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out5 "* This file lists the logarithmic multiplication factors on either PWA boundary.\n\n";
print $fh_out5 join ("\t", '#{Dec_sites}\E[#{indels}]', @E_INDEL_CTS), "\n\n";

for (my $n_D = 1; $n_D <= $UB_CT_DECSITES; $n_D++) { # 5th main for-loop (over #{descendant sites}). #

	my @lmfacs = ();

	foreach my $tlapse (@TIME_LAPSES) { # Inner foreach-loop (over time-lapses). #

	    my $mfac = mfac_agaps_vs_dsites_1event ($n_D, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS_BD, $CUTOFF_INS_LEN) ;
	    unless (defined $mfac) {
		print STDOUT "[ERROR '$program_name'] 'mfac_agaps_vs_dsites_1event' (with the insertion on a boundary) FAILED on ", '($n_D, $tlapse) = ', "($n_D, $tlapse) !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    push @lmfacs, ($mfac > 0) ? log ($mfac) : 'log(zero)' ;

	}			# End of the inner foreach-loop (over time-lapses). #

	print $fh_out5 join ("\t", $n_D, @lmfacs), "\n";

}	# End of the 2nd main for-loop (over #{descendant sites}). #

close ($fh_out5);



print STDOUT "\n<<< Main Process 6: Calculating the log-multiplication factors (ancestral sites & non-homologous descendant sites; on either boundary) >>>\n\n";

my $fh_out6 = open_output_file ($out_bd_as_ds, 0);
unless (defined $fh_out6) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file, '$out_bd_as_ds' !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out6 "* This file lists the logarithmic multiplication factors on either PWA boundary.\n\n";

print $fh_out6 join ("\t", '#{Anc_sites}:#{dec_sites}\E[#{indels}]', @E_INDEL_CTS), "\n\n";

for (my $n_A = 1; $n_A <= $UB_CT_ANCSITES; $n_A++) { # 6th main outer for-loop (over #{ancestral sites}). #
   for (my $n_D = 1; $n_D <= $UB_CT_DECSITES; $n_D++) { # main middle for-loop (over #{descendant sites}). #

	my @lmfacs = ();

	foreach my $tlapse (@TIME_LAPSES) { # Inner foreach-loop (over time-lapses). #

	    my $mfac1 = tld_mu_P_pwa_e ($n_A, $n_D, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS_BD, @FREQS_DEL_BD, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) ;
	    unless (defined $mfac1) {
		print STDOUT "[ERROR '$program_name'] 'tld_mu_P_pwa_e' (with both the deletion and the insetion on a boundary) FAILED on ", '($n_A, $n_D, $tlapse) = ', "($n_A, $n_D, $tlapse) !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    my $mfac2 = tld_mu_P_pwa_fg ($n_A, $n_D, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL_BD, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) ;
	    unless (defined $mfac2) {
		print STDOUT "[ERROR '$program_name'] 'tld_mu_P_pwa_fg' (with the deletion on a boundary) FAILED on ", '($n_A, $n_D, $tlapse) = ', "($n_A, $n_D, $tlapse) !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    my $mfac3 = tld_mu_P_pwa_fg ($n_A, $n_D, $tlapse, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS_BD, @FREQS_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) ;
	    unless (defined $mfac3) {
		print STDOUT "[ERROR '$program_name'] 'tld_mu_P_pwa_fg' (with the insertion on a boundary) FAILED on ", '($n_A, $n_D, $tlapse) = ', "($n_A, $n_D, $tlapse) !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    my $mfac = $mfac1 + $mfac2 + $mfac3 ;
    
	    push @lmfacs, ($mfac > 0) ? log ($mfac) : 'log(zero)' ;

	}			# End of the inner foreach-loop (over time-lapses). #

	print $fh_out6 join ("\t", "$n_A:$n_D", @lmfacs), "\n";

    }		# End of the main middle for-loop (over #{descendant sites}). #
}	# End of the 6th main outer for-loop (over #{ancestral sites}). #

close ($fh_out6);


print STDOUT "\n\n\n=== ALL Processes FINISHED !! ===\n\n\n";


