package MyCluster_distribution;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyCluster_distribution.pm: a package to construct and print distribution tables of clusters,
#				according to the cluster size.
#
# REVISED on Feb 20 (Thu), 2003: I added two modules, "batch_initializer" and "batch_pr_distri",
#				which manipulates multiple distributions.
#

use strict;
use warnings;
use Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
	bin_initializer cluster_distributor  pr_cluster_distri
	batch_initializer batch_distributor batch_pr_distri
		);

our @EXPORT_OK = qw(
	cluster_distributor_wgt batch_distributor_wgts
	pr_file_cluster_distri
	batch_pr_file_distri batch_pr_file_distri_xl
	batch_pr_file_distri_fmt batch_pr_file_distri_xl_fmt
	array_distributor window_distributor init_2dim_distri distributor_2dim
	pr_file_2dim_distri pr_file_2dim_distri_xl
	create_key_intervals classify_by_measure
		);


# bin_initializer (\@\@) : a module which initializes "bins" by referencing the number of boundaries #
#
sub bin_initializer (\@\@) {

	my ($rf_bins, $rf_boundaries) = @_;
	my $num_bins = scalar(@{$rf_boundaries}) + 1;
	for (my $i = 0; $i < $num_bins; $i++) { ${$rf_bins}[$i] = 0;}

}						# end of the module bin_initializer #

# batch_initializer(\@\@) : a module which initializes MULTIPLE "bins"
#
sub batch_initializer (\@\@) {

	my ($rf_array_bins, $rf_boundaries) = @_;
	foreach my $rf_bins (@{$rf_array_bins}) {

		bin_initializer (@{$rf_bins}, @{$rf_boundaries});
	}
}						# end of the module batch_initializer #

 
# cluster_distributor ($\@\@)  : a module which distributes clusters into "bin"s of different sizes #
#
# PREREQUISITE1: size of @{$rf_boundaries} should be less than the size of @{$rf_bins} by 1.
# PREREQUISITE2: @{$rf_bins} shoud be initialized to (0,0,...,0).
# PREREQUISITE3: elements of @{$rf_boundaries} should be in increasing order.
#
sub cluster_distributor ($\@\@) {

	my ($size, $rf_bins, $rf_boundaries) = @_;
	my $num_boundaries = @{$rf_boundaries};

	if ($size < ${$rf_boundaries}[0] ) {
		${$rf_bins}[0]++;
		return 1;
	}

	for (my $i = 1; $i < $num_boundaries; $i++) {

		if ( $size < ${$rf_boundaries}[$i]) {
			${$rf_bins}[$i]++;
			return 1;
		}
	}
	
	if( $size >= ${$rf_boundaries}[$num_boundaries -1]) {
		${$rf_bins}[$num_boundaries]++;
		return 1;
	} else {
		return 0
	}
}						# end of the module cluster_distributor #

 
# cluster_distributor_wgt ($\@\@$):
#	This revised version of the subroutine 'cluster_distributor' takes
#	a weighted distribution by adding the weight factor specified by the 3rd argument
#	to the relevant bin.
#
# PREREQUISITE1: size of @{$rf_boundaries} should be less than the size of @{$rf_bins} by 1.
# PREREQUISITE2: @{$rf_bins} shoud be initialized to (0,0,...,0).
# PREREQUISITE3: elements of @{$rf_boundaries} should be in increasing order.
#
sub cluster_distributor_wgt ($\@\@$) {

	my ($size, $rf_bins, $rf_boundaries, $weight) = @_;
	my $num_boundaries = @{$rf_boundaries};

	if ($size < ${$rf_boundaries}[0] ) {
		${$rf_bins}[0] += $weight;
		return 1;
	}

	for (my $i = 1; $i < $num_boundaries; $i++) {

		if ( $size < ${$rf_boundaries}[$i]) {
			${$rf_bins}[$i] += $weight;
			return 1;
		}
	}
	
	if( $size >= ${$rf_boundaries}[$num_boundaries -1]) {
		${$rf_bins}[$num_boundaries] += $weight;
		return 1;
	} else {
		return 0
	}
}						# end of the module cluster_distributor_wgt #


# batch_distributor: This subroutine distributes several quantities into
#			 corresponding sets of bins at a time.
#
# PREREQUISITE: All sets of bins need to be delimited by the same boundary values.
#
sub batch_distributor (\@\@\@\@) {

	my ($rf_quantities, $rf_batch_bins, $rf_bound, $rf_magratios) = @_;

	if (@{$rf_magratios}< @{$rf_batch_bins}) {
		for (my $i=@{$rf_magratios}; $i<@{$rf_batch_bins}; $i++) { $rf_magratios->[$i] = 1; }
	}

	for (my $i=0; $i<@{$rf_batch_bins}; $i++) {

		(defined $rf_quantities->[$i]) or next;

		cluster_distributor($rf_magratios->[$i]*$rf_quantities->[$i], @{$rf_batch_bins->[$i]}, @{$rf_bound});
	}

	return 1;

}				# end of the subroutine batch_distributor #


# batch_distributor_wgts:
#	This subroutine distributes several different weights into
#	corresponding sets of bins at a time,
#	under a single quantity to be classified.
#
# PREREQUISITE: All sets of bins need to be delimited by the same boundary values.
#
# INPUT: ($quantity, \@batch_bins, \@bounds, \@wgts), where
#	$quantity is the quantity to be classified according to
#	a set of boundaries defined by @bounds;
#	$wgts[$i] will be distributed into bins in  @{$batch_bins[$i]},
#	according to the class of $quantity.
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub batch_distributor_wgts ($\@\@\@) {

	my ($quantity, $rf_batch_bins, $rf_bounds, $rf_wgts) = @_;

	my $ct_binsets = @{$rf_batch_bins};
	unless ($ct_binsets == @{$rf_wgts}) {
	    print STDERR "[ERROR 'batch_distributor_wgts'] The number of weights (", scalar (@{$rf_wgts}), ") CLASHES with the number of the sets of bins ($ct_binsets): $! !\n";
	    return 0;
	}

	for (my $i=0; $i<$ct_binsets; $i++) {
	    my $wgt = $rf_wgts->[$i];
	    unless ((defined $wgt) and ($wgt >0)) {
	
		(defined $wgt) or ($wgt = 'undef');
		print STDERR "[ERROR 'batch_distributor_wgts'] ILLEGAL WEIGHT FACTOR ($wgt) for the $i th set of bins: $! !\n";
		return 0;
	    }
	    cluster_distributor_wgt ($quantity, @{$rf_batch_bins->[$i]}, @{$rf_bounds}, $rf_wgts->[$i]);
	}

	return 1;

}				# end of the subroutine batch_distributor_wgts #


# pr_cluster_distri (\@\@$) : a module which prints out a distribution of cluster sizes #
#
sub pr_cluster_distri (\@\@$) {

	my ($rf_bins, $rf_boundaries, $measured_entity) = @_;
	my $delim_width = 50;
	my $num_boundaries = @{$rf_boundaries};
	my $total = 0;
	(defined $measured_entity) or ( $measured_entity = "cluster size");
	#my $measured_entity = "sequence length";

	for (my $j = 0; $j < @{$rf_bins}; $j++) {
		$total += ${$rf_bins}[$j];
	}

	print '-' x $delim_width . "\n";
	print "\($measured_entity\) < ${$rf_boundaries}[0]:\t\t${$rf_bins}[0]\n";
	
	for (my $i = 1; $i < $num_boundaries; $i++) {

		print "${$rf_boundaries}[$i -1] =< \($measured_entity\) < ${$rf_boundaries}[$i]:\t${$rf_bins}[$i]\n";
	}

	print "${$rf_boundaries}[$num_boundaries -1] =< \($measured_entity\):\t\t${$rf_bins}[$num_boundaries]\n";

	print '-' x $delim_width . "\n";
	print "Total\t\t\t\t$total\n";

}						# end of the module pr_cluster_distri #


# batch_pr_distri (\@\@$) : a module which prints out MULTIPLE  distributions (of cluster sizes) #
#
sub batch_pr_distri (\@\@$) {

	my ($rf_array_bins, $rf_boundaries, $measured_entity) = @_;
	my $num_boundaries = @{$rf_boundaries};
	my $border = '-' x (35 + 10 * scalar(@{$rf_array_bins}) ) . "\n";
	my @total = ();
	my @outputs = ();
	for (my $i = 0; $i < @{$rf_array_bins}; $i++) { push @total, 0; }
	for (my $i = 0; $i < @{$rf_array_bins->[0]}; $i++) { push @outputs, ''; }

	(defined $measured_entity) or ( $measured_entity = "cluster size");
	#my $measured_entity = "sequence length";

	for (my $i = 0; $i < @{$rf_array_bins}; $i++) {
		for (my $j = 0; $j < @{$rf_array_bins->[$i]}; $j++) {
			$total[$i] += $rf_array_bins->[$i][$j];
			$outputs[$j] .= "\t" . $rf_array_bins->[$i][$j];
		}
	}

	print $border; 
	print "\($measured_entity\) < $rf_boundaries->[0]:\t\t$outputs[0]\n";
	
	for (my $i = 1; $i < $num_boundaries; $i++) {

		print "$rf_boundaries->[$i -1] =< \($measured_entity\) < $rf_boundaries->[$i]:\t$outputs[$i]\n";
	}

	print "$rf_boundaries->[$num_boundaries -1] =< \($measured_entity\):\t$outputs[$num_boundaries]\n";

	my $total = join "\t", @total;
	print $border;
	print "Total\t\t\t\t$total\n";

}						# end of the module batch_pr_distri #

# pr_file_cluster_distri (\@\@$\*) : a module which prints out a distribution of cluster sizes
#					to a file handle specified by the 3rd argument #
#
sub pr_file_cluster_distri (\@\@$\*) {

	my ($rf_bins, $rf_boundaries, $measured_entity, $fh) = @_;
	my $num_boundaries = @{$rf_boundaries};
	my $total = 0;
	(defined $measured_entity) or ( $measured_entity = "cluster size");
	#my $measured_entity = "sequence length";

	for (my $j = 0; $j < @{$rf_bins}; $j++) {
		$total += ${$rf_bins}[$j];
	}

	print $fh "\($measured_entity\) < ${$rf_boundaries}[0]:\t\t${$rf_bins}[0]\n";
	
	for (my $i = 1; $i < $num_boundaries; $i++) {

		print $fh "${$rf_boundaries}[$i -1] =< \($measured_entity\) < ${$rf_boundaries}[$i]:\t${$rf_bins}[$i]\n";
	}

	print $fh "${$rf_boundaries}[$num_boundaries -1] =< \($measured_entity\):\t\t${$rf_bins}[$num_boundaries]\n";

	print $fh "------------------------------------------------\nTotal\t\t\t\t$total\n";

}						# end of the module pr_file_cluster_distri #


# batch_pr_file_distri (\@\@$\*$\@) : a module which prints out MULTIPLE  distributions (of cluster sizes) 
#				to a filehandle specified by the 3rd argument.
# a minor revision: The subroutine now receives a 4th argument 'format',
#			which specifies the format of boundary parameters
# a second minor revision: The subroutine now receives a 5th argument 'tags',
#			   which specifies the labels of columns in the table.
#
sub batch_pr_file_distri (\@\@$\*$\@) {

	my ($rf_array_bins, $rf_boundaries, $measured_entity, $fh, $format, $tags) = @_;
	my $num_boundaries = @{$rf_boundaries};
	my $mytab = ' ' x 2;
	my $fwidth  =  7;
	my ($format_len) = ($format =~ m/^\%-?(\d+)/ );
	(defined $format_len) or ($format_len = 4);
	my $len_title = 2*$format_len + 10 + length($measured_entity);
	my $border = '-' x ($len_title + length($mytab) + 9 * scalar(@{$rf_array_bins}) ) . "\n";
	my @total = ();
	my @outputs = ();
	for (my $i = 0; $i < @{$rf_array_bins}; $i++) { push @total, 0; }
	for (my $i = 0; $i < @{$rf_array_bins->[0]}; $i++) { push @outputs, ''; }

	my $label = '';

	if (@{$tags}) {

		$label = $tags->[0]; 
		$label .= ' ' x ($len_title -1 - length($label) ) . ":$mytab";
		my $mfwidth = $fwidth + length($mytab);
		for (my $i=1; $i<@{$tags}; $i++) {

			if (length($tags->[$i]) < $mfwidth) {

				$label .= ' ' x ($mfwidth -length($tags->[$i]) ) . $tags->[$i];
			} else {
				$label .= ' ' . $tags->[$i];
			}
		}

		$label .= "\n";
	}

	(defined $measured_entity) or ( $measured_entity = "cluster size");

	for (my $i = 0; $i < @{$rf_array_bins}; $i++) {
		for (my $j = 0; $j < @{$rf_array_bins->[$i]}; $j++) {
			$total[$i] += $rf_array_bins->[$i][$j];
			#$outputs[$j] .= "\t" . $rf_array_bins->[$i][$j];
			$outputs[$j] .= sprintf "$mytab%${fwidth}d" , $rf_array_bins->[$i][$j];
		}
	}

	print $fh $label;
	print $fh $border; 
	#print $fh "\($measured_entity\) < $rf_boundaries->[0]:\t\t$outputs[0]\n";
	print $fh ' ' x ($format_len + 4);
	printf $fh "\($measured_entity\) < $format:$mytab$outputs[0]\n", $rf_boundaries->[0];
	
	for (my $i = 1; $i < $num_boundaries; $i++) {

		#print $fh "$rf_boundaries->[$i -1] =< \($measured_entity\) < $rf_boundaries->[$i]:\t$outputs[$i]\n";
		printf $fh "$format =< \($measured_entity\) < $format:$mytab$outputs[$i]\n", $rf_boundaries->[$i-1], $rf_boundaries->[$i];
	}

	#print $fh "$rf_boundaries->[$num_boundaries -1] =< \($measured_entity\):\t$outputs[$num_boundaries]\n";
	printf $fh "$format =< \($measured_entity\)", $rf_boundaries->[$num_boundaries -1];
	print $fh ' ' x ($format_len + 3) . ":$mytab$outputs[$num_boundaries]\n";

	#my $total = join "\t", @total;
	my $total = '';
	foreach my $tmp (@total) { $total .= sprintf "$mytab%${fwidth}d" , $tmp; }
	print $fh $border;
	print $fh "Total" . ' ' x ($len_title - 1 - length('Total') );
	print $fh ":$mytab$total\n";

}						# end of the module batch_pr_file_distri #


# batch_pr_file_distri_xl (\@\@$\*$\@) : This subroutine is a revised version of
#		"batch_pr_file_distri" specialized  to print a table for Excel.
#
# batch_pr_file_distri:	 a module which prints out MULTIPLE  distributions (of cluster sizes) 
#				to a filehandle specified by the 3rd argument.
# a minor revision: The subroutine now receives a 4th argument 'format',
#			which specifies the format of boundary parameters
# a second minor revision: The subroutine now receives a 5th argument 'tags',
#			   which specifies the labels of columns in the table.
#
sub batch_pr_file_distri_xl (\@\@$\*$\@) {

	my ($rf_array_bins, $rf_boundaries, $measured_entity, $fh, $format, $tags) = @_;
	my $num_boundaries = @{$rf_boundaries};
	my $mytab = "\t";
	my $fwidth  =  1;
	my ($format_len) = ($format =~ m/^\%-?(\d+)/ );
	my $fl_int = ($format =~ m/d\s*$/ );
	(defined $format_len) or ($format_len = 4);
	my $len_title = 2*$format_len + 10 + length($measured_entity);
	my $border = "\n";
	my @total = ();
	my @outputs = ();
	for (my $i = 0; $i < @{$rf_array_bins}; $i++) { push @total, 0; }
	for (my $i = 0; $i < @{$rf_array_bins->[0]}; $i++) { push @outputs, ''; }

	my $label = '';

	if (@{$tags}) {

		$label = $tags->[0]; 
		$label .= ' ' x ($len_title -1 - length($label) ) . "$mytab";
		my $mfwidth = $fwidth + length($mytab);
		for (my $i=1; $i<@{$tags}; $i++) {

				$label .= "\t" . $tags->[$i];
		}

		$label .= "\n";
	}

	(defined $measured_entity) or ( $measured_entity = "cluster size");

	for (my $i = 0; $i < @{$rf_array_bins}; $i++) {
		for (my $j = 0; $j < @{$rf_array_bins->[$i]}; $j++) {
			$total[$i] += $rf_array_bins->[$i][$j];
			#$outputs[$j] .= "\t" . $rf_array_bins->[$i][$j];
			$outputs[$j] .= sprintf "$mytab%${fwidth}d" , $rf_array_bins->[$i][$j];
		}
	}

	print $fh $label;
	print $fh $border; 
	print $fh ' ' x ($format_len + 4);
	printf $fh "\($measured_entity\) < $format$mytab$outputs[0]\n", $rf_boundaries->[0];
	
	for (my $i = 1; $i < $num_boundaries; $i++) {

	    if ($fl_int and ($rf_boundaries->[$i-1] + 1 == $rf_boundaries->[$i]) ) {

		print $fh ' ' x ($format_len + 4); ## ADDED on June 8 (Mon), 2009 ##

		printf $fh "\($measured_entity\) = $format$mytab$outputs[$i]\n", $rf_boundaries->[$i-1];
	    } else {
		printf $fh "$format =< \($measured_entity\) < $format$mytab$outputs[$i]\n", $rf_boundaries->[$i-1], $rf_boundaries->[$i];
	    }
	}

	printf $fh "$format =< \($measured_entity\)", $rf_boundaries->[$num_boundaries -1];
	print $fh ' ' x ($format_len + 3) . "$mytab$outputs[$num_boundaries]\n";

	my $total = '';
	foreach my $tmp (@total) { $total .= sprintf "$mytab%${fwidth}d" , $tmp; }
	print $fh $border;
	print $fh "Total" . ' ' x ($len_title - 1 - length('Total') );
	print $fh "$mytab$total\n";

}			# end of the module batch_pr_file_distri_xl #


# batch_pr_file_distri_xl_fmt (\@\@$\*$$$\@) : This subroutine is a revised version of
#		"batch_pr_file_distri_xl" specialized  to print a table for Excel,
#		with each field formatted in a specified way.
#
# A minor revision:  Added a flag $fl_total,
#		which specifies whether the 'total' is already available
#		(=1) or must be constructed from scratch (=0).
#
# batch_pr_file_distri:	 a module which prints out MULTIPLE  distributions (of cluster sizes) 
#				to a filehandle specified by the 3rd argument.
# a minor revision: The subroutine now receives a 4th argument 'format_bd',
#			which specifies the format of boundary parameters
# a second minor revision: The subroutine now receives a 5th argument 'tags',
#			   which specifies the labels of columns in the table.
#
sub batch_pr_file_distri_xl_fmt (\@\@$\*$$$\@) {

	my ($rf_array_bins, $rf_boundaries, $measured_entity, $fh, $format_bd, $format_fld, $fl_total, $tags) = @_;
	my $num_boundaries = @{$rf_boundaries};
	my $mytab = "\t";
	my $fwidth  =  1;
	my ($format_len) = ($format_bd =~ m/^\%-?(\d+)/ );
	my $fl_int = ($format_bd =~ m/d\s*$/ );
	(defined $format_len) or ($format_len = 4);
	my $len_title = 2*$format_len + 10 + length($measured_entity);
	my $border = "\n";
	my @total = ();
	my @outputs = ();

		# ADDED on June 6 (Mon), 2011 #
	my $ct_clms = @{$rf_array_bins};
	my $ct_rows = @{$rf_array_bins->[0]};
	for (my $i = 0; $i < $ct_clms; $i++) { push @total, $fl_total ? $rf_array_bins->[$i][$ct_rows -1] : 0; }
	$fl_total and ($ct_rows--);
	for (my $i = 0; $i < $ct_rows; $i++) { push @outputs, ''; }
	##for (my $i = 0; $i < @$rf_array_bins}; $i++) { push @total, 0; }
	##for (my $i = 0; $i < @{$rf_array_bins->[0]}; $i++) { push @outputs, ''; }

	my $label = '';

	if (@{$tags}) {

		$label = $tags->[0]; 
		$label .= ' ' x ($len_title -1 - length($label) ) . "$mytab";
		my $mfwidth = $fwidth + length($mytab);
		for (my $i=1; $i<@{$tags}; $i++) {

				$label .= "\t" . $tags->[$i];
		}

		$label .= "\n";
	}

	(defined $measured_entity) or ( $measured_entity = "cluster size");

	for (my $i = 0; $i < $ct_clms; $i++) {
		for (my $j = 0; $j < $ct_rows; $j++) {
	##for (my $i = 0; $i < @{$rf_array_bins}; $i++) {
		##for (my $j = 0; $j < @{$rf_array_bins->[$i]}; $j++) {
			my $fld = $rf_array_bins->[$i][$j];
			(!($fl_total) and (uc($fld) !~ m/NA/)) and ($total[$i] += $fld);
			##$total[$i] += $rf_array_bins->[$i][$j];

			#$outputs[$j] .= "\t" . $rf_array_bins->[$i][$j];
			##$outputs[$j] .= sprintf "$mytab%${fwidth}d" , $rf_array_bins->[$i][$j];
			$outputs[$j] .= sprintf "$mytab$format_fld" , $fld;
		}
	}

	print $fh $label;
	print $fh $border; 
	print $fh ' ' x ($format_len + 4);
	printf $fh "\($measured_entity\) < $format_bd$mytab$outputs[0]\n", $rf_boundaries->[0];
	
	for (my $i = 1; $i < $num_boundaries; $i++) {

	    if ($fl_int and ($rf_boundaries->[$i-1] + 1 == $rf_boundaries->[$i]) ) {

		print $fh ' ' x ($format_len + 4); ## ADDED on June 8 (Mon), 2009 ##

		printf $fh "\($measured_entity\) = $format_bd$mytab$outputs[$i]\n", $rf_boundaries->[$i-1];
	    } else {
		printf $fh "$format_bd =< \($measured_entity\) < $format_bd$mytab$outputs[$i]\n", $rf_boundaries->[$i-1], $rf_boundaries->[$i];
	    }
	}

	printf $fh "$format_bd =< \($measured_entity\)", $rf_boundaries->[$num_boundaries -1];
	print $fh ' ' x ($format_len + 3) . "$mytab$outputs[$num_boundaries]\n";

	my $total = '';
	foreach my $tmp (@total) {
	    ##$total .= sprintf "$mytab%${fwidth}d" , $tmp;
	    $total .= sprintf "$mytab$format_fld" , $tmp;
	}
	print $fh $border;
	print $fh "Total" . ' ' x ($len_title - 1 - length('Total') );
	print $fh "$mytab$total\n";

}			# end of the module batch_pr_file_distri_xl_fmt #


# batch_pr_file_distri_fmt (\@\@$\*$\@) : a module which prints out MULTIPLE  distributions (of cluster sizes) 
#				to a filehandle specified by the 3rd argument.
#		 REVISION from the subroutine 'batch_pr_file_distri':
#			This new subroutine receives a 6th argument '$fmt_fields' 
#			So that it can handle decimals and fractions. 
#		(NOTE: $fmt_fields should include initial tab
#			 which separates the interval and the numerical field.)
#		ANOTHER REVISION is that 'tags' is now (a reference of) a string ,
#				and not (a reference of) an array.
#
# a minor revision: The subroutine now receives a 4th argument 'fmt_bd',
#			which specifies the format of boundary parameters
# a second minor revision: The subroutine now receives a 5th argument 'tags',
#			   which specifies the labels of columns in the table.
#
sub batch_pr_file_distri_fmt (\@\@$\*$\$$) {

	my ($rf_array_bins, $rf_boundaries, $measured_entity, $fh, $fmt_bd, $tags, $fmt_fields) = @_;
	my $num_boundaries = @{$rf_boundaries};
	my $mytab = ' ' x 2;
	my $fwidth  =  9;
	my ($fmt_bd_len) = ($fmt_bd =~ m/^\%-?(\d+)/ );
	(defined $fmt_bd_len) or ($fmt_bd_len = 4);
	my $len_title = 2*$fmt_bd_len + 10 + length($measured_entity);
	#my $border = '-' x ($len_title + length($mytab) + 9 * scalar(@{$rf_array_bins}) ) . "\n";
	my $border = '-' x ($len_title + length($mytab) + length (${$tags}) ) . "\n";

	my $ct_intervals = @{$rf_array_bins->[0]};
	my $ct_fields = @{$rf_array_bins};
	my @total = ();
	my @outputs = ();
	for (my $i = 0; $i < $ct_fields; $i++) { push @total, 0; }
	for (my $i = 0; $i < $ct_intervals; $i++) { push @outputs, ''; }

	my $slen_title = $len_title -1;
	my $label = sprintf ("%-${slen_title}s", 'Categories') . ":${$tags}\n";
=comment out
	my $label = '';

	if (@{$tags}) {

		$label = $tags->[0]; 
		$label .= ' ' x ($len_title -1 - length($label) ) . ":$mytab";
		my $mfwidth = $fwidth + length($mytab);
		for (my $i=1; $i<@{$tags}; $i++) {

			if (length($tags->[$i]) < $mfwidth) {

				$label .= ' ' x ($mfwidth -length($tags->[$i]) ) . $tags->[$i];
			} else {
				$label .= ' ' . $tags->[$i];
			}
		}

		$label .= "\n";
	}
=cut

	(defined $measured_entity) or ( $measured_entity = "cluster size");

	for (my $j = 0; $j < $ct_intervals; $j++) {

		my @reservoir  = ();
		for (my $i = 0; $i < $ct_fields; $i++) {
			$total[$i] += $rf_array_bins->[$i][$j];
			#$outputs[$j] .= sprintf "$mytab%${fwidth}d" , $rf_array_bins->[$i][$j];
 			push @reservoir , $rf_array_bins->[$i][$j];
		}

		$outputs[$j] = sprintf $fmt_fields, @reservoir; 
	}

	print $fh $label;
	print $fh $border; 
	#print $fh "\($measured_entity\) < $rf_boundaries->[0]:\t\t$outputs[0]\n";
	print $fh ' ' x ($fmt_bd_len + 4);
	printf $fh "\($measured_entity\) < $fmt_bd:$outputs[0]\n", $rf_boundaries->[0];
	
	for (my $i = 1; $i < $num_boundaries; $i++) {

		#print $fh "$rf_boundaries->[$i -1] =< \($measured_entity\) < $rf_boundaries->[$i]:\t$outputs[$i]\n";
		printf $fh "$fmt_bd =< \($measured_entity\) < $fmt_bd:$outputs[$i]\n", $rf_boundaries->[$i-1], $rf_boundaries->[$i];
	}

	#print $fh "$rf_boundaries->[$num_boundaries -1] =< \($measured_entity\):\t$outputs[$num_boundaries]\n";
	printf $fh "$fmt_bd =< \($measured_entity\)", $rf_boundaries->[$num_boundaries -1];
	print $fh ' ' x ($fmt_bd_len + 3) . ":$outputs[$num_boundaries]\n";

	#my $total = join "\t", @total;
	my $total = sprintf $fmt_fields, @total;
	print $fh $border;
	print $fh "Total" . ' ' x ($len_title - 1 - length('Total') );
	print $fh ":$total\n";

}						# end of the module batch_pr_file_distri_fmt #


# array_distributor: This subroutine distributes an array of quantities into a set of 'bin's
#		    via a serial use of 'cluster_distributor'.
#
sub array_distributor (\@\@\@) {

	my ($rf_quantities, $rf_bins, $rf_bounds) = @_;

	foreach my $quantity (@{$rf_quantities}) {

		cluster_distributor ($quantity, @{$rf_bins}, @{$rf_bounds});
	}

}					# end of the subroutine array_distributor #

# window_distributor: This subroutine construct a set of histogram-like distributions
#			out of a set of (probably potitional) distributions obtained
#			e.g. by a sliding-window analysis.
#
# Expected input format: $_[0] = [ $window_distribution_1, $window_distribution_2, ... ],
#			where $window_distribution_i = [quantity_wd1, quantity_wd2, ... ]_i-th_set,
#			their sizes do not have to be identical.
#
sub window_distributor (\@\@) {

	my ($rf_window_distri, $rf_bounds) = @_;

	my $ct_wd_sets = @{$rf_window_distri};
	$ct_wd_sets or (return []);

	my @batch_windowsets = ();	# statistics collected for each of the windowsets #
	my @bin_tot_windows = ();	# total statistics #

	foreach (1 .. $ct_wd_sets) { push @batch_windowsets, []; }
	push @batch_windowsets, \@bin_tot_windows;
	batch_initializer (@batch_windowsets, @{$rf_bounds});

	for (my $i=0; $i<$ct_wd_sets; $i++) {

		my $wd_ct = @{$rf_window_distri->[$i]};
		for (my $l=0; $l<$wd_ct; $l++) {

			my $wd_quantity = $rf_window_distri->[$i][$l];

			cluster_distributor ($wd_quantity, @{$batch_windowsets[$i]}, @{$rf_bounds});
			cluster_distributor ($wd_quantity, @bin_tot_windows, @{$rf_bounds});

		}		# end of the inner for #
	}			# end of the outer for #

		return (\@batch_windowsets);

}				# end of the subroutine window_distributor #

# init_2dim_distri: This subroutine receives 2 sets of boundaries,
#			one for 'abscissas' and the other for 'ordinates',
#			and returns initialized sets of bins and tags.
#
# RULE: Abscissas are used for the tags of bin sets,
#	ordinates are used as usual boundaries,
#	both of which can be used e.g., in 'batch_pr_file_distri'.
#
sub init_2dim_distri (\@\@$) {

	my ($rf_abscissa, $rf_ordinate, $fmt_abs) = @_;

	my $ct_abs = @{$rf_abscissa};
	$ct_abs or (return (undef, undef, undef) );

	my @tags = ('Y \ X', sprintf ( "[\,$fmt_abs)", $rf_abscissa->[0] ) );
	for (my $i=1; $i<$ct_abs; $i++) {
		push @tags, sprintf ( "[$fmt_abs\,$fmt_abs)", @{$rf_abscissa}[$i-1, $i] ) ;
	}

	push @tags, sprintf ( "[$fmt_abs\,)", $rf_abscissa->[$ct_abs-1] ) , 'Total';

	my $ct_bins = @tags - 1;

	 my @mags = my @batch_bins = ();

	for (1 .. $ct_bins) { push @batch_bins, []; }

	batch_initializer (@batch_bins, @{$rf_ordinate});

	return (\@batch_bins, \@tags);

}			# end of the subroutine init_2dim_distri #

# distributor_2dim: This subroutine  distributes a pair of quantities
#				into a 2dimensinal histogram.
#
sub distributor_2dim ($$\@\@\@) {

	my ($x, $y, $rf_bins, $rf_abscissa, $rf_ordinate) = @_;

	my $ct_abs = @{$rf_abscissa};
	unless (@{$rf_bins} == $ct_abs +2) {
		print STDERR "Illegal input for \'distributor_2dimensional\'.\n";
		return 0;
	}

	for (my $i=0; $i<$ct_abs; $i++) {

		if ($x < $rf_abscissa->[$i]) {
			cluster_distributor($y, @{$rf_bins->[$i]}, @{$rf_ordinate});
			cluster_distributor ($y, @{$rf_bins->[$ct_abs + 1]}, @{$rf_ordinate} ) ;
			return 1;
		}
	}					# end of the for #

	unless ($x >= $rf_abscissa->[$ct_abs-1] ) {
		print STDERR "Some unsusual thing has happened for \'distributor_2dimensional\'.\n";
		return 0;
	}

	cluster_distributor ($y, @{$rf_bins->[$ct_abs]}, @{$rf_ordinate});
	cluster_distributor ($y, @{$rf_bins->[$ct_abs + 1]}, @{$rf_ordinate});

	return 1;
}			# end of the subroutine distributor_2dim #


# pr_file_2dim_distri: This subroutine prints out a 2dimensional distribution
#			by using the subroutine 'batch_pr_file_distri'.
#
sub pr_file_2dim_distri (\@\@\@$$\*$$) {

	my ($rf_batch_bins, $rf_abscissa, $rf_ordinate, $measure_abs, $measure_ord, 
		$fh,  $fmt_abs, $fmt_ord) =  @_;

	my $ct_abs = @{$rf_abscissa};
	$ct_abs or (return (undef, undef, undef) );

	my @tags = ("Y \\ X ($measure_abs)", sprintf ( "[\,$fmt_abs)" , $rf_abscissa->[0] ) );
	for (my $i=1; $i<$ct_abs; $i++) {
		push @tags, sprintf ( "[$fmt_abs\,$fmt_abs)", @{$rf_abscissa}[$i-1, $i] ) ;
	}

	push @tags, sprintf ("[$fmt_abs\,)", $rf_abscissa->[$ct_abs-1] ), 'Total';

	batch_pr_file_distri (@{$rf_batch_bins},@{$rf_ordinate},  $measure_ord, *${$fh}, $fmt_ord, @tags);

}		# end of the subroutine pr_file_2dim_distri #


# pr_file_2dim_distri_xl: This subroutine prints out a 2dimensional distribution
#			by using the subroutine 'batch_pr_file_distri'.
#
sub pr_file_2dim_distri_xl (\@\@\@$$\*$$) {

	my ($rf_batch_bins, $rf_abscissa, $rf_ordinate, $measure_abs, $measure_ord, 
		$fh,  $fmt_abs, $fmt_ord) =  @_;

	my $ct_abs = @{$rf_abscissa};
	$ct_abs or (return (undef, undef, undef) );

	my @tags = ("Y \\ X ($measure_abs)", sprintf ( "[\,$fmt_abs)" , $rf_abscissa->[0] ) );
	for (my $i=1; $i<$ct_abs; $i++) {
		push @tags, sprintf ( "[$fmt_abs\,$fmt_abs)", @{$rf_abscissa}[$i-1, $i] ) ;
	}

	push @tags, sprintf ("[$fmt_abs\,)", $rf_abscissa->[$ct_abs-1] ), 'Total';

	batch_pr_file_distri_xl (@{$rf_batch_bins},@{$rf_ordinate},  $measure_ord, *${$fh}, $fmt_ord, @tags);

}		# end of the subroutine pr_file_2dim_distri_xl #


# create_key_intervals:
#	This subroutine receives a set of boundaries and
#	create the key of interval categories using the set.
#
# INPUT: (\@boundaries, $format),
#
# OUTPUT: \@key, where
#	@key = (" x < $boundaries[0]", "$boundaries[0] <= x < $bounaries[1]", 
#		..., "$boundaries[$#boundaries] <= x ").
#
sub create_key_intervals (\@$) {

	my ($bound, $format) = @_;

	my $ct_bd = @{$bound};

	my @key_intervals = ( sprintf(" x < $format", $bound->[0]) );
	for (my $i=1; $i<$ct_bd; $i++) {
	    push @key_intervals,
		sprintf ("$format <= x < $format", @{$bound}[$i-1, $i] ) ;
	}

	push @key_intervals, sprintf ("$format <= x ", $bound->[$ct_bd-1] );

	return \@key_intervals;

}		# end of the subroutine 'create_key_intervals' #

# classify_by_measure:
#	This subroutine receives a numerical measure accompanied by
#	a set of boundaries, and classify the measure according to the boundaries.
#
# INPUT: ($measure, \@boundaries), where
#	@boundaries stores the boundaries for the measure.
#
# OUTPUT: $class, which is defined by
#	$k+1 if ($boundaries[$k-1] (if any) <= $measure < $boundaries[$k]),
#	scalar (@boundaries) + 1 if ($measure >= $boundaries[$#boundaries]).
#
sub classify_by_measure ($\@) {

	my ($measure, $bounds) = @_;

	my $ct_bds = @{$bounds};

	for (my $k=0; $k<$ct_bds; $k++) {
	    ($measure < $bounds->[$k]) and (return $k+1);
	}

	return $ct_bds + 1;

}	# End of the subroutine 'classify_by_measure' #


1;			# a "spell" necessary for the package to be "use"-able or "require"-able #


