package MyInferAncStates_parser;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.5: Copyright (C) 2014 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyInferAncStates_parser.pm:
#	This package provides modules to parse the outputs of programs inferring the ancestral states
#	of the sequences, such as those in PAML and the Perl script, 'br_infer_psm_indel_events.pl'.
#	(CREATION STARTED on Nov 23 (Sat), 2013, ENDED on Nov 26 (Tue), 2013)
#
#	MODIFIED on Dec 31 (Tue), 2013: 
#	A SERIOUS BUG, which concerns the complementary bases, was fixed in the subroutine 'revcom_anc_base_state_distr'.
#


use strict;
use warnings;

use MyArrayManipu qw( diff_array union_array );
use MyFile_manipu qw( open_input_file2 );
use MyTaxonomy qw( fetch_ancestry fetch_last_common_ancestor_ov2 fetch_common_ancestors
		 fetch_ext_offsprings fetch_all_offsprings );

use MyTreeManipu_II qw(
	extract_sgltree_from_file
	input_tree_file read_tree_file input_subtree
	input_tree_with_nodeids
	make_node2children
	);
	## copy_tree
	## output_tree draw_subtree
	## br_output_tree br_draw_subtree topo_subtree
	## br_output_tree_sep br_draw_subtree_sep
	## make_node2parent

use MyTreeManipu_II_spe qw(
	replace_nodeids
	input_ext_tree_file
	);
	## read_ext_tree_file input_ext_subtree

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	map_treenode2treenode
	reconstruct_rst_tree
	extract_sitewise_anc_state_distri_baseml
	from_1based_to_0based_paml_hash
	revcom_anc_base_state_distr
	parse_brief_output_infer_psm_indel_events
	revcom_distr_psm_indel_events
	linear_order_indel_state_distributions
	);

sub reconstruct_rst_tree ($$$);


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

=for the package 'MyTreeManipu_II.pm'
our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.
=cut

	# Levels of test outputs:

our $TEST_LEVEL_SBST = 0;
our $TEST_LEVEL_INDEL = 3;


### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.



# map_treenode2treenode:
#	This subroutine receives two independent representations of the same tree 
#	(both based on the same Newick format with the same top node,
#	and in the same format, but with likely different node ID assignments),
#	then construct a map from the node ID in one representation to thos in the other.
#
#	PREREQUISITE: The exterior nodes are assigned the same names in both representations.
#
# INPUT: (\%node2ch_src, \%node2pa_src, \%name2id_src, \%node2pa_dst, \%name2id_dst), where
#	%node2ch_src = ($node_id => \@children,...) for the interior nodes, in the 'source' representation ;
#	%node2pa_src/dst = ($node_id => $parent_id, ...) for the nodes other than the top node, in the 'source'/'destination' representation;
#	%name2id_src/dst = ($name => $node_id, ...) for the nodes, including all exterior nodes, in the 'source'/'destination' representation;
#
# OUTPUT: \%nodeid_src2nodeid_dst, where
#	%nodeid_src2nodeid_dst = ($node_id_src => $node_id_dst, ...) for all the nodes,
#	maps the node IDs in the 'source' representation to the node IDs in the 'destination' representation.
#
sub map_treenode2treenode (\%\%\%\%\%) {

	my ($node2ch_src, $node2pa_src, $name2id_src, $node2pa_dst, $name2id_dst) = @_;

	my %nodeid_src2nodeid_dst;

	foreach my $name (keys %{$name2id_src}) { # Main outer foreach-loop (over exterior nodes). #

	    my $id_src = $name2id_src->{$name};
	    (defined $id_src) or next; # Skip if the node ID is undefined.

	    my $children = $node2ch_src->{$id_src};
	    ((defined $children) and (@{$children}>0)) and next; # Skip if the node is an interior one.

	    my $id_dst = $name2id_dst->{$name};
	    unless (defined $id_dst) {
		print STDOUT "[ERROR 'map_treenode2treenode'] The node with the name '$name' (source node ID = '$id_src') has NO ID in the destination tree representation: $! !\n";
		return undef;
	    }

	    $nodeid_src2nodeid_dst{$id_src} = $id_dst;

	    while (defined $node2pa_src->{$id_src}) { # Main inner while-loop (climbing the ancestral lineage). #

		$id_src = $node2pa_src->{$id_src};
		(defined $nodeid_src2nodeid_dst{$id_src}) and last; # Stop climbing the lineage if the map is already given.

		$id_dst = $node2pa_dst->{$id_dst};
		unless (defined $id_dst) {
		    print STDOUT "[ERROR 'map_treenode2treenode'] The source node ID '$id_src' has NO COUNTERPART in the destination tree representation: $! !\n";
		    return undef;
		}

		$nodeid_src2nodeid_dst{$id_src} = $id_dst;

	    }				# End of the main inner while-loop  (climbing the ancestral lineage). #

	}				# End of the main outer foreach-loop (over exterior nodes). #

	return \%nodeid_src2nodeid_dst;

}	# End of the subroutine 'map_treenode2treenode' #


# reconstruct_rst_tree:
#	This subroutine reconstructs the tree recorded in the 'rst' file
#	out of the three output lines.
#
# INPUT: ($tree_seqname, $tree_exnodeid, $parent_children), where
#	$tree_seqname is the 1st tree (in the Newick format) written in terms of the sequence names;
#	$tree_exnodeid is the 2nd tree (also in the Newick format) written in terms of the exterior node IDs;
#	$parent_children is the 3rd tree, given in the parent..children format.
#
# OUTPUT: \@tree_rst, where
#	@tree_rst = ($topnode_id, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr),
#	where the hashes follow the standard format (described around the top),
#	with node IDs being those assigned in the 'rst' file;
#	
sub reconstruct_rst_tree ($$$) {

	my ($tree_seqname, $tree_exnodeid, $parents_children) = @_;

		# $in_nodeid is the node ID internally-assigned by the tree-reader subroutine.
		# $ex_nodeid is the exterior node ID displayed in the 'rst' file.
		# $rst_nodeid is the node ID displayed in the 'rst' file, (which includes $ex_nodeid).

	my %in_nodeid2pa; my %in_nodeid2ch; my %name2in_nodeid; my %in_nodeid2name; my %in_brid2attr;

	my $cp_tree_seqname = $tree_seqname;
	unless ( input_tree_file ($tree_seqname, %in_nodeid2pa, %in_nodeid2ch, %name2in_nodeid, %in_nodeid2name, %in_brid2attr) ) {
	    $cp_tree_seqname =~ s/\s+$//;
	    print STDOUT "[ERROR 'reconstruct_rst_tree'] Subroutine 'input_tree_file' FAILED on the first input tree in the 'rst' file: $! !\n";
	    print STDOUT  "'$cp_tree_seqname' .\n\n";
	    return undef;
	}
	$cp_tree_seqname = '';

	my %in_nodeid2pa2; my %in_nodeid2ch2; my %ex_nodeid2in_nodeid; my %in_nodeid2ex_nodeid; my %dummy;

	my $cp_tree_exnodeid = $tree_exnodeid;
	unless (  input_tree_file ($tree_exnodeid, %in_nodeid2pa2, %in_nodeid2ch2, %ex_nodeid2in_nodeid, %in_nodeid2ex_nodeid, %dummy) ) {
	    $cp_tree_exnodeid =~ s/\s+$//;
	    print STDOUT "[ERROR 'reconstruct_rst_tree'] Subroutine 'input_tree_file' FAILED on the second input tree in the 'rst' file: $! !\n";
	    print STDOUT "'$cp_tree_exnodeid' .\n\n";
	    return undef;
	}
	$cp_tree_exnodeid = '';

		# Construct the tree from the 'parent..child' format. #
	my %rst_node2pa;
	my $rst_topnode_id;
	foreach my $rel (split ' ', $parents_children) { # 1st foreach-loop (over parent-child). #

	    my ($pa, $ch) = split /\.\./, $rel;
	    unless ((defined $ch) and ($ch =~ m/\S/)) {
		$parents_children =~ s/\s+$//;
		print STDOUT "[ERROR 'reconstruct_rst_tree'] ILLEGAL INPUT: '$rel' does NOT CONFORM TO the format, 'parent..child': $! !\n";
		print STDOUT  "'$parents_children' .\n\n";
	    }
	    $rst_node2pa{$ch} = $pa;

	    (defined $rst_topnode_id) or ($rst_topnode_id = $pa);
	}				# End of the 1st foreach-loop (over parent-child). #

	my $rst_node2ch = make_node2children (%rst_node2pa);
	unless (defined $rst_node2ch) {
	    print STDOUT "[ERROR 'reconstruct_rst_tree'] Subroutine 'make_node2children' FAILED: $! !\n";
	    return undef;
	}

		# Look for the true top-node ID in the rst-displayed tree. #

	while (defined $rst_node2pa{$rst_topnode_id}) { $rst_topnode_id = $rst_node2pa{$rst_topnode_id}; }

		# Recover the name <->  node ID relationships, #
		# as well as the branch ID -> attribute relatiopships. #

	my %name2rst_nodeid = ($TOP_NODE => $rst_topnode_id);
	my %rst_nodeid2name;
	my %rst_brid2attr;

	foreach my $in_nodeid (keys %in_nodeid2name) { # 3rd foreach-loop (over exterior nodes) #
	##foreach my $in_nodeid (%in_nodeid2name) { # 3rd foreach-loop (over exterior nodes) # A BUG fixed on Nov 26 (Tue), 2013. #

	    my $name = $in_nodeid2name{$in_nodeid};
	    my $rst_nodeid = my $ex_nodeid = $in_nodeid2ex_nodeid{$in_nodeid};

	    $rst_nodeid2name{$rst_nodeid} = $name;
	    $name2rst_nodeid{$name} = $rst_nodeid;

	    my $in_brid = $in_nodeid;
	    my $rst_brid = $ex_nodeid;
	    while (defined $rst_node2pa{$rst_brid}) { # Inner while-loop (climbing the lineage) #

		(defined $rst_brid2attr{$rst_brid}) and last;

		$rst_brid2attr{$rst_brid} = $in_brid2attr{$in_brid};

		$in_brid = $in_nodeid2pa{$in_brid};
		$rst_brid = $rst_node2pa{$rst_brid};

	    }				# End of the inner while-loop (climbing the lineage) # 

	}			 # End of the 3rd foreach-loop 

	my @tree_rst = ($rst_topnode_id, \%rst_node2pa, $rst_node2ch, \%name2rst_nodeid, \%rst_nodeid2name, \%rst_brid2attr);

		# @tree_rst = ($topnode_id, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr),

	return \@tree_rst;

}	#  End of the subroutine 'reconstruct_rst_tree' #



# extract_sitewise_anc_state_distri_baseml:
#	This subroutine reads in an 'rst' file of the program 'baseml' in the PAML package
#	under the 'verbose = 2' option,
#	and extracts the state distributions at all sites at the ancestral nodes.
#
# INPUT: (\*{$fh_rst}, \@bases, $fl_full_anc), where
#	$fh_rst is the reference to the file-handle for the relevant 'rst' file;
#	@bases specifies the probabilities of the bases to be stored in the output.
#	(More specifically, The probability of the base $base[$b] will be stored
#	as the $b th element of the array for the output distribution);
#	$fl_full_anc = 1 if you want distributions for ALL ancestral sequences,
#		     = 0 if you want distributions only for the TOP (i.e., root) sequence.
#
# OUTPUT: ($tree_no, \@tree_rst, \%node2state_distr, \%node2freq_patterns, \%node2patterns), where
#	$tree_no is the tree number assigned by the 'rst' file;
#	@tree_rst = ($topnode_id, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr),
#	where the hashes follow the standard format (described around the top),
#	with node IDs being those assigned in the 'rst' file;
#
#	%node2state_distri = ($node_id => \@state_distri, ...) for interior nodes, with
#	$node_id being consistent with @tree_rst, and 
#	$state_distri[$s]->[$b] being the probability of the $b th base (i.e. $bases[$b]) at the $s th site;
#	
#	%node2freq_patterns = ($node_id => \@freq_patterns, ...) for interior nodes, with
#	$freq_patterns[$s] being the frequency of the character pattern at the $s th site;
#
#	%node2patterns = ($node_id => \@patterns, ...) for interior nodes, with
#	$patterns[$s] being the character pattern at the $s th site.
#
#	NOTE1: @{$state_distri[0]} is empty, and $freq_patterns[0] and $patterns[0] are undefined,
#	because PAML employs the 1-based full-closed convention of the site coordinate.
#
#	NOTE2: $tree_no = undef indicates that there are no more records to be extracted;
#	A defined $tree_no and undefined \@tree_rst (and \%node2state_distri) indicate a failure of the subroutine.
#
sub extract_sitewise_anc_state_distri_baseml (\*\@$) {

	my ($fh_rst, $bases, $fl_full_anc) = @_;
	local $/ = ''; # Use blank lines as input-record-separators.

		## Extract the tree no. ##

	my $tree_no;

	while (my $in = <$fh_rst>) { # The 1st preliminary while-loop. #
	    ($tree_no) = ($in =~ m/^TREE\s+\#\s+(\S+)/);
	     (defined $tree_no) and last;
	}		# End of the 1st preliminary while-loop. #

	unless (defined $tree_no) {
	     return (undef, undef, undef, undef, undef); # There should be no more tree the ancestral inference is based on.
	}

		## Extract the tree right as displayed in the 'rst' file. ##

	while (my $in = <$fh_rst>) { ($in =~ m/^Ancestral\s+reconstruction/) and last; }

	my $tree_seqname = <$fh_rst>;
	my $tree_extnodeid = <$fh_rst>;
	my $parents_children = <$fh_rst>;

	unless (defined $parents_children) {
	    print STDOUT "[ERROR 'extract_sitewise_anc_state_distri_baseml'] The input 'rst' file does NOT contain adequate information on the tree: $! !\n";
	    return ($tree_no, undef, undef, undef, undef);
	}

	# @{$tree_rst} = ($topnode_id, \%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr),
	#
	my $tree_rst = reconstruct_rst_tree ($tree_seqname, $tree_extnodeid, $parents_children);
	unless (defined $tree_rst) {
	    print STDOUT "[ERROR 'extract_sitewise_anc_state_distri_baseml'] The subroutine 'reconstruct_rst_tree' FAILED: $! !\n";
	    return ($tree_no, undef, undef, undef, undef);
	}

	my $topnode_id = $tree_rst->[0];

		## (Final preliminary process) Prepare the base -> index correspondence. #
	my $ct_bases=@{$bases};
	my %base2indx;
	for (my $b=0; $b<$ct_bases; $b++) { $base2indx{$bases->[$b]} = $b; }

		## MAIN PROCESS:
		## Extract the distributions of the ancestral states at all sites (and at interior nodes) ##

	my %node2state_distri;
	my %node2freq_patterns;
	my %node2patterns;

	my $head;
	while ($head = <$fh_rst>) { # The main outer while-loop (over nodes). #

	     ($head =~ m/^Prob\s+of\s+best\s+state/) and last;

	     my ($node) = ($head =~ m/^Prob\s+distribution\s+at\s+node\s+([^\s\,]+)\,/);
	     (defined $node) or next;

	    my $fld_tags = <$fh_rst>;
	    unless ($fld_tags =~ m/site\s+Freq/) {
		$fld_tags =~ s/\s+$//;
		print STDOUT "[ERROR 'extract_sitewise_anc_state_distri_baseml'] Found ANOMALOUS FIELD TAGS for the node '$node': $! !\n";
		print STDOUT "'$fld_tags' .\n\n";
		return ($tree_no, undef, undef, undef, undef);
	    }

	    my @state_distri = ([]);
	    my @freq_patterns = my @patterns = ();

	    my $chunk = <$fh_rst>; 

	    unless ($fl_full_anc) {
		($node eq $topnode_id) or next; # Skip unless the node is the root.
	    }

	    my @lines = split /\n+/, $chunk;

	    while (my $line = shift @lines) { # Main middle while-loop (over sites). #

		my @fields = split ' ', $line;
		unless (@fields >= 3 + $ct_bases) {
		    $line =~ s/\s+$//;
		    print STDOUT "[ERROR 'extract_sitewise_anc_state_distri_baseml'] Found a LINE with TOO FEW FIELDS for the node '$node': $! !\n";
		    print STDOUT "'$line' .\n\n";
		    return ($tree_no, undef, undef, undef, undef);
		}

		my $site = shift @fields;
		unless ($site =~ m/^\d+$/) {
		    $line =~ s/\s+$//;
		    print STDOUT "[ERROR 'extract_sitewise_anc_state_distri_baseml'] Found an ANOMALOUS SITE LABEL, '$site,' in a line  for the node '$node': $! !\n";
		    print STDOUT "'$line' .\n\n";
		    return ($tree_no, undef, undef, undef, undef);
		}

		my $pfreq = shift @fields;
		my $pattern = shift @fields;

		
		$freq_patterns[$site] = $pfreq; 
		$patterns[$site] = $pattern;

		my @distri = ();
		for (my $b=0; $b<$ct_bases; $b++) { # Inner for-loop (over bases). #

		    my ($base, $bfreq) = ($fields[$b] =~ m/^(\w+)\(([0-9\.]+)\)/);
		    unless (defined $bfreq) {
			my $fld = $fields[$b];
			print STDOUT "[ERROR 'extract_sitewise_anc_state_distri_baseml'] Found an ANOMALOUS FIELD, '$fld,'  for the $site th site at the node '$node': $! !\n";
			return ($tree_no, undef, undef, undef, undef);
		    }

		    my $indx = $base2indx{$base};
		    unless (defined $indx) {
			print STDOUT "[ERROR 'extract_sitewise_anc_state_distri_baseml'] Found an ANOMALOUS BASE, '$base,'  for the $site th site at the node '$node': $! !\n";
			return ($tree_no, undef, undef, undef, undef);
		    }

		    $distri[$indx] = $bfreq;
		}			# End of the inner for-loop (over bases). #

		$state_distri[$site] = \@distri;

	    }			# End of the main middle while-loop (over sites).

	    $node2state_distri{$node} = \@state_distri;
	    $node2freq_patterns{$node} = \@freq_patterns;
	    $node2patterns{$node} = \@patterns;

	}		# End of the main outer while-loop (over nodes). #

	return ($tree_no, $tree_rst, \%node2state_distri, \%node2freq_patterns, \%node2patterns);

		# OUTPUT: ($tree_no, \@tree_rst, \%node2state_distr, \%node2freq_patterns, \%node2patterns).

}	# End of the subroutine 'extract_sitewise_anc_state_distri_baseml' #


# from_1based_to_0based_paml_hash:
#	This subroutine converts the coordinate system represented by the array index for the values of a hash storing some PAML output
#	from 1-based (as in the raw PAML output) to 0-based (as commonly seen in public genome annotation data).
#	
# INPUT: \%hash_out_paml, where
#	%hash_out_paml = ($key => \@array, ...), with
#	$array[$s] gives information on the $s th site in the 0-based coordinate convention.
#	Especially, $array[0] is supposed to contain no information.
#
# OUTPUT: 1 if the subroutine successfully ended.
#	  0 if it failed.
#
#	The subroutine converts the coordinate system for \@array, which is a value in  %hash_out_paml,
#	into 0-based by removing the 0 th element in the original @array.
#
sub from_1based_to_0based_paml_hash (\%) {

	my ($hash_out_paml) = @_;

	foreach my $key (keys %{$hash_out_paml}) { # Main foreach-loop (over hash keys). #

	   my $array = $hash_out_paml->{$key};
	   ((defined $array) and (@{$array}>0)) or next; # Skip if the array is undefined or empty.

	   my $null_element = shift @{$array}; # (ESSENTIAL!!) Remove the 0 th element.

	   (defined $null_element) or next;  # An undefined zero-th element is within the expectation. # 

	   my $ref = ref ($null_element);

	   unless (defined $ref) { # The 0 th element is a scalar. #

		print STDOUT "[ERROR 'from_1based_to_0based_paml_hash'] ILLEGAL INPUT: The 0-th element (='$null_element') is a scalar and DEFINED for the key '$key': $! !\n";
		return 0;
	   }

	   unless ($ref eq 'ARRAY') {
		print STDOUT "[ERROR 'from_1based_to_0based_paml_hash'] ILLEGAL INPUT: The 0-th element for the key '$key' is NEITHER a scalar NOR an array reference, but the reference to a $ref, contradicting the expectation: $! !\n";
		return 0;
	   }

		# Now, the 0 th element must be an array reference. #

	   unless (@{$null_element} == 0) {
		print STDOUT "[ERROR 'from_1based_to_0based_paml_hash'] ILLEGAL INPUT: The 0-th element for the key '$key' is a reference to an array, (", join (', ', @{$null_element}), "), which is NOT EMPTY, contradicting the expectation: $! !\n";
		return 0;
	   }
	}			# End of the main foreach-loop (over hash keys). #

		## The conversion completed. ##

	return 1;

}		# End of the subroutine 'from_1based_to_0based_paml_hash'. #


# INPUT: (\%node2state_distri, \@bases), where
#	%node2state_distri = ($node_id => \@state_distri, ...) for interior nodes, with
#	$node_id being consistent with @tree_rst, and 
#	$state_distri[$s]->[$b] being the probability of the $b th base (i.e. $bases[$b]) at the $s th site;
#	@bases stores the four bases. It MUST BE the same as that inputted to 'extract_sitewise_anc_state_distri_baseml'.
#
#	NOTE:  @{$state_distri[0]} could be either empty (for 1-based coordinates) or filled (for 0-based coordinates).
#
# OUTPUT: \%node2state_distri_revcom, where
#	%node2state_distri_revcom follows the same format as %node2state_distri,
#	but gives the ancestral base distributions for the reverse-complemented MSA.
#	If the input is 0-based, the output is also 0-based.
#	If the input is 1-based, the output is also 1-based.
#
#	Namely,
#	%node2state_distri_revcom = ($node_id => \@state_distri_revcom, ...) with
#	$state_distri_revcom[$s]->[$b] = $state_distri[$s_rev]->[$b_cmpl], where
#	$s_rev = $msa_len -1 - $s for 0-based,
#	       = $msa_len + 1 - $s for 1-based, and
#	$bases[$b_cmpl] being the complement of $bases[$b].
#
sub revcom_anc_base_state_distr (\%\@) {

	my ($node2state_distri, $bases) = @_;

		# Preliminary processes. #

	my $ct_bases = @{$bases};
	unless ($ct_bases == 4) {
	    print STDOUT "[ERROR 'revcom_anc_base_state_distr] ILLEGAL INPUT: The number of bases (= ", join (', ', @{$bases}), ") is NOT 4 : $! !\n"; 
	    return undef;
	}

	my %base2cmpl = ('T' => 'A', 'C' => 'G', 'A' => 'T', 'G' => 'C');
	##my %base2cmpl = ('T' => 'C', 'C' => 'T', 'A' => 'G', 'G' => 'A'); # A SERIOUS BUG fixed on Dec 31 (Tue), 2013. #

	my %base2indx;
	for (my $b=0; $b<$ct_bases; $b++) { $base2indx{uc ($bases->[$b])} = $b; }

	my %indx2cmpl; # = ($base_indx => $complement_base_indx, ...)

	foreach my $base (@{$bases}) { # Preliminary foreach-loop (over bases). #
	    $base = uc ($base);
	    my $indx = $base2indx{$base};
	    my $cmpl = $base2cmpl{$base};
	    unless (defined $cmpl) {
		print STDOUT "[ERROR 'revcom_anc_base_state_distr'] ILLEGAL INPUT: The $indx th base '$base' was NO COMPLEMENT: $! !\n";
		return undef;
	    }
	    my $cmpl_indx = $base2indx{$cmpl};
	    $indx2cmpl{$indx} = $cmpl_indx;
	}		# End of the preliminary foreach-loop (over bases). #

		## The MAIN PROCESS: Calculate the reverse-complement distribution. ##

	my %node2state_distri_revcom;

	foreach my $node (keys %{$node2state_distri}) { # Main outer foreach-loop (over ancestral nodes). #

	    my $state_distri = $node2state_distri->{$node};
	    ((defined $state_distri) and (@{$state_distri}>0)) or next;

	    my @state_distri_revcom = ();
	    my $start_coord = 0;
	    my $size = @{$state_distri};
	    if (@{$state_distri->[0]} == 0) { # 1-based coordinate system. #
		$start_coord = 1;
		push @state_distri_revcom, [];
	    }

	    for (my $s = $size-1; $s >= $start_coord; $s--) { # Main middle for-loop (over sites). #

		my $base_probs = $state_distri->[$s];
		my @cmpl_base_probs = ();

		for (my $b=0; $b<$ct_bases; $b++) { # Main inner for-loop (over bases). #
		    my $cmpl_indx = $indx2cmpl{$b};
		    push @cmpl_base_probs, $base_probs->[$cmpl_indx];
		}			# End of the main inner for-loop (over bases). #

		push @state_distri_revcom, \@cmpl_base_probs;

	    }					# End of the main middle for-loop (over sites). #

	    $node2state_distri_revcom{$node} = \@state_distri_revcom;

	}					# End of the main outer foreach-loop (over ancestral nodes). #

	return \%node2state_distri_revcom;

}		# End of the subroutine 'revcom_anc_base_state_distr'. #


# parse_brief_output_infer_psm_indel_events:
#	This subroutine parses the brief output of the Perl script 'br_infer_psm_indel_events.pl',
#	which calculates the 'probability distribution' of ancestral gap states (and indel scenarios),
#	based on the parsimonious indel scenarios.
#
# INPUT: (\$infile, $fl_full_nodes), where
#	$infile is the path to the input file, which is an brief output of 'br_infer_psm_indel_events.pl';
#	$fl_full_nodes determines on what nodes the ancestral states should be stored in the output:
#	$fl_full_nodes = 0 when you need only the state distribution on the top node,
#	               = 1 when you need those on all interior nodes,
#		       = 2 when you need those on all nodes (including exterior nodes).
#
# OUTPUT: (\@parameters, \@info_segments, \@indel_state_distributions), where
#	@parameters = (\@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with
#	@full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr) for the full MSA,
#	@seqids storing the sequence IDs,
#	$ct_segments = scalar (@info_segments) = scalar (@indel_state_distributions);
#
#	[ @{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, \@red_tree), with ] --- replaced with:
#	@{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_id), with 
#	$seg_start/end in the 0-based half-open convention,
#	[ @red_tree = ($rtopnode_id, \%rnode2pa, \%rnode2ch, \%rname2id, \%rid2name, \%rbr_attr) for the MSA reduced
#	according to the usability pattern; ] --- obsolete.
#	$red_topnode_id is the ID of the top-node (root) of the tree reduced according to the usability pattern
#	 ($red_topnode_id = undef for the segment skipped to infer ancestral states) ;
#
#	@sw_indel_state_distributons = @{$indel_state_distributons[$s]}
#	stores the indel state distributions for the $s th segment, and
#
#	@{$sw_indel_state_distributions[2*$cl]} = ($cl_start, $cl_end, $state)
#	for the gapless block left-flanking the $cl th gap-block cluster
#	(or on the right-end of the segment (with $cl = #{clusters})),
#	or (always with $cl = 0) for the skipped segment due to inadquate information.
#	$state	= 'X' for a gapless block,
#		= '?' for a skipped segment due to inadequate information.
#
#	@{$sw_indel_state_distributions[2*$cl + 1]}
#	= ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution) 
#	is for the $cl th gap-block cluster, where
#
#	@{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and
#
#	@{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
#	 is for the $ps th parsimonious indel scenario, with
#	
#	%node2bw_gap_states = ($node_id => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and
#	%br2indel_events = ($branch_id => \@events_on_branch, ...) only for branches experiencing indel events, with
#	@{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event along the branch $branch_id.
#
sub parse_brief_output_infer_psm_indel_events (\$$) {

	my ($infile, $fl_full_nodes) = @_;
	local $/ = ''; # Use blank lines as input-record-separators. #

	my $fh_in = open_input_file2 (${$infile});
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] 'open_input_file2' FAILED on the input file '${$infile}': $! !\n";
	    return (undef, undef, undef); 
	}

		## Read the input sequence tree. ##

	my $top_node_id0;

	while (my $in = <$fh_in>) {
	    ($in =~ m/^\[\[\s*Input\s+Tree/) or next;
	    ##($in =~ m/^\[\[\s*Input\s+tree/) or next; # A BUG fixed on Nov 26 (Tue), 2013. #
	    ($top_node_id0) = ($in =~ m/top_node_ID\s*\=\s*([^\s\(\)]+)/);
	    last;
	}
	unless (defined $top_node_id0) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT FIND the top-node ID: $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $full_tree = <$fh_in>;

	my %fnode2pa; my %fnode2ch; my %fname2id; my %fid2name; my %fbr_attr;

	unless ( input_ext_tree_file ($full_tree, %fnode2pa, %fnode2ch, %fname2id, %fid2name, %fbr_attr, 1) ) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] The subroutine 'input_ext_tree_file' FAILED to extract the full tree: $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $ftopnode_id = $fname2id{$TOP_NODE};
	unless ($ftopnode_id eq $top_node_id0) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: The top-node ID in the label (=$top_node_id0) CONFLICTS with the ID displayed in the tree (=$ftopnode_id): $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr);

		# @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr) for the full MSA,

		## Read the MSA information ##

	my $label_msa = <$fh_in>;

	my ($ct_seqs0) = ($label_msa =~ m/\(\#\{sequences\}\s*\=\s*(\d+)\)/);
	unless (defined $ct_seqs0) {
	    $label_msa =~ s/\s+$//;
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: The number of sequences canNOT be extracted from the label: $!\n";
	    print STDOUT "'$label_msa' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	(my $list_seqids = <$fh_in>) =~ s/\s*\.\s+$//; 
	##(my $list_seqids = <$fh_in>) =~ s/\.\s+$//; # A BUG fixed on Dec 19 (Thu), 2013. #
	$list_seqids =~ s/^Sequences\:\s+//;
	my @seqids = split /\,\s*/, $list_seqids;

	unless (@seqids == $ct_seqs0) {
	    my $ct_seqids = @seqids;
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: The number of sequences in the label (=$ct_seqs0) CONFLICTS with $ct_seqids, which is the number of IDs in the list: $! !\n";
	    print STDOUT ' ' x 3, join (', ', @seqids), " .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $descr_msa_len = <$fh_in>;
	my ($msa_len) = ($descr_msa_len =~ m/^\#\{columns\s+in\s+MSA\}\s*\=\s*(\d+)/);
	unless (defined $msa_len) {
	    $descr_msa_len =~ s/\s+$//;
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: The length of the MSA canNOT be extracted from the line: $!\n";
	    print STDOUT "'$descr_msa_len' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $descr_tgt_indx = <$fh_in>;
	my ($tgt_seqid, $indx_tgt) = ($descr_tgt_indx =~ m/^\{Index\s+for\s+the\s+target\s+sequence\s+\'([^\s\']+)\'\}\s*\=\s*(\S+)/ );
	unless (defined $indx_tgt) {
	    $descr_tgt_indx =~ s/\s+$//; 
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: The index of the target sequence canNOT be extracted from the line: $!\n";
	    print STDOUT "'$descr_tgt_indx' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $descr_ct_segs = <$fh_in>;
	my ($ct_segments) = ($descr_ct_segs =~ m/^\#\{MSA\s+Segments\}\s*\=\s*(\d+)/);
	unless (defined $ct_segments) {
	    $descr_ct_segs =~ s/\s+$//; 
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: The number of MSA segments canNOT be extracted from the line: $! !\n";
	    print STDOUT "'$descr_ct_segs' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my @parameters = (\@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments);

		# @parameters = (\@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with


		### MAIN PROCESS: Read the cluster-wise distributions of the indel scenarios
		### (and the associated ancestral gap states), which were bundled into segments.

	my @info_segments = my @indel_state_distributions = ();

	while (my $head_seg = <$fh_in>) { # Main outer while-loop (over segments) #

	    my ($seg_no) = ($head_seg =~ m/^\<\<\<\<\s+The\s+(\d+)\s+th\s+Segment/) ;
	    unless (defined $seg_no) {
		$head_seg =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT Extract the segment no. from the segment heading: $! !\n";
		print STDOUT "'$head_seg' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }

	    my ($seg_start, $seg_end) = ($head_seg =~ m/Region\s+\=\s+\[(\d+)\,\s*(\d+)\)/);
	    unless (defined $seg_end) {
		$head_seg =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT Extract the segment end coordinate from the segment heading: $! !\n";
		print STDOUT "'$head_seg' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }

	    my ($us_pattern) = ($head_seg =~ m/Usability\s+pattern\s+\=\s*\'([^\s\']+)\'/);
	    unless (defined $us_pattern) {
		$head_seg =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT Extract the segment end coordinate from the segment heading: $! !\n";
		print STDOUT "'$head_seg' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }

	    my @seg_info = ($seg_start, $seg_end, $us_pattern);
	    $info_segments[$seg_no] = \@seg_info;

		# @{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_id), with 
		# $seg_start/end in the 0-based half-open convention,
		# $red_topnode_id is the ID of the top-node (root) of the tree reduced according to the usability pattern
		#  ($red_topnode_id = undef for the segment skipped to infer ancestral states) ;

	    my $second_chunk = <$fh_in>;

	    if ($second_chunk =~ s/^\[\[\s+Reduced\s+Tree\s+//) {

			# Normal case. #

		my ($red_topnode_id) = ($second_chunk =~ m/\(top_node_ID\s*\=\s*([^\s\(\)]+)\)/);
		unless (defined $red_topnode_id) {
		    $second_chunk =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT Extract the top-node ID for the reduced tree from the second chunk: $! !\n";
		    print STDOUT "'$second_chunk' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}
		push @seg_info, $red_topnode_id;

			# Go ahead with reading the indel scenario distribution.

	    } elsif ($second_chunk =~ m/^SKIPPED\s+INFERRING/) {
	    ##} elsif ($second_chunk =~ m/^SKIPPED\s+INDERRING/) { # A BUG fixed on Nov 26 (Tue), 2013. #

			# The ancestral state inference was skipped for this segment. #

		my @sw_indel_state_distributions = ( [$seg_start, $seg_end, '?'] );
		$indel_state_distributions[$seg_no] = \@sw_indel_state_distributions;

		next;	# Go ahead to the next segment.
		
	    } else {
			# An Unexpected Case. #

		$second_chunk =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: an ANOMALOUS SECOND CHUNK for the $seg_no th segment: $! !\n";
		print STDOUT "'$second_chunk' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }

	    my $red_topnode_id = $seg_info[3];

	    my $red_tree = <$fh_in>; # Emit the chunk for the reduced tree.
	    my $heading_main = <$fh_in>; # Emit the heading that tells that the main description begins. 

	    my @sw_indel_state_distributions = ();

	    while (my $head_cls = <$fh_in>) { # Main middle while-loop (over clusters) #

		if ($head_cls =~ s/^\<\s+Gapless\s+block\s+// ) {

			# Gapless block. #

		    my ($bl_start, $bl_end) = ($head_cls =~ m/region\s+\=\s+\[(\d+)\,\s*(\d+)\)/ );
		    unless (defined $bl_end) {
			$head_cls =~ s/\s+$//;
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT extract the block-end coordinate from the following heading of the gapless block in the $seg_no th segment: $! !\n";
			print STDOUT "'< Gapless block $head_cls' .\n\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

		    my ($rfl_cl_no) = ($head_cls =~ m/^left\-flanking\s+the\s+(\d+)/) ;
		    ##my ($rfl_cl_no) = ($head_cls =~ m/^left\-flanking\s+the\s+(\d)/) ; # A BUG fixed on Nov 26 (Tue), 2013. #

		    unless (defined $rfl_cl_no) {
			if ($head_cls =~ m/^on\s+the\s+right\-end/) {

				# The right-end of the segment. #

			    push @sw_indel_state_distributions, [$bl_start, $bl_end, 'X'];

			    last; # End the middle while-loop.

			} else {
				# An unexpected case. #

			    $head_cls =~ s/\s+$//;
			    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT extract the right-flanking cluster number from the following heading of the gapless block in the $seg_no th segment: $! !\n";
			    print STDOUT "'< Gapless block $head_cls' .\n\n";
			    close ($fh_in);
			    return (undef, undef, undef);
			}

		    } # END of "unless (defined $right_cl_no) {...}"

		    unless (@sw_indel_state_distributions == 2* $rfl_cl_no) {
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] The size of the array, ", '@sw_indel_state_distributions (=', scalar (@sw_indel_state_distributions), ") is NOT EQUAL to twice the number of the right-flanking cluster (=$rfl_cl_no): $! !\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

		    $sw_indel_state_distributions[2*$rfl_cl_no] = [$bl_start, $bl_end, 'X'];

		    next; # Go ahead to the gap-block cluster on the right. 

			# @{$sw_indel_state_distributions[2*$cl]} = ($cl_start, $cl_end, $state)
			# for the gapless block left-flanking the $cl th gap-block cluster
			# (or on the right-end of the segment (with $cl = #{clusters})),
			# or (always with $cl = 0) for the skipped segment due to inadquate information.
			# $state	= 'X' for the gapless block,
			# 	= '?' for the skipped segment due to inadequate information.

		} # END of "if ($head_cls =~ s/^\<\s+Gapless\s+block\s+/) {...}"


			## Read information on the gap-block cluster. ##

		my ($cl_no) = ($head_cls =~ m/^\<\s+The\s+(\d+)\s+th\s+cluster/ );
		unless (defined $cl_no) {
		    $head_cls =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT extract the cluster number from the following heading of a cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_cls' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		unless (@sw_indel_state_distributions == 2 * $cl_no + 1) {
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] The temporary size of the array, ", '@sw_indel_state_distributions (=', scalar (@sw_indel_state_distributions), ') is NOT EQUAL to 2 * $cl_no + 1, where $cl_no = ', "$cl_no, for the $seg_no th segment: $! !\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my ($cl_start, $cl_end) = ($head_cls =~ m/region\s+\=\s+\[(\d+)\,\s*(\d+)\)/ );
		unless (defined $cl_end) {
		    $head_cls =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT extract the end-coordinate of the cluster from the following heading of a cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_cls' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my ($ct_gblks) = ($head_cls =~ m/\#\{gap\-blocks\}\s+\=\s+(\d+)/ );
		unless (defined $ct_gblks) {
		    $head_cls =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT extract the number of gap-blocks from the following heading of a cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_cls' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			# Read information on the gap-blocks. #

		my $head_gblocks = <$fh_in>;
		my $fld_tags_gblocks = <$fh_in>;
		my $info_gblocks = <$fh_in>;

		my @gap_blocks = ();

		foreach my $gblock (split /\n+/, $info_gblocks) { # 1st inner foreach-loop (over gap-blocks) #
		     my ($bl_no, $bl_len, $bl_start, $bl_end, $indx_gp, $gpattern) = split /\s+/, $gblock;
		     push @gap_blocks, [$bl_start, $bl_end, $indx_gp, $gpattern];

			# @{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and

		}					# End of the 1st inner foreach-loop (over gap-blocks) #

		unless (@gap_blocks == $ct_gblks) {
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] The size of ", '@gap_blocks (=', scalar (@gap_blocks), ") CONFLICTS with the number of gap-blocks (=$ct_gblks) described in the heading: $! !\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			# Read the distribution of parsimonious scenarios for a gap-block cluster. #

		my $heading_list_psm_scns = <$fh_in>;

		my ($min_totct_indels) = ($heading_list_psm_scns =~ m/Min_Total\#\{indel_events\}\s+\=\s+(\d+)/ );
		unless (defined $min_totct_indels) {
		    $heading_list_psm_scns =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT extract the minimum total count of indel events from the following heading for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$heading_list_psm_scns' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my ($ct_psm_scns) = ($heading_list_psm_scns =~ m/\#\{scenarios\}\s+\=\s+(\d+)/ );
		unless (defined $ct_psm_scns) {
		    $heading_list_psm_scns =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT extract the count of parsimonious indel scenarios from the following heading for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$heading_list_psm_scns' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my @scenario_distribution = ();

		while (my $head_scn = <$fh_in>) { # 2nd inner whlie-loop (over parsimonious scenarios). #

		    my ($scn_no, $rel_prob) = ($head_scn =~ m/^\{\s+Scenario\s+no\.\s+(\d+)\s*\;\s+rel_prob\.\s*\=\s*([^\s\{\}]+)/ );
		    unless (defined $rel_prob) {
			$head_scn =~ s/\s+$//;
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] ILLEGAL INPUT: CanNOT extract the relative probability from the following heading of a parsimonious scenario for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
			print STDOUT "'$head_scn' .\n\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }
		    unless (@scenario_distribution == $scn_no) {
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] The temporary size of ", '@scenario_distribution (=', scalar (@scenario_distribution), ") CLASHES with the parsimonious scenario no. (=$scn_no) for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

			# Read block-wise gap states mapped onto the nodes. #

		    my $head_bw_gstates = <$fh_in>;
		    my $fld_tags_bw_gstates = <$fh_in>;
		    my $info_bw_gstates = <$fh_in>;

		    my %node2bw_gap_states;
		    my $ct_rlv_nodes = my $ct_gaponly = 0; # A measure to avoid a buggy behavior. (introduced on Nov 26 (Tue), 2013). #

		    foreach my $line (split /\n+/, $info_bw_gstates) { # 1st inner-inner foreach-loop (over nodes). #
			my ($node_id, $cnct_bw_gstats) = split /\s+/, $line;

				# Introduced (on Nov 26 (Tue), 2013) in order to avoid a buggy behavior when encountering gap-only columns. #
			$ct_rlv_nodes++;
			($cnct_bw_gstats =~ m/X/) or ($ct_gaponly++);

			if ($fl_full_nodes <= 0) {
			    ($node_id eq $red_topnode_id) or next; # Skip except the top node.
			} elsif ($fl_full_nodes < 2) {
			    my $children = $fnode2ch{$node_id};
			    ((defined $children) and (@{$children}>0)) or next; # Skip an exterior node.
			}
			$node2bw_gap_states{$node_id} = $cnct_bw_gstats;

				# %node2bw_gap_states = ($node_id => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and

		    }						# End of the 1st inner-inner foreach-loop (over nodes). #

			# Read indel events mapped onto the branches. #

		    my $head_indels = <$fh_in>;
		    my $fld_tags_indels = <$fh_in>;

				#  MODIFIED (on Nov 26 (Tue), 2013) in order to avoid a buggy behavior when encountering gap-only columns. #
		    my $info_indels = ($ct_gaponly < $ct_rlv_nodes) ? <$fh_in> : '';
		    ##my $info_indels = <$fh_in>;

		    my %br2indel_events;

		    foreach my $line (split /\n+/, $info_indels) { # 2nd inner-inner foreach-loop (over indel-undergoing branches). #

			my @fields = split /\s+/, $line;
			my $br_id = shift @fields;

			my @events_on_branch = ();

			foreach my $fld (@fields) { # Inner-most foreach-loop (over indel events on the branch). # 
			    my ($stat_ue, $stat_le, $cnct_blks) = split /\//, $fld;
			    my @involved_blocks = split /\,/, $cnct_blks;
			    push @events_on_branch, [$stat_ue, $stat_le, \@involved_blocks];

			}		# End of the inner-most foreach-loop (over indel events on the branch). #

			$br2indel_events{$br_id} = \@events_on_branch;

			    # %br2indel_events = ($branch_id => \@events_on_branch, ...) only for branches experiencing indel events, with
			    # @{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event

		    }				# End of the 2nd inner-inner foreach-loop (over indel-undergoing branches). #

		    $scenario_distribution[$scn_no] = [$rel_prob, \%node2bw_gap_states, \%br2indel_events];


			# @{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
			#  is for the $ps th parsimonious indel scenario, with

		    ($scn_no == $ct_psm_scns - 1) and last; # We know that this is the last scenario. #


		}			# End of the 2nd inner while-loop (over parsimonious scenarios). #

		unless (@scenario_distribution == $ct_psm_scns) {
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] The size of ", '@scenario_distribution (=', scalar (@scenario_distribution), ") CLASHES with the number of parsimonious scenarios (=$ct_psm_scns) recorded in the heading: $! !\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}


		$sw_indel_state_distributions[2 * $cl_no + 1] = [$cl_start, $cl_end, 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution];

			# @{$sw_indel_state_distributions[2*$cl + 1]}
			# = ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution) 
			# is for the $cl th gap-block cluster, where

	    }			# End of the main middle while-loop (over clusters) #

	    $indel_state_distributions[$seg_no] = \@sw_indel_state_distributions;

		# @sw_indel_state_distributons = @{$indel_state_distributons[$s]}
		# stores the indel state distributions for the $s th segment, and

	}		# End of the main outer while-loop (over segments) #

	close ($fh_in);

=An earlier attempt (TO BE REMOVED if all are successful).
	push @indel_state_distributions, \@sw_indel_state_distributions;

	unless ((defined $seg_no) and ($seg_no == $ct_segments - 1)) {
	    (defined $seg_no) or ($seg_no = 'undef');
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] The final segment number (=$seg_no) CLASHES with the number of MSA segments (=$ct_segments) minus one: $! !\n";
	    return (undef, undef, undef);
	}
=cut

	unless (@info_segments == $ct_segments) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] The size of ", '@info_segments', " (=", scalar (@info_segments), ") CLASHES with the number of MSA segments (=$ct_segments): $! !\n";
	    return (undef, undef, undef);
	}
	unless (@indel_state_distributions == $ct_segments) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events'] The size of ", '@indel_state_distributions', " (=", scalar (@indel_state_distributions), ") CLASHES with the number of MSA segments (=$ct_segments): $! !\n";
	    return (undef, undef, undef);
	}

	return (\@parameters, \@info_segments, \@indel_state_distributions);

}	# End of the subroutine 'parse_brief_output_infer_psm_indel_events'


# revcom_distr_psm_indel_events:
#	This subroutine receives a main output of 'parse_brief_output_infer_psm_indel_events',
#	and returns the reverse-complemented version of the input.
#
# INPUT: ($msa_len, \@info_segments, \@indel_state_distributions), where
#	$msa_len is the length of the MSA from which the main inputs resulted;
#
#	@{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_id), with 
#	$seg_start/end in the 0-based half-open convention,
#	$red_topnode_id is the ID of the top-node (root) of the tree reduced according to the usability pattern
#	 ($red_topnode_id = undef for the segment skipped to infer ancestral states) ;
#
#	@sw_indel_state_distributons = @{$indel_state_distributons[$s]}
#	stores the indel state distributions for the $s th segment, and
#
#	@{$sw_indel_state_distributions[2*$cl]} = ($cl_start, $cl_end, $state)
#	for the gapless block left-flanking the $cl th gap-block cluster
#	(or on the right-end of the segment (with $cl = #{clusters})),
#	or (always with $cl = 0) for the skipped segment due to inadquate information.
#	$state	= 'X' for a gapless block,
#		= '?' for a skipped segment due to inadequate information.
#
#	@{$sw_indel_state_distributions[2*$cl + 1]}
#	= ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution) 
#	is for the $cl th gap-block cluster, where
#
#	@{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and
#
#	@{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
#	 is for the $ps th parsimonious indel scenario, with
#	
#	%node2bw_gap_states = ($node_id => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and
#	%br2indel_events = ($branch_id => \@events_on_branch, ...) only for branches experiencing indel events, with
#	@{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event along the branch $branch_id.
#
# OUTPUT: (\@revcom_info_segments, \@revcom_indel_state_distributions), where
#	@revcom_info_segments is the reverse-complemented version of @info_segments. More specifically, 
#	@{$rev_info_segments[$ct_segs - 1 - $s]} = ($msa_len - $seg_end, $msa_len - $seg_start, $usability_pattern, $red_topnode_id)
#	based on @{$info_segments[$s]} above;
#
#	@revcom_indel_state_distributions is the reverse-complemented version of @indel_state_distributions. More specifically,
#	@{$revcom_indel_state_distributions[$ct_segs - 1 - $s]} is based on @{$indel_state_distributions[$s]} above.
#	Then, all the clusters (and gapless blocks) were reversed in order,
#	all the coordinates were complemented (i.e., ($start, $end) -> ($msa_len - $end, $msa_len - $start)),
#	all the elements in @gap_blocks were reversed, 
#	all the elements (e.g., $indx_blk) in @involved_blocks were replaced with the complements (e.g., $ct_blocks - 1 - $indx_blk) with their order reversed,
#	and all the concatenated gap states in %node2bw_gap_states were reversed.
#
sub revcom_distr_psm_indel_events ($\@\@) {

	my ($msa_len, $info_segments, $indel_state_distrs) = @_;

	my $ct_segs = @{$info_segments};

	unless (@{$indel_state_distrs} == $ct_segs) {
	    print STDOUT "[ERROR 'revcom_distr_psm_indel_events'] ILLEGAL INPUT: The size of ", '@{$indel_state_distrs} (=', scalar (@{$indel_state_distrs}), ') CLASHES with that of @{$info_segments} (=', "$ct_segs\): $! !\n";
	    return (undef, undef);
	}

	my @rc_info_segments = my @rc_indel_state_distrs = ();

	for (my $s = 0; $s < $ct_segs; $s++) { # Main outer for-loop (over segments) #

	    my ($seg_start, $seg_end, $us_pattern, $red_topnode_id) = @{$info_segments->[$ct_segs - 1 - $s]};
	    my $sw_indel_state_distrs = $indel_state_distrs->[$ct_segs - 1 - $s];

	    my $rc_seg_start = $msa_len - $seg_end;
	    my $rc_seg_end = $msa_len - $seg_start;

	    $rc_info_segments[$s] = [$rc_seg_start, $rc_seg_end, $us_pattern, $red_topnode_id];

	    my @rc_sw_indel_state_distrs = ();

	    my $ct_cls_glblks = @{$sw_indel_state_distrs}; # Number of gap-block clusters and gapless blocks.

	    for (my $indx_cl = 0; $indx_cl < $ct_cls_glblks; $indx_cl++) { # Main middle for-loop (over gap-block clusters and gapless blocks). #
		my ($cl_start, $cl_end, $state, $gblocks, $min_totct_indels, $scn_distr) = @{$sw_indel_state_distrs->[$ct_cls_glblks - 1 - $indx_cl]};

			# Construct the basic information of the reverse-complemented cluster. #

		my $rc_cl_start = $msa_len - $cl_end;
		my $rc_cl_end = $msa_len - $cl_start;

		my @rc_cl_info = ($rc_cl_start, $rc_cl_end, $state);

		$rc_sw_indel_state_distrs[$indx_cl] = \@rc_cl_info;

		##unless ($state = 'X/-') { # A BUG fixed on Dec 4 (Wed), 2013. #
		unless ($state eq 'X/-') {
			# Either a gapless block or a segment with inadequate information. #
		    next;
		}

			# Construct the revserse-complemented set of gap-blocks. #

		my @rc_gblocks = ();
		my $ct_gblks = @{$gblocks};

		for (my $g = 0; $g < $ct_gblks; $g++) { # 1st main inner for-loop (over gap-blocks) #

		    my ($bl_start, $bl_end, $indx_gp, $gpattern) = @{$gblocks->[$ct_gblks-1 - $g]};

			# @{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and

		    $rc_gblocks[$g] = [$msa_len - $bl_end, $msa_len - $bl_start, $indx_gp, $gpattern];

		}			# End of the 1st main inner for-loop (over gap-blocks) #


			# Construct the reverse-complemented distribution of indel scenarios. #

		my @rc_scn_distr = ();
		my $ct_psm_scns = @{$scn_distr};

		for (my $ps = 0; $ps < $ct_psm_scns; $ps++) { # 2nd main inner for-loop (over parsimonious indel scenarios). #

		    my ($rel_prob, $node2bw_gstates, $br2indel_events) = @{$scn_distr->[$ps]};

			# @{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
			#  is for the $ps th parsimonious indel scenario, with

				# Map: node -> reversecomplemented block-wise gap states #
		    my %rc_node2bw_gstates;

		    foreach my $node (keys %{$node2bw_gstates}) { # 1st inner-inner foreach-loop (over nodes). #
			my $bw_gstates = $node2bw_gstates->{$node};
				# %node2bw_gap_states = ($node_id => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and

			(defined $bw_gstates) or next;
			$rc_node2bw_gstates{$node} = reverse ($bw_gstates) ;

		    }					# End of the 1st inner-inner foreach-loop (over nodes). #


				# Map: branch -> revserse-complemented indel events. #
		    my %rc_br2indel_events;

		    foreach my $br (keys %{$br2indel_events}) { # 2nd inner-inner foreach-loop (over branches undergoing indels). #

			my $indels_on_br = $br2indel_events->{$br};
			    # %br2indel_events = ($branch_id => \@events_on_branch, ...) only for branches experiencing indel events, with

			(defined $indels_on_br) or next;

			my @rc_indels_on_br = ();
			my $ct_eob = @{$indels_on_br};

			for (my $e=0; $e<$ct_eob; $e++) { # Inner-inner-inner for-loop (over events on the branch). #
			    my ($state_ue, $state_le, $involved_gblks) = @{$indels_on_br->[$e]};
				# @{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event along the branch $branch_id.

			    my @rc_involved_gblks = ();
			    my $ct_involved = @{$involved_gblks};
			    for (my $igb=0; $igb<$ct_involved; $igb++) { # Inner-most for-loop (over blocks involved in the event). #
				my $indx_gblk = $involved_gblks->[$ct_involved-1 - $igb];
				$rc_involved_gblks[$igb] = $ct_gblks - 1 - $indx_gblk;
			    }				# End of the inner-most for-looop (over blocks involved in the event). #

			    $rc_indels_on_br[$e] = [$state_ue, $state_le, \@rc_involved_gblks];

			}			# End of the inner-inner-inner for-loop (over events on the branch). #

			$rc_br2indel_events{$br} = \@rc_indels_on_br;

		    }					# End of the 2nd inner-inner foreach-loop (over branches undergoing indels). #

		    $rc_scn_distr[$ps] = [$rel_prob,  \%rc_node2bw_gstates, \%rc_br2indel_events];

		}				# End of the 2nd main inner for-loop (over parsimonious indel scenarios). #

		push @rc_cl_info, \@rc_gblocks, $min_totct_indels, \@rc_scn_distr;

	    }						# End of the main middle for-loop (over gap-block clusters and gapless blocks). #

	    $rc_indel_state_distrs[$s] = \@rc_sw_indel_state_distrs;

	}			# End of the main outer for-loop (over segments) #


	return (\@rc_info_segments, \@rc_indel_state_distrs);

}	# End of the subroutine 'revcom_distr_psm_indel_events'. #


# linear_order_indel_state_distribution:
#	This subroutine first receives an array, @indel_state_distributions,  which stores indel state distributions organized in a dual-hierarchical manner as:
#	the whole MSA ->  segments (of sites sharing the same usable/missing pattern) -> gap-block clusters (or gapless blocks),
#	and which is one of the outputs of 'parse_brief_output_infer_psm_indel_events'.
#	Then the subroutine re-organized @indel_state_distributions into a linear set, @linear_indel_state_distributions, which stores indel state distributions in a linear manner as:
#	the whole MSA -> gap-block clusters (or gapless blocks).
#
# INPUT: (\@indel_state_distributions), where
#	@sw_indel_state_distributons = @{$indel_state_distributons[$s]}
#	stores the indel state distributions for the $s th segment, and
#
#	@{$sw_indel_state_distributions[2*$cl]} = ($cl_start, $cl_end, $state)
#	for the gapless block left-flanking the $cl th gap-block cluster
#	(or on the right-end of the segment (with $cl = #{clusters})),
#	or (always with $cl = 0) for the skipped segment due to inadquate information.
#	$state	= 'X' for a gapless block,
#		= '?' for a skipped segment due to inadequate information.
#
#	@{$sw_indel_state_distributions[2*$cl + 1]}
#	= ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution) 
#	is for the $cl th gap-block cluster, where
#
#	@{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and
#
#	@{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
#	 is for the $ps th parsimonious indel scenario, with
#	
#	%node2bw_gap_states = ($node_id => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and
#	%br2indel_events = ($branch_id => \@events_on_branch, ...) only for branches experiencing indel events, with
#	@{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event along the branch $branch_id.
#
# OUTPUT: \@linear_indel_state_distributions, where
#
# 	@linear_indel_state_distributions stores indel state distributions in a linear manner, so that
#
#	@{$linear_indel_state_distributions[$ser_cl]} is for the $ser_cl th cluster in the serial order from the left-end to the right-end.
#	= ($indx_seg, $cl_start, $cl_end, $state)
#		if $state = '?' (for segments skipped due to inadequate informaiton) or 'X' (for gapless blocks), and
#	= ($indx_seg, $cl_start, $cl_end, $state, \@gap_blocks, $min_totct_indels, \@scenario_distribution)
#		if $state = 'X/-' (for gap-block clusters, excluding gapless blocks).
#	Here $indx_seg is the index of the segment that accommodated the cluster in the original @indel_state_distributions, 
#
sub linear_order_indel_state_distributions (\@) {

	my ($indel_state_distributions) = @_;

	my $ct_segments = @{$indel_state_distributions};

	my @linear_indel_state_distributions = ();

	for (my $indx_seg = 0; $indx_seg < $ct_segments; $indx_seg++) { # Main for-loop (over segments). #

	    foreach my $clwise_distri (@{$indel_state_distributions->[$indx_seg]}) { # Inner foreach-loop (over clusters and gapless blocks). #
		my @new_clwise_distri = ($indx_seg, @{$clwise_distri});
		push @linear_indel_state_distributions, \@new_clwise_distri;

	    }			# End of the inner foreach-loop (over clusters and gapless blocks). #

	}		# End of the main for-loop (over segments). #

	return \@linear_indel_state_distributions;

}	# End of the subroutine 'linear_order_indel_state_distributions' #


1;		# a "spell" which makes the module "use"-able and/or "require"-able #


