package MyLinearRegression;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.5: Copyright (C) 2014 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyLinearRegression.pm:
#	This package contains subroutines that perform linear regression analyses, as well as correlation analyses.
#	(CREATION STARTED on Apr 17 (Thu), 2014, ENDED on Apr 18 (Fri), 2014)
#

use strict;
use warnings;

use Exporter;

use MyProbDistri qw( pv_t_distri ); # Necessary for examining whether the correlation is significant or not.

our @ISA = qw( Exporter ); 

our @EXPORT_OK = qw(
	correl_linregr_sglexpl_wgt
	);


	# Level of the test output. #
	# (0, silent; 1, very brief; 2, brief; 3, medium; 4, verbose; 5, very verbose) #
our $TEST_LEVEL = 0; 

our $INFTY = 1E+50;


# correl_linregr_sglexpl_wgt:
#	This subroutine performs a correlation and a linear regression analyses,
#	with a single explanatory variable and with each data point weighted.
#
#	The analyses, especially the significance analyses,
#	are justified only when the following assumptions are fulfilled:
#	(i) the data points are independent; and
#	(ii) the (expected) standard error of the response variable (i.e., y) is 1/sqrt(weight_factor).
#
# INPUT: (\@weights, \@xs, \@ys), where
#	$weights[$i], $xs[$i], $ys[$i] are the weight, explanatory variable, and the response variable
#	for the $i th datapoint.
#
# OUTPUT: (\@wgtd_sums, \@wgtd_aves_vars, \@results_correl, \@results_linregr), where
#
#	@wgtd_sums = ($ct_eff_datapoints, $sum_wgt, $sum_x_wgtd, $sum_y_wgtd, $sum_xx_wgtd, $sum_yy_wgtd, $sum_xy_wgtd), where
#	$ct_eff_datapoints is the number of "effective datapoints", at each of which $weights[$i], $xs[$i], and $ys[$i] all have defined numerical values),
#	$sum_wgt = Sum_i $weights[$i], where the summation is over all "effective datapoints",
#	$sum_x/y_wgtd = Sum_i $weights[$i] * $xs/ys[$i], over all "effective datapoints",
#	$sum_xx_wgtd = Sum_i $weights[$i] * $xs[$i] * $xs[$i], over all "effective datapoints",
#	$sum_yy_wgtd and $sum_xy_wgtd are also defined similarly to $sum_xx_wgtd;
#
#	@wgtd_aves_vars = ($wgtd_ave_x, $wgtd_ave_y, $wgtd_var_x, $wgtd_var_y, $wgtd_cov_xy), where 
#	$wgtd_ave_x/y = $sum_x/y_wgtd / $sum_wgt ,
#	$wgtd_var_x/y = ($sum_xx/yy_wgtd - $sum_wgt * $wgtd_ave_x/y * $wgtd_ave_x/y) /($ct_eff_datapoints -1),
#	$wgtd_cov_xy  = ($sum_xy_wgtd - $sum_wgt * $wgtd_ave_x * $wgtd_ave_y) /($ct_eff_datapoints -1);
#
#	@results_correl = ($corr_coeff_wgtd, $dof, $t_stat_wgtd, $p_value), where
#	$corr_coeff_wgtd = $wgtd_cov_xy /sqrt($wgtd_var_x * $wgtd_var_y),
#	$dof = $ct_eff_datapoints -2,
#	$t_stat_wgtd = $corr_coeff_wgtd * sqrt ( $dof /(1 - $corr_coeff_wgtd*$corr_coeff_wgtd) ),
#	$p_value is calculated as a T-test with $dof degrees of freedom using $t_stat_wgtd as a T-statistic;
#
#	@results_linregr = ($slope, $se_slope, $y_intercept, $se_y_intercept, $s2_xy), where
#	$s2_xy = ($wgtd_var_y - $wgtd_cov_xy*$wgtd_cov_xy/$wgtd_var_x) * ($ct_eff_datapoints-1)/$dof ,
#	$slope = $wgtd_cov_xy /$wgtd_var_x ,
#	$se_slope = sqrt ($s2_xy/(($ct_eff_datapoints-1)*$wgtd_var_x)),
#	$y_intercept = $wgtd_ave_y - $slope * $wgtd_ave_x,
#	$se_y_intercept = sqrt ($s2_xy * (1/$sum_wgt + $wgtd_ave_x*$wgtd_ave_x/(($ct_eff_datapoints-1)*$wgtd_var_x)) ). 
#
sub correl_linregr_sglexpl_wgt (\@\@\@) {

	my ($weights, $xs, $ys) = @_;

	my $datasize = @{$weights};

    (print STDOUT "  < Subroutine 'correl_linregr_sglexpl_wgt'  (datasize = $datasize) >\n\n") if ($TEST_LEVEL>0);

	my $ct_eff_dpts = my $sum_wgt = my $sum_x_wgtd = my $sum_y_wgtd = 0;
	my $sum_xx_wgtd = my $sum_yy_wgtd = my $sum_xy_wgtd = 0;

	for (my $i=0; $i<$datasize; $i++) { # Main for-loop (over datapoints). #

	    my $w = $weights->[$i]; # weight factor. 
	    my $x = $xs->[$i]; # Explanatory variable.
	    my $y = $ys->[$i]; # Response variable.

	    ((defined $w) and (defined $x) and (defined $y)) or next; # Skip if any of them are undefined.
	    ($w > 0) or next; # Skip if the weight is non-positive (a negative weight indicates that the datapoint is NOT well-defined).

    (print STDOUT "    [ The $i th datapoint (weight = $w, expl_variable = $x, resp_variable = $y) ]\n\n") if ($TEST_LEVEL>2);

	    $ct_eff_dpts++;

	    $sum_wgt += $w;
	    $sum_x_wgtd += $w*$x;
	    $sum_y_wgtd += $w*$y;

	    $sum_xx_wgtd += $w*$x*$x;
	    $sum_yy_wgtd += $w*$y*$y;
	    $sum_xy_wgtd += $w*$x*$y;

    if ($TEST_LEVEL>3) {
	print STDOUT "    ( Interim Result )\n\n";
	print STDOUT '     $ct_eff_dpts = ', $ct_eff_dpts, " ,\n";
	print STDOUT '     $sum_wgt     = ', $sum_wgt, " ,\n";
	print STDOUT '     $sum_x_wgtd  = ', $sum_x_wgtd, " ,\n";
	print STDOUT '     $sum_y_wgtd  = ', $sum_y_wgtd, " ,\n\n";
	print STDOUT '     $sum_xx_wgtd = ', $sum_xx_wgtd, " ,\n";
	print STDOUT '     $sum_yy_wgtd = ', $sum_yy_wgtd, " ,\n";
	print STDOUT '     $sum_xy_wgtd = ', $sum_xy_wgtd, " .\n\n\n";
    }

	}			# End of the main for-loop (over datapoints). #

    if ($TEST_LEVEL>1) {
	print STDOUT "\n";
	print STDOUT "   [[ FINAL Result (I: basic summations) ]]\n\n";
	print STDOUT '   $ct_eff_dpts = ', $ct_eff_dpts, " ,\n";
	print STDOUT '   $sum_wgt     = ', $sum_wgt, " ,\n";
	print STDOUT '   $sum_x_wgtd  = ', $sum_x_wgtd, " ,\n";
	print STDOUT '   $sum_y_wgtd  = ', $sum_y_wgtd, " ,\n\n";
	print STDOUT '   $sum_xx_wgtd = ', $sum_xx_wgtd, " ,\n";
	print STDOUT '   $sum_yy_wgtd = ', $sum_yy_wgtd, " ,\n";
	print STDOUT '   $sum_xy_wgtd = ', $sum_xy_wgtd, " .\n\n\n";
    }

	my @wgtd_sums = ($ct_eff_dpts, $sum_wgt, $sum_x_wgtd, $sum_y_wgtd, $sum_xx_wgtd, $sum_yy_wgtd, $sum_xy_wgtd);

	if ($ct_eff_dpts < 3) {
	##if ($ct_eff_dpts>2) {...} # A BUG fixed on Apr 18 (Fri), 2014. #
	    print STDOUT "[ERROR 'correl_linregr_sglexpl_wgt'] The number of effective datapoints (=$ct_eff_dpts) is NOT enough for performing the correlation & regression analyses: $! !\n\n";
	    return (\@wgtd_sums, undef, undef, undef);
	}

	my $wgtd_ave_x = $sum_x_wgtd /$sum_wgt;
	my $wgtd_ave_y = $sum_y_wgtd /$sum_wgt;
	my $wgtd_var_x = ($sum_xx_wgtd - $sum_x_wgtd *$sum_x_wgtd/$sum_wgt)/($ct_eff_dpts-1);
	my $wgtd_var_y = ($sum_yy_wgtd - $sum_y_wgtd *$sum_y_wgtd/$sum_wgt)/($ct_eff_dpts-1);
	my $wgtd_cov_xy = ($sum_xy_wgtd - $sum_x_wgtd *$sum_y_wgtd/$sum_wgt)/($ct_eff_dpts-1);

	my @wgtd_aves_vars = ($wgtd_ave_x, $wgtd_ave_y, $wgtd_var_x, $wgtd_var_y, $wgtd_cov_xy);

    if ($TEST_LEVEL>1) {
	print STDOUT "   [[ FINAL Result (II: averages & (co)variances) ]]\n\n";
	print STDOUT '   $wgtd_ave_x  = ', $wgtd_ave_x, " ,\n";
	print STDOUT '   $wgtd_ave_y  = ', $wgtd_ave_y, " ,\n\n";
	print STDOUT '   $wgtd_var_x  = ', $wgtd_var_x, " ,\n";
	print STDOUT '   $wgtd_var_y  = ', $wgtd_var_y, " ,\n";
	print STDOUT '   $wgtd_cov_xy = ', $wgtd_cov_xy, " .\n\n\n";
    }

	unless ( ($wgtd_var_x > 0) and ($wgtd_var_y > 0) ) {
	    print STDOUT "[ERROR 'correl_linregr_sglexpl_wgt'] Found (a) NON-POSITIVE VARIANCE(s): ", '($wgtd_var_x, $wgtd_var_y) = ', "($wgtd_var_x, $wgtd_var_y): $! !\n\n";
	    return (\@wgtd_sums, \@wgtd_aves_vars, undef, undef);
	}


	my $corr_coeff_wgtd = $wgtd_cov_xy /sqrt($wgtd_var_x * $wgtd_var_y);
	my $if_perfect = ($corr_coeff_wgtd *$corr_coeff_wgtd >= 1) ? 1 : 0;

	my $dof = $ct_eff_dpts -2;
	my $t_stat_wgtd = $if_perfect ? $INFTY : $corr_coeff_wgtd * sqrt ( $dof /(1 - $corr_coeff_wgtd*$corr_coeff_wgtd) );

	my $p_value = $if_perfect ? 0 : pv_t_distri ($t_stat_wgtd, $dof);

    if ($TEST_LEVEL>1) {
	print STDOUT "   [[ FINAL Result (IIIA: correlation analysis) ]]\n\n";
	print STDOUT '   $corr_coeff_wgtd = ', $corr_coeff_wgtd, " ,\n"; 
	print STDOUT '   $dof             = ', $dof, " ,\n";
	print STDOUT '   $t_stat_wgtd     = ', $t_stat_wgtd, " ,\n";
	print STDOUT '   $p_value         = ', ((defined $p_value) ? $p_value : 'undef') , " .\n\n\n";
    }

	unless (defined $p_value) {
	    print STDOUT "[ERROR 'correl_linregr_sglexpl_wgt'] 'pv_t_distri' FAILED to calculate the P-value for the correlation: $! !\n\n";
	    my @results_correl = ($corr_coeff_wgtd, $dof, $t_stat_wgtd, $p_value);
	    return (\@wgtd_sums, \@wgtd_aves_vars, \@results_correl, undef);
	}

	my @results_correl = ($corr_coeff_wgtd, $dof, $t_stat_wgtd, $p_value);


		# MODIFIED to handle situations where ($wgtd_var_x*$wgtd_var_y - $wgtd_cov_xy*$wgtd_cov_xy) is infinitesimally negative (on Apr 18 (Fri), 2014). #
	my $s2_xy = $if_perfect ? 0 : ($wgtd_var_y - $wgtd_cov_xy*$wgtd_cov_xy/$wgtd_var_x) * ($ct_eff_dpts-1)/$dof ;
	##my $s2_xy = ($wgtd_var_y - $wgtd_cov_xy*$wgtd_cov_xy/$wgtd_var_x) * ($ct_eff_dpts-1)/$dof 

	my $slope = $wgtd_cov_xy /$wgtd_var_x;
	##my $slope = $wgtd_cov_xy /$wgtd_var_y; # A BUG fixed on Apr 18 (Fri), 2014.#

	my $se_slope = sqrt ($s2_xy/(($ct_eff_dpts-1)*$wgtd_var_x));

	my $y_intercept = $wgtd_ave_y - $slope * $wgtd_ave_x;
	my $se_y_intercept = sqrt ($s2_xy * (1/$sum_wgt + $wgtd_ave_x*$wgtd_ave_x/(($ct_eff_dpts-1)*$wgtd_var_x)) );

	my @results_linregr = ($slope, $se_slope, $y_intercept, $se_y_intercept, $s2_xy);

    if ($TEST_LEVEL>1) {
	print STDOUT "   [[ FINAL Result (IIIB: linear regression) ]]\n\n";
	print STDOUT '   $s2_xy          = ', $s2_xy, " ,\n\n";
	print STDOUT '   $slope          = ', $slope, " ,\n";
	print STDOUT '   $se_slope       = ', $se_slope, " ,\n";
	print STDOUT '   $y_intercept    = ', $y_intercept, " ,\n";
	print STDOUT '   $se_y_intercept = ', $se_y_intercept, " .\n\n\n";
    }

   (print STDOUT "\n  < END of the subroutine 'correl_linregr_sglexpl_wgt' >\n\n") if ($TEST_LEVEL>0);

	return (\@wgtd_sums, \@wgtd_aves_vars, \@results_correl, \@results_linregr) ;

}	# End of the subroutine 'correl_linregr_sglexpl_wgt'. #


1; 	# a 'spell' which makes the module 'use'-able and/or 'require'-able #


