package MyMisc;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyMisc.pm : a module for subroutines that are convenient
#	but not classified into any specific categories
#	defined by my modules.
#
use strict;
use warnings;
use Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
	ordinal_suffix construct_label2indx
	count_spec_chars
	indices_relv_elements elements_in_intvls
	exact_match_strings string_intvl_set convert_string2intvls
	enumerate_indx_combinations
	);


# ordinal_suffix:
#	This subroutine makes a suffix for an ordinal number,
#	like 'st' for 1st, 'nd' for 2nd, etc.
#
# INPUT: $cardinal, which is an integer.
#
# OUTPUT: $ord_suffix.
#
sub ordinal_suffix ($) {

	my ($cardinal) = @_;

	unless ($cardinal  == int $cardinal) {
	    print STDERR "ILLEGAL INPUT for 'ordinal_suffix': $cardinal.\n";
	    return undef;
	}

	$cardinal = sprintf "%d", $cardinal; # We cannot be too careful.

	my ($last) = ($cardinal =~ m/(\d)$/ );
	my ($sec_last) = ($cardinal =~ m/(\d)\d$/ );

	if ($last == 1) {
	   return ((defined $sec_last) and ($sec_last ==1)) ? 'th' : 'st' ;
	} elsif ($last == 2) {
	   return ((defined $sec_last) and ($sec_last ==1)) ? 'th' : 'nd' ;
	} elsif ($last == 3) {
	   return ((defined $sec_last) and ($sec_last ==1)) ? 'th' : 'rd' ;
	} else {
	    return 'th';
	}

	print STDERR "Some ERROR has occurred in 'ordinal_suffix'.\n";

	return undef;

}	# End of the subroutine 'ordinal_suffix' #


# construct_label2indx:
#	This subroutine constructs a hash table
#	that associates each label
#	to a list of corresponding indices in the input array .
#
# INPUT: (\@array, $indx_label), where
#	$array[$m]->[$indx_label] (or $array[$m] if $indx_label < 0)
#	gives a label that specifies the $m th element.
#
# OUTPUT: \%label2indx, where
#	%label2indx = ($label => [a list of indices], ...).
#
sub construct_label2indx (\@$) {

	my ($array, $indx_label) = @_;
	my $ct_ele = @{$array};

	my %label2indx;

	for (my $k=0; $k<$ct_ele; $k++ ) { # The outer for-loop #
	    my $label = ($indx_label < 0) ?
		$array->[$k] : $array->[$k][$indx_label] ;
	    if (defined $label2indx{$label}) {
		push @{$label2indx{$label}}, $k;
	    } else {
		$label2indx{$label} = [$k];
	    }
	}			# End of the outer for-loop #

	return \%label2indx;

}		# End of the subroutine 'construct_label2indx' #


# count_spec_chars:
#	This subroutine counts specified characters in the input string.
#
# INPUT: (\$string, \@spec_chars, $fl_caseignore), where
#	@spec_chars stores the characters to be counted;
#	$fl_caseignore = 1 if you ignore the differences
#		between upper- and lower- case letters,
#			=0 if the differences matter.
#
# OUTPUT: $ct_characters.
#
sub count_spec_chars (\$\@$) { 

	my ($string, $spec_chars, $fl_caseignore) = @_;

	my %spec_chars;
	foreach my $char (@{$spec_chars}) { # The 1st foreach-loop # 
	    ($fl_caseignore) and ($char = uc $char);
	     $spec_chars{$char} = 1;
	}			# End of the 1st foreach-loop # 

	my $cp_string = ($fl_caseignore) ? uc (${$string}) : ${$string} ;
	my $len = length $cp_string;

	my $ct_chars = 0;
	for (my $l=0; $l<$len; $l++) {
	   my $char = substr ($cp_string, 0, 1, '');
	   (defined $spec_chars{$char}) and ($ct_chars++);
	}

	return $ct_chars;

}		# End of the subroutine 'count_spec_chars' #


# indices_relv_elements:
#	This subroutine enumerates the indices of relevant elements contained in an input array,
#
# INPUT: (\@relv_labels, \@in_labels), where
#	@in_labels is the input set of labels;
#	@relv_labels = (a set of labels for the element we want).
#
# OUTPUT: \@relv_indices, where
#	@relv_indices is the set of indies for relevant elements.
#
sub indices_relv_elements (\@\@) {

	my ($relv_labels, $in_labels) = @_;

	my $ct_els = @{$in_labels};
	($ct_els>0) or (return []);

	my %label2relv; 
	foreach my $label (@{$relv_labels}) { $label2relv{$label} = 1; }

	my @relv_indices = ();
	for (my $m=0; $m<$ct_els; $m++) { # The main for-loop #
	    (defined $label2relv{$in_labels->[$m]}) and (push @relv_indices, $m);
	}			# End of the main for-loop #

	return \@relv_indices;

}		# end of the subroutine 'indices_relv_elements' #


# elements_in_intvls:
#	This subroutine exaimes whether each element in an inpu array
#	falls into a set of intervals or not.
#
# INPUT: (\@array, $indx, \@intvls, $indx_beg, $indx_end, $fl_leq, $fl_req), where
#	$array[$m]->[$indx] (or $array[$m] if $indx <0 or $indx = undef)
#	is the quantity used to see if
#	the $m th element is included in the set of intervals;
#	($beg_intvl, $end_intvl) = @{$intvls[$k]}[$indx_beg, $indx_end]
#	gives the boundaries of the $k th interval;
#	$fl_leq/req = 1 if the equation for the left/right boundary
#		also indicates that the element reside in the interval,
#		    = 0 otherwise.
#	The subroutine ASSUMES that @array and @intvls are appropriately sorted
#	beforehand, and the intervals do NOT overlap with one another.
#
# OUTPUT: (\@in_intvls, \@out_intvls), where
#	@in_intvls is a list of indices in @array
#	that reside in the set of intervals;
#	@out_intvls is a list of indices that lie out of the set.
#
sub elements_in_intvls (\@$\@$$$$) {

	my ($array, $indx, $intvls, $indx_beg, $indx_end, $fl_leq, $fl_req) = @_;

	my $ct_els = @{$array};
	my @intvls = @{$intvls};
	(@intvls >0) or (return ([], [0.. $ct_els-1]) ); # no elements cannot be in the intervals.

	my @in_intvls = my @out_intvls = ();

	my $pres_intvl = shift @intvls;
	my ($beg_intvl, $end_intvl) = @{$pres_intvl}[$indx_beg, $indx_end];

	OUTER:
	for (my $m=0; $m<$ct_els; $m++) { # The outermost for-loop #

	    my $quantity = ((defined $indx) and ($indx >=0) ) ?
			$array->[$m][$indx] : $array->[$m] ;

		# skip intervals while they are on the left of the element.
	    while ( ($quantity > $end_intvl) or
		(($fl_req == 0) and ($quantity == $end_intvl)) ) {
				# the inner while-loop #

		$pres_intvl = shift @intvls;
		unless (defined $pres_intvl) { # we've run out of intervals
		    push @out_intvls, ($m .. $ct_els-1);
		    last OUTER;
		}

		($beg_intvl, $end_intvl) = @{$pres_intvl}[$indx_beg, $indx_end];

	    }			# End of the inner while-loop #

	    if ( ($quantity < $beg_intvl) or
		 (($fl_leq == 0) and ($quantity == $beg_intvl)) ) {
		   # the element is on the left of the interval #

		push @out_intvls, $m;
		next;
	     }

	    if ( ($quantity < $end_intvl) or
		 (($fl_req >0) and ($quantity == $end_intvl)) ) {
		# the element is in the interval #

		push @in_intvls, $m;
		next;
	    }
	}			# End of the outermost for-loop #

	return (\@in_intvls, \@out_intvls);

}		# end of the subroutine 'elements_in_intvls' #


# exact_match_strings;
#	This subroutine examine whether a 'query' string ('words')
#	exactly matches a part of the 'subject' string.
#	NOTE that this subroutine chews the subject from the upstream.
#
# INPUT: ($query, \$sbjct, $fl_bwords), where
#	$query is the 'query' string;
#	$sbjct is the 'subject' string;
#	$fl_bwords = 1 if we limit the start of $query to the starts of words,
#		   = 0 if we search everywhere.
#
# OUTPUT: \@start_pos, where
#	@start_pos stores starting positions if $query exists (counting from 1).
#	It is empty it $query does NOT exist.
#
sub exact_match_strings ($\$$) {

	my ($query, $sbjct, $fl_bwords) = @_;

	my $len_qu = length $query;
	my $len_sb = length ${$sbjct};
	my $up_limit = $len_sb - $len_qu + 1;
	my $first = substr ($query, 0, 1, '');
		# NOTE that we have removed the 1st character from the query.
	my @start_pos = ();

=for debugging the subroutine
	print STDERR "\n < 'exact_match_strings' >\n\n";
=cut

	OUTER_EXACT:
	for (my $i=0; $i<$up_limit; $i++) { # The outer for-loop #

	    if ($fl_bwords) {
		my ($spacer) = (${$sbjct} =~ m/^(\W+)/);
		if (defined $spacer) {
		    my $len_sp = length ($spacer);
		    $i += $len_sp;
		    substr (${$sbjct}, 0, $len_sp, '');
		}
	    }

=for debugging the subroutine
	    print STDERR "Start: ", $i+1, "\n";
	    print STDERR "'${$sbjct}'\n\n";
=cut
	    my $char = substr (${$sbjct}, 0, 1, '');
	    if ($char eq $first) {
		my $rest = ($len_qu >1) ? substr (${$sbjct}, 0, $len_qu -1) : '';
		($query eq $rest) and (push @start_pos, $i+1);
	    }

	   if ($fl_bwords) {
		my ($wfrag) = (${$sbjct} =~ m/^(\w+)/);
		if (defined $wfrag) {
		     my $len_frag = length ($wfrag);
		     $i += $len_frag;
		     substr (${$sbjct}, 0, $len_frag, '');
		}
	   }
	}			# end of the outer for-loop #

	return \@start_pos;

}		# end of the subroutine 'exact_match_strings' #


# string_intvl_set:
#	This subroutine makes a string showing a set of intervals.
#
# INPUT: (\@intvl_set, $indx_beg, $indx_end), where
#	@{$intvl_set[$r]}[$indx_beg, $indx_end] = ($beg, $end)
#	gives the boundaries of the $r th interval.
#
# OUTPUT: \$string_intvl_set, where
#	$string_intvl_set = "[$beg1,$end1][$beg2,$end2]...".
#
sub string_intvl_set (\@$$) {

	my ($intvl_set, $indx_beg, $indx_end) = @_;

	my $string = '';

	foreach my $intvl (@{$intvl_set}) { $string .= sprintf ("\[%s\,%s\]", @{$intvl}[$indx_beg, $indx_end] ); }

	return \$string;

}		# End of the subroutine 'string_intvl_set' #


# convert_string2intvls:
#	This subroutine is an 'inverse function' of 'string_intvl_set',
#	and makes a set of intervals out of the string representing one.
#
# INPUT: \$string, which follows the format
#	"[$beg1,$end1][$beg2,$end2]...".
#
# OUTPUT: \@intvl_set, where
#	@{$intvl_set[$r]} = ($beg, $end) gives the boundaries of the $r th interval.
#
#	NOTE that it doesn't care about any consistency of the intervals
#	(, or even whether $beg and $end are numbers or not).
#
sub convert_string2intvls (\$) {

	my ($string) = @_;
	$string = ${$string};

	$string =~ s/^\s*\[\s*//;
	$string =~ s/\s*\]\s*$//;

	my @bd_pairs = split /\s*\]\[\s*/, $string;

	my @intvl_set = ();

	while (my $bd_pair = shift @bd_pairs) {
	    my ($beg, $end) = split /\s*\,\s*/, $bd_pair;
	    unless (defined $end) {
		print STDERR "[ERROR 'convert_string2intvls'] Found an ANOMALOUS 'BOUNDARY  PAIR': '$bd_pair' !\n";
		return undef;
	    }

	    push  @intvl_set, [$beg, $end];
	}

	return \@intvl_set ;

}			# End of the subroutine 'convert_string2intvls' #


# enumerate_indx_combinations:
#	This subroutine returns all possible combinations
#	of the indices that an array of arrays can take.
#
# INPUT: \@array_of_arrays, where
#	@{$array_of_arrays[$k]} is the $k th array.
#
# OUTPUT: \@set_indx_combi, where
#	$set_indx_combi[$p]->[$k] gives the index for the $k th array
#	in the $p th combination.
#	It is undef if the $k th array is empty or undefined.
#
sub enumerate_indx_combinations (\@) {

	my ($array_of_arrays) = @_;

	my $ct_arrays = @{$array_of_arrays};
	my @set_indx_combi = ([]);

	for (my $k=0; $k<$ct_arrays; $k++) { # Main for-loop #
	    my $array = $array_of_arrays->[$k];
	    unless ((defined $array) and (@{$array}>0)) {
		foreach my $combi (@set_indx_combi) { $combi->[$k] = undef; }
		next;
	    }

	    my $size_array = @{$array};
	    my @new_set_indx_combi = ();
	    foreach my $combi (@set_indx_combi) { # Middle foreach-loop #
		for (my $i=0; $i<$size_array; $i++) { # Inner foreach-loop #
		    my @cp_combi = @{$combi};
		    $cp_combi[$k] = $i;
		    push @new_set_indx_combi, \@cp_combi; 
		}			# End of the inner foreach-loop #
	    }				# End of the middle foreach-loop #

		# Update the set #
		@set_indx_combi = @new_set_indx_combi;
	}			# End of the main for-loop #

	return \@set_indx_combi;

}	# End of the subroutine 'enumerate_indx_combinations' #


1;	# a spell to enable the subroutines to be used as modules #


