package MyPertProb_locPWA_dawg;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyPertProb_locPWA_dawg.pm:
#	This Perl module is a collection of subroutines
#	that perform perturbation calculations of the probability
#	of a given local pairwise sequence alignment (PWA),
#	under Dawg's indel model (Cartwright 2005) with the space- and time-homogeneous indel rates.
#	(Creation STARTED on April 28 (Tue), 2015, ENDED on ...)
#

use strict;
use warnings;

use MyArrayManipu_num qw( diff_nmarray log_sum_exp );
	## union_nmarray

=set aside for later use
use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree

use MyArrayManipu qw( diff_array union_array nr_array);

use MyMalnIO qw( pr_ODEN_input );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	mfac_null_2events
	mfac_asites_vs_dgaps_1event
	mfac_asites_vs_dgaps_2events tld_mu_P_pwa_a tld_mu_P_pwa_b
	mfac_agaps_vs_dsites_1event
	mfac_agaps_vs_dsites_2events tld_mu_P_pwa_c tld_mu_P_pwa_d 
	mfac_asites_nhom_dsites_2events tld_mu_P_pwa_e tld_mu_P_pwa_fg

	mfac_asites_nhom_dsites_3events
	tld_mu_P_pwa_h tld_mu_P_pwa_i1_and_i2 tld_mu_P_pwa_i3
	tld_mu_P_pwa_j1_and_j2 tld_mu_P_pwa_j3p4_and_j5p6 tld_mu_P_pwa_j7p8
	tld_mu_P_pwa_k tld_mu_P_pwa_l1_and_l2 tld_mu_P_pwa_l3
	tld_mu_P_pwa_m1_and_m2 tld_mu_P_pwa_m3p4_and_m5p6 tld_mu_P_pwa_m7p8

	coeff_quad_asites_vs_dgaps
	coeff_quad_agaps_vs_dsites
	coeff_quad_asites_nhom_dsites
	);


	# This specifies the level of tests on the main subroutines,
	# 'xxx', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
##our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
our $TEST_LEVEL = 5;

	# This specifies the level of tests on the primary satellite subroutines,
	# 'cal_lnlk1_indel_scn_cls', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
##our $TEST_LEVEL1 = 1;
##our $TEST_LEVEL1 = 2;
##our $TEST_LEVEL1 = 3;
our $TEST_LEVEL1 = 5;

	### DEFINITIONS of the SUBROUTINES ###


	## (I) Fundamental subroutines that give basic functions necessary for the practical perturbation calculations. ##

# F_e_neg:
#	This basic subroutine calculates the basic function:
#	F_e- (x,t) := \int_{0}^{t} dt1 exp(-x*t)
#		= (1 - exp(-x*t))/x if x,
#		= 0, := t if x == 0.
#	It is defined in Eq.(SR-8B.2) in Supplementary Results 8B.
#	[CREATED on April 28 (Tue), 2015.]
#
# INPUT: ($x, $t), both of which are real numbers.
#
# OUTPUT: $f_en = F_e- (x,t).
#
sub F_e_neg ($$) {

	my ($x, $t) = @_;

	my $f_en = ($x == 0) ? $t : (1 - exp (-$x*$t))/$x ;

	return $f_en;

}	# End of the subroutine 'F_e_neg'. #


# I_3hatM:
#	This basic subroutine calculates the basic function (defined in Eq.(SR-8B.5b) in Supplementary Results):
#	I_{3\hat{M}} (l1,l2, l3; t, lamba)
#	:= \int_{0}^{t}dt1 \int_{t1}^{t}dt2 \int_{t2}^{t}dt3 exp[-lambda*l1*(t-t1) -lambda*l2*(t-t2) - lambda*l3*(t-t3)].
#
#	(Here t := t_F - t_I, lambda := lambda_I + lambda_D. The "lx" is short for "delta_lx.")
#
#	[CREATED on April 28 (Tue), 2015.]
#
# INPUT: ($l1, $l2, $l3, $t, $lambda),
#	all of which are real numbers.
#	And all except $t are non-zero.
#
# OUTPUT: $i_3hm = I_{3\hat{M}} ($l1, $l2, $l3; $t, $lambda),
#	or undef if we encounter an error.
#
sub I_3hatM ($$$$$) {

	my ($l1, $l2, $l3, $t, $lambda) = @_;

	unless (($l1 != 0) and ($l2 != 0) and ($l3 != 0)) {
	    print STDERR "[ERROR 'I_3hatM'] ILLEGAL INPUT: (l1, l2, l3) = ($l1, $l2, $l3): $! !\n\n";
	    return undef;
	}

	unless ($lambda != 0) {
	    print STDERR "[ERROR 'I_3hatM'] ILLEGAL INPUT: lambda = $lambda: $! !\n\n";
	    return undef;
	}

	my $l23 = $l2+$l3;

	if ($l23 == 0) {
	    my $l12 = $l1+$l2;
	    my $i_3hm = ((1-exp(-$lambda*$l1*$t))/($lambda*$l1) - F_e_neg ($lambda*$l12, $t))/($l3*$l2) ;
	    ##my $i_3hm = ((1-exp(-$lambda*$l1*$t))/($lambda*$l1) - F_e_neg ($lambda*$l23, $t))/($l3*$l2) ; # A BUG fixed on May 4 (Mon), 2014. #
	    $i_3hm -=  (1-(1+$lambda*$l1*$t)*exp(-$lambda*$l1*$t)) /($lambda*$l1) /($l3*$l1) ;
	    ##$i_3hm -=  (1-(1+$lambda*$l1*$t))*exp(-$lambda*$l1*$t) /($lambda*$l1) /($l3*$l1) ; # A BUG fixed on May 4 (Mon), 2014. #
	    $i_3hm /= $lambda*$lambda ;

	    return $i_3hm ;
	}

	my $l12 = $l1+$l2;
	my $l123 = $l12 + $l3;

	my $i_3hm = (1-exp(-$lambda*$l1*$t))/($lambda*$l1) /($l2*$l23) ;
	$i_3hm -= F_e_neg ($lambda*$l12, $t) /($l3*$l2) ;
	$i_3hm += F_e_neg ($lambda*$l123, $t) /($l3*$l23) ;
	$i_3hm /= $lambda*$lambda;

	return $i_3hm ;

}	# End of the subroutine 'I_3hatM'. #


# I_hatM_pls_2hatM:
#	This basic subroutine calculates the basic function (defined in Eq.(SR-8B.6b) in Supplementary Results):
#	I_{\hat{M}+2\hat{M}} (l1, {l2, l3}; t, lambda)
#	:= \int_{0}^{t}dt1 [ exp(-lambda*l1*(t-t1)) * \int_{t1}^{t}dt2 exp(-lambda*l2*(t-t2)) * \int_{t1}^{t}dt3 exp(-lambda*l3*(t-t3)) ].
#
#	(Here t := t_F - t_I, lambda := lambda_I + lambda_D. The "lx" is short for "delta_lx.")
#
#	[CREATED on Aprit 29 (Wed), 2015.]
#
# INPUT: ($l1, $l2, $l3, $t, $lambda),
#	all of which are real numbers.
#	And all except $t are non-zero.
#
# OUTPUT: $i_hm_2hm = I_{\hat{M}+2\hat{M}} ($l1, {$l2, $l3}; $t, $lambda),
#	or undef if we encounter an error.
#
sub I_hatM_pls_2hatM ($$$$$) {

	my ($l1, $l2, $l3, $t, $lambda) = @_;

	unless (($l1 != 0) and ($l2 != 0) and ($l3 != 0)) {
	    print STDOUT "[ERROR 'I_hatM_pls_2hatM'] ILLEGAL INPUT: (l1, l2, l3) = ($l1, $l2, $l3): $! !\n\n";
	    return undef;
	}

	unless ($lambda != 0) {
	    print STDOUT "[ERROR 'I_hatM_pls_2hatM'] ILLEGAL INPUT: lambda = $lambda: $! !\n\n";
	    return undef;
	}

	my $l12 = $l1 + $l2;
	my $l13 = $l1 + $l3;
	my $l123 = $l12 + $l3;

	my $i_hm_2hm = (1-exp(-$lambda*$l1*$t))/($lambda*$l1);
	$i_hm_2hm += F_e_neg ($lambda*$l123, $t);
	$i_hm_2hm -= F_e_neg ($lambda*$l12, $t);
	$i_hm_2hm -= F_e_neg ($lambda*$l13, $t);

	$i_hm_2hm /= $lambda*$lambda*$l2*$l3;

	return $i_hm_2hm;

}	# End of the subroutine 'I_hatM_pls_2hatM'. #


# I_2hatM_pls_hatM:
#	This basic subroutine calculates the basic function (defined in Eq.(SR-8B.7b) in Supplementary Results):
#	I_{2\hat{M}+\hat{M}} ({l1, l2}, l3; t, lambda)
#	:= \int_{0}^{t}dt3 [ exp(-lambda*l3*(t-t3)) * \int_{0}^{t3}dt1 exp(-lambda*l1*(t-t1)) * \int_{0}^{t3}dt2 exp(-lambda*l2*(t-t2))] .
#
#	(Here t := t_F - t_I, lambda := lambda_I + lambda_D. The "lx" is short for "delta_lx.")
#	The calculation is based on the identity, Eq.(SR-8B.7b').
#
#	[CREATED on April 29 (Wed), 2015.]
#
# INPUT: ($l1, $l2, $l3, $t, $lambda),
#	all of which are real numbers.
#	And all except $t are non-zero.
#
# OUTPUT: $i_hm_2hm = I_{\hat{M}+2\hat{M}} ($l1, {$l2, $l3}; $t, $lambda),
#	or undef if we encounter an error.
#
sub I_2hatM_pls_hatM ($$$$$) {

	my ($l1, $l2, $l3, $t, $lambda) = @_;

	unless (($l1 != 0) and ($l2 != 0) and ($l3 != 0)) {
	    print STDOUT "[ERROR 'I_2hatM_pls_hatM'] ILLEGAL INPUT: (l1, l2, l3) = ($l1, $l2, $l3): $! !\n\n";
	    return undef;
	}

	unless ($lambda != 0) {
	    print STDOUT "[ERROR 'I_2hatM_pls_hatM'] ILLEGAL INPUT: lambda = $lambda: $! !\n\n";
	    return undef;
	}

	my $i_2hatM_hatM = I_hatM_pls_2hatM (-$l3, -$l2, -$l1, $t, $lambda);

	unless (defined $i_2hatM_hatM) {
	    print STDOUT "[ERROR 'I_2hatM_pls_hatM'] The subroutine, 'I_hatM_pls_2hatM', FAILED: $! !\n\n";
	    return undef;
	}

	$i_2hatM_hatM *= exp (-$lambda*($l1+$l2+$l3)*$t) ;

	return $i_2hatM_hatM ;

}	# End of the subroutine 'I_2hatM_pls_hatM'. #


	## (II) 2nd-grade subroutines that calculate each term of the practical perturbation calculations. ##

# mu_P_3events_gnl:
#	This subroutine calculates the contribution from a general triple-indel history
#	(defined in Eq.(SR-8B.5a) in Supplementary Results):
#	\mu_P (l1, l2, l3; t) := \mu_P [([\hat{M}_1, \hat{M}_2, \hat{M}_3], [t_I, t_F]) | (s^A, t_I) ]|_{Dawg}
#		= r(l1)*r(l2)*r(l3)*I_{3\hat{M}}(l1, l2, l3; t, lambda) .
#
#	(Here t := t_F - t_I, lambda := lambda_I + lambda_D. The "lx" is short for "delta_lx.")
#
#	[CREATED on April 29 (Wed), 2015.] 
#
# INPUT: ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D), where
#	$l1, $l2, $l3, $t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $l1, $l2, $l3, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $mu_p = \mu_P ($l1, $l2, $l3; $t) with the parameters $lambda_I, $lambda_D, \@freq_I and \@freq_D.
#	Or undef if we encounter an error.
#
sub mu_P_3events_gnl ($$$$$$\@\@) {

	my ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, $freq_I, $freq_D) = @_;

	unless (($l1 != 0) and ($l2 != 0) and ($l3 != 0)) {
	    print STDOUT "[ERROR 'mu_P_3events_gnl'] ILLEGAL INPUT: (l1, l2, l3) = ($l1, $l2, $l3): $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'mu_P_3events_gnl'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	my $lambda = $lambda_I+$lambda_D;
	my $mu_P_3events = I_3hatM ($l1, $l2, $l3, $t, $lambda) ;

	unless (defined $mu_P_3events) {
	    print STDOUT "[ERROR 'mu_P_3events_gnl'] The subroutine 'I_3hatM' FAILED: $! !\n\n";
	    return undef;
	}

	$mu_P_3events *= ($l1 > 0) ? $lambda_I * $freq_I->[$l1] : $lambda_D * $freq_D->[-$l1] ;
	$mu_P_3events *= ($l2 > 0) ? $lambda_I * $freq_I->[$l2] : $lambda_D * $freq_D->[-$l2] ; 
	$mu_P_3events *= ($l3 > 0) ? $lambda_I * $freq_I->[$l3] : $lambda_D * $freq_D->[-$l3] ; 

	return $mu_P_3events ;

}	# End of the subroutine 'mu_P_3events_gnl'. #


# mu_P_3events_S23:
#	This subroutine calculates the contribution from a pair of general triple-indel histories
#	(defined in Eq.(SR-8B.6a) in Supplementary Results):
#	\mu_P^{(S23)} (l1, l2, l3; t)
#	:= \mu_P [([\hat{M}_1, \hat{M}_2, \hat{M}_3], [t_I, t_F]) | (s^A, t_I) ]|_{Dawg}
#	   + \mu_P [([\hat{M}_1, \hat{M'}_3, \hat{M'}_2], [t_I, t_F]) | (s^A, t_I) ]|_{Dawg}
#
#		= r(l1)*r(l2)*r(l3)*I_{\hat{M}+2\hat{M}}(l1, {l2, l3}; t, lambda) .
#
#	(Here t := t_F - t_I, lambda := lambda_I + lambda_D. The "lx" is short for "delta_lx.")
#
#	[CREATED on April 29 (Wed), 2015.] 
#
# INPUT: ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D), where
#	$l1, $l2, $l3, $t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $l1, $l2, $l3, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $mu_p = \mu_P^{(S23)} ($l1, $l2, $l3; $t) with the parameters $lambda_I, $lambda_D, \@freq_I and \@freq_D.
#	Or undef if we encounter an error.
#
sub mu_P_3events_S23 ($$$$$$\@\@) {

	my ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, $freq_I, $freq_D) = @_;

	unless (($l1 != 0) and ($l2 != 0) and ($l3 != 0)) {
	    print STDOUT "[ERROR 'mu_P_3events_S23'] ILLEGAL INPUT: (l1, l2, l3) = ($l1, $l2, $l3): $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'mu_P_3events_S23'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	my $lambda = $lambda_I+$lambda_D;
	my $mu_P_3events = I_hatM_pls_2hatM ($l1, $l2, $l3, $t, $lambda) ;

	unless (defined $mu_P_3events) {
	    print STDOUT "[ERROR 'mu_P_3events_S23'] The subroutine 'I_hatM_pls_2hatM' FAILED: $! !\n\n";
	    return undef;
	}

	$mu_P_3events *= ($l1 > 0) ? $lambda_I * $freq_I->[$l1] : $lambda_D * $freq_D->[-$l1] ;
	$mu_P_3events *= ($l2 > 0) ? $lambda_I * $freq_I->[$l2] : $lambda_D * $freq_D->[-$l2] ; 
	$mu_P_3events *= ($l3 > 0) ? $lambda_I * $freq_I->[$l3] : $lambda_D * $freq_D->[-$l3] ; 

	return $mu_P_3events ;

}	# End of the subroutine 'mu_P_3events_S23'. #


# mu_P_3events_S12:
#	This subroutine calculates the contribution from a pair of general triple-indel histories
#	(defined in Eq.(SR-8B.7a) in Supplementary Results):
#	\mu_P^{(S12)} (l1, l2, l3; t)
#	:= \mu_P [([\hat{M}_1, \hat{M}_2, \hat{M}_3], [t_I, t_F]) | (s^A, t_I) ]|_{Dawg}
#	   + \mu_P [([\hat{M'}_2, \hat{M'}_1, \hat{M}_3], [t_I, t_F]) | (s^A, t_I) ]|_{Dawg}
#
#		= r(l1)*r(l2)*r(l3)*I_{\hat{M}+2\hat{M}}(l1, {l2, l3}; t, lambda) .
#
#	(Here t := t_F - t_I, lambda := lambda_I + lambda_D. The "lx" is short for "delta_lx.")
#
#	[CREATED on April 29 (Wed), 2015.] 
#
# INPUT: ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D), where
#	$l1, $l2, $l3, $t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $l1, $l2, $l3, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $mu_p = \mu_P^{(S12)} ($l1, $l2, $l3; $t) with the parameters $lambda_I, $lambda_D, \@freq_I and \@freq_D.
#	Or undef if we encounter an error.
#
sub mu_P_3events_S12 ($$$$$$\@\@) {

	my ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, $freq_I, $freq_D) = @_;

	unless (($l1 != 0) and ($l2 != 0) and ($l3 != 0)) {
	    print STDOUT "[ERROR 'mu_P_3events_S12'] ILLEGAL INPUT: (l1, l2, l3) = ($l1, $l2, $l3): $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'mu_P_3events_S12'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	my $lambda = $lambda_I+$lambda_D;
	my $mu_P_3events = I_2hatM_pls_hatM ($l1, $l2, $l3, $t, $lambda) ;

	unless (defined $mu_P_3events) {
	    print STDOUT "[ERROR 'mu_P_3events_S12'] The subroutine 'I_2hatM_pls_hatM' FAILED: $! !\n\n";
	    return undef;
	}

	$mu_P_3events *= ($l1 > 0) ? $lambda_I * $freq_I->[$l1] : $lambda_D * $freq_D->[-$l1] ;
	$mu_P_3events *= ($l2 > 0) ? $lambda_I * $freq_I->[$l2] : $lambda_D * $freq_D->[-$l2] ; 
	$mu_P_3events *= ($l3 > 0) ? $lambda_I * $freq_I->[$l3] : $lambda_D * $freq_D->[-$l3] ; 

	return $mu_P_3events ;

}	# End of the subroutine 'mu_P_3events_S12'. #



	## (III) Main subroutines that perform the practical perturbation calculations. ##


# mfac_null_2events:
#	This subroutine calculates the portion of the multiplication factor of the region between two adjacent
#	preserved ancestral sites (PASs) contributed by the 2-event (next-parsimonious) local indel histories (given in Eq.(VI-2.2a) of the main Results):
#	   \sum_{l=1}^{min{L_I^{CO}, L_D^{CO}}} \mu_P [([\hat{M}_I(1,l), \hat{M}_D(2, l+1)], [t_I, t_F]) | (s^A, t_I) ].
#
#	It uses the formula, Eq.(VI-2.2b') in Results VI.
#
#	[CREATED on April 29 (Wed), 2015.]
#
# INPUT: ($t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $mfac_2ev, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub mfac_null_2events ($$$\@\@$$) {

	my ($t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'mfac_null_2events'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $mfac_2ev = 0;

	my $cutoff = ($cutoff_I < $cutoff_D) ? $cutoff_I : $cutoff_D ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main for-loop (over $l). #

	    my $mu_P = exp(-$lambda * $l * $t) - 1 + $lambda * $l * $t ;
	    $mu_P *= $freq_I->[$l] * $freq_D->[$l] /$l /$l ;

	    $mfac_2ev += $mu_P ;
	}			# End of the main for-loop (over $l). #

	$mfac_2ev *= $lambda_I * $lambda_D /($lambda*$lambda) ;

	return $mfac_2ev ;

}	# End of the subroutine 'mfac_null_2events'. #


# mfac_asites_vs_dgaps_1event:
#	This subroutine calculates the the portion of the multiplicatiion factor for a run of $n_A (= Delta L^A) ancestral sites aligned with a run of gaps in the descendant sequence, contributed by the 1-event (parsimonious) local indel history.
#	The portion is calculated according to Eq.(VI-2.4') of the main Results.
#
#	[CREATED on April 29 (Wed), 2015.]
#
# INPUT: ($n_A, $t, $lambda_I, $lambda_D, \@freq_D, $cutoff_D), where
#	$n_A (>0) (denoted as '\Delta L^A in the paper) is the number of ancestral sites in the run aligned with the descendant gaps;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_D is the cutoff length for deletions.
#
# OUTPUT: $mfac_1ev, which is the portion we desire.
#	Or undef if we encounter an error.
#
sub mfac_asites_vs_dgaps_1event ($$$$\@$) {

	my ($n_A, $t, $lambda_I, $lambda_D, $freq_D, $cutoff_D) = @_;

	unless ($n_A > 0) {
	    print STDOUT "[ERROR 'mfac_asites_vs_dgaps_1event'] ILLEGAL INPUT: n_A = $n_A: $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'mfac_asites_vs_dgaps_1event'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_A > $cutoff_D) { # This history cannot contribute a probability.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $mfac_1ev = exp($lambda * $n_A * $t) - 1;
	$mfac_1ev *= $lambda_D * $freq_D->[$n_A] /$lambda /$n_A ;

	return $mfac_1ev ;

}	# End of the subroutine 'mfac_asites_vs_dgaps_1event'. #


# tld_mu_P_pwa_a:
#	This satellite subroutine of 'mfac_asites_vs_dgaps_2events'
#	calculates the \tilde{\mu}_P[(a)], which is a sum of contributions from histories consisting of 2 deletions each.
#	The portion is calculated according to Eqs.(VI-2.5b,d') of the main Results.
#
#	[CREATED on April 29 (Wed), 2015.]
#
# INPUT: ($n_A, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites aligned with gaps in the run;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_a, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_a ($$$$\@\@$$) {

	my ($n_A, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless ($n_A > 0) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_a'] ILLEGAL INPUT: n_A = $n_A : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_a'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_A > 2 * $cutoff_D ) { # No history of this type can contribute a probability.
	    return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_a = 0;

	my $cutoff = $n_A -1;
	##my $cutoff = ($cutoff_I < $cutoff_D) ? $cutoff_I : $cutoff_D ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main for-loop (over $l). #

	    my $l2 = $n_A - $l;

	    (($l > $cutoff_D) or ($l2 > $cutoff_D)) and next; # This history cannot contribute a probability.

	    my $mu_P = (exp ($lambda * $n_A * $t) - 1) /$n_A ;
	    $mu_P -= (exp ($lambda * $l * $t) -1) /$l ;
	    $mu_P *= $freq_D->[$l] * $freq_D->[$l2] /$l2 ;

	    $tmu_P_a += ($n_A - $l + 1) * $mu_P ;
	}			# End of the main for-loop (over $l). #

	$tmu_P_a *= $lambda_D * $lambda_D /($lambda*$lambda) ;

	return $tmu_P_a ;

}	# End of the subroutine 'tld_mu_P_pwa_a'. #


# tld_mu_P_pwa_b:
#	This satellite subroutine of 'mfac_asites_vs_dgaps_2events'
#	calculates the \tilde{\mu}_P[(b)], which is a sum of contributions from histories each of which consists of an insertion and a subsequent deletion.
#	The portion is calculated according to Eqs.(VI-2.5c,e') of the main Results.
#
#	[CREATED on April 29 (Wed), 2015.]
#
# INPUT: ($n_A, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites aligned with gaps in the run;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_b, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_b ($$$$\@\@$$) {

	my ($n_A, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless ($n_A > 0) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_b'] ILLEGAL INPUT: n_A = $n_A : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_b'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_A > $cutoff_D) { # This type of histories cannot contribute the probabilities.
	    return undef;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_b = 0;

	my $cutoff = ($cutoff_I < $cutoff_D - $n_A) ? $cutoff_I : $cutoff_D - $n_A ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main for-loop (over $l). #

	    my $l2 = $n_A + $l;

	    my $mu_P = (exp ($lambda * $n_A * $t) - 1) /$n_A ;
	    $mu_P -= (1 - exp (- $lambda * $l * $t)) /$l ;
	    $mu_P *= $freq_I->[$l] * $freq_D->[$l2] /$l2 ;

	    $tmu_P_b += ($n_A + 1) * $mu_P ;
	}			# End of the main for-loop (over $l). #

	$tmu_P_b *= $lambda_I * $lambda_D /($lambda*$lambda) ;

	return $tmu_P_b ;

}	# End of the subroutine 'tld_mu_P_pwa_b'. #


# mfac_asites_vs_dgaps_2events:
#	This subroutine calculates the the portion of the multiplicatiion factor for a run of $n_A (= \Delta L^A) ancestral sites aligned with a run of gaps in the descendant sequence, contributed by the 2-event (next-parsimonious) local indel histories.
#	The portion is calculated according to Eq.(VI-2.5a) of the main Results.
#
#	[CREATED on April 29 (Wed), 2015.]
#
# INPUT: ($n_A, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites aligned with gaps in the run;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: ($mfac_2ev, \@components), where
#	$mfac_2ev is the portion of the multiplcation factor we desire;
#	@components stores the components of $mfac_2ev.
#
#	The subroutine returns undef if it encounters an error.
#
sub mfac_asites_vs_dgaps_2events ($$$$\@\@$$) {

	my ($n_A, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	my $tmu_P_a = tld_mu_P_pwa_a ($n_A, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_a) {
	    print STDOUT "[ERROR 'mfac_asites_vs_dgaps_2events'] The subroutine 'tld_mu_P_pwa_a' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_b = tld_mu_P_pwa_b ($n_A, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_b) {
	    print STDOUT "[ERROR 'mfac_asites_vs_dgaps_2events'] The subroutine 'tld_mu_P_pwa_b' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $mfac_2ev = $tmu_P_a + $tmu_P_b ;
	my @components = ($tmu_P_a, $tmu_P_b);

	return ($mfac_2ev, \@components) ;

}	# End of the subroutine 'mfac_asites_vs_dgaps_2events'. #


# mfac_agaps_vs_dsites_1event:
#	This subroutine calculates the the portion of the multiplicatiion factor for a run of $n_D (= Delta L^D) descendant sites aligned with a run of gaps in the ancestral sequence, contributed by the 1-event (parsimonious) local indel history.
#	The portion is calculated according to Eq.(SR-8A.1') of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_D, $t, $lambda_I, $lambda_D, \@freq_I, $cutoff_I), where
#	$n_D (>0) (denoted as '\Delta L^D in the paper) is the number of descendant sites in the run aligned with the ancestral gaps;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero;
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$cutoff_I is the cutoff length for insertions.
#
# OUTPUT: $mfac_1ev, which is the portion we desire.
#	Or undef if we encounter an error.
#
sub mfac_agaps_vs_dsites_1event ($$$$\@$) {

	my ($n_D, $t, $lambda_I, $lambda_D, $freq_I, $cutoff_I) = @_;

	unless ($n_D > 0) {
	    print STDOUT "[ERROR 'mfac_agaps_vs_dsites_1event'] ILLEGAL INPUT: n_D = $n_D: $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'mfac_agaps_vs_dsites_1event'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_D > $cutoff_I) { # This history cannot contribute to the probability.
	    return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $mfac_1ev = 1- exp( - $lambda * $n_D * $t) ;
	$mfac_1ev *= $lambda_I * $freq_I->[$n_D] /$lambda /$n_D ;

	return $mfac_1ev ;

}	# End of the subroutine 'mfac_agaps_vs_dsites_1event'. #


# tld_mu_P_pwa_c:
#	This satellite subroutine of 'mfac_agaps_vs_dsites_2events'
#	calculates the \tilde{\mu}_P[(c)], which is a sum of contributions from histories each of which consists of two insertions.
#	The portion is calculated according to Eqs.(SR-8A.2b,d') of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites aligned with gaps in the run;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_c, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_c ($$$$\@\@$$) {

	my ($n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless ($n_D > 0) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_c'] ILLEGAL INPUT: n_D = $n_D : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_c'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_D > 2*$cutoff_I) { # Any history of this type cannot contribute a probability.
	    return 0;
	}
	

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_c = 0;

	my $cutoff = $n_D -1;
	##my $cutoff = ($cutoff_D < $cutoff_I - $n_D) ? $cutoff_D : $cutoff_I - $n_D ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main for-loop (over $l). #

	    my $l1 = $n_D - $l;
	    my $l2 = $l;

	    (($l1 > $cutoff_I) or ($l2 > $cutoff_I)) and next; # This history cannot contribute.

	    my $mu_P = (1 - exp (- $lambda * $l1 * $t)) /$l1 ;
	    $mu_P -= (1 - exp (- $lambda * $n_D * $t)) /$n_D ;
	    $mu_P *= $freq_I->[$l1] * $freq_I->[$l2] /$l2 ;

	    $tmu_P_c += ($n_D - $l + 1) * $mu_P ;
	}			# End of the main for-loop (over $l). #

	$tmu_P_c *= $lambda_I * $lambda_I /($lambda*$lambda) ;

	return $tmu_P_c ;

}	# End of the subroutine 'tld_mu_P_pwa_c'. #


# tld_mu_P_pwa_d:
#	This satellite subroutine of 'mfac_agaps_vs_dsites_2events'
#	calculates the \tilde{\mu}_P[(d)], which is a sum of contributions from histories each of which consists of an insertion and a subsequent deletion.
#	The portion is calculated according to Eqs.(SR-8A.2c,e') of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites aligned with gaps in the run;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_d, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_d ($$$$\@\@$$) {

	my ($n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless ($n_D > 0) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_d'] ILLEGAL INPUT: n_D = $n_D : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_d'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_D > $cutoff_I) { # In this case, there can be no histories resulting in the pattern.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_d = 0;

	my $cutoff = ($cutoff_D < $cutoff_I - $n_D) ? $cutoff_D : $cutoff_I - $n_D ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main for-loop (over $l). #

	    my $l1 = $n_D + $l;
	    my $l2 = $l;

	    my $mu_P = (1 - exp (- $lambda * $n_D * $t)) /$n_D ;
	    $mu_P -= (1 - exp (- $lambda * $l1 * $t)) /$l1 ;
	    $mu_P *= $freq_I->[$l1] * $freq_D->[$l2] /$l2 ;

	    $tmu_P_d += ($n_D + 1) * $mu_P ;
	}			# End of the main for-loop (over $l). #

	$tmu_P_d *= $lambda_I * $lambda_D /($lambda*$lambda) ;

	return $tmu_P_d ;

}	# End of the subroutine 'tld_mu_P_pwa_d'. #


# mfac_agaps_vs_dsites_2events:
#	This subroutine calculates the the portion of the multiplicatiion factor for a run of $n_D (= \Delta L^D) descendant sites aligned with a run of gaps in the ancestral sequence, contributed by the 2-event (next-parsimonious) local indel histories.
#	The portion is calculated according to Eq.(SR-8A.2a) of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites aligned with gaps in the run;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residues insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residues deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: ($mfac_2ev, \@components), where
#	$mfac_2ev is the portion of the multiplcation factor we desire;
#	@components stores the compoenents of $mfac_2ev.
#
#	The subroutine returns undef if it encounters an error.
#
sub mfac_agaps_vs_dsites_2events ($$$$\@\@$$) {

	my ($n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	my $tmu_P_c = tld_mu_P_pwa_c ($n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_c) {
	    print STDOUT "[ERROR 'mfac_agaps_vs_dsites_2events'] The subroutine 'tld_mu_P_pwa_c' FAILED: $! !\n\n";
	    return undef;
	}

	my $tmu_P_d = tld_mu_P_pwa_d ($n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_d) {
	    print STDOUT "[ERROR 'mfac_agaps_vs_dsites_2events'] The subroutine 'tld_mu_P_pwa_d' FAILED: $! !\n\n";
	    return undef;
	}

	my $mfac_2ev = $tmu_P_c + $tmu_P_d ;
	my @components = ($tmu_P_c, $tmu_P_d) ;

	return ($mfac_2ev, \@components) ;

}	# End of the subroutine 'mfac_agaps_vs_dsites_2events'. #


# tld_mu_P_pwa_e:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_2events'
#	calculates \tilde{\mu}_P[(e)], which is a sum of contributions from histories each of which consists of a deletion and a subsequent insertion.
#	The portion is calculated according to Eqs.(SR-8B.1b,e') of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $mu_P, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_e ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_fg'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_fg'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > $cutoff_I) or ($n_A > $cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $mu_P1 = (exp($lambda * $n_A * $t) - 1) /($lambda * $n_A) ;

	my $mu_P2 = F_e_neg ($lambda * ($n_D - $n_A), $t) ;
	unless (defined $mu_P2) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_fg'] The subroutine 'F_e_neg' FAILED: $! !\n\n";
	    return undef;
	}

	my $mu_P = $mu_P1 - $mu_P2 ;
	$mu_P *= $lambda_D * $freq_D->[$n_A] * $lambda_I * $freq_I->[$n_D] /($lambda * $n_D) ;

	return $mu_P ;

}	# End of the subroutine 'tld_mu_P_pwa_e'. #


# tld_mu_P_pwa_fg:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_2events'
#	calculates \tilde{\mu}_P[(f)] = \tilde{\mu}_P [(g)], which is a sum of contributions from histories each of which consists of an insertion and a subsequent deletion.
#	The portion is calculated according to Eqs.(SR-8B.1c,d,f',g') of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_fg, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_fg ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_fg'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_fg'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > $cutoff_I) or ($n_A > $cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $mu_P0 = F_e_neg ($lambda * ($n_D - $n_A), $t) ;
	unless (defined $mu_P0) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_fg'] The subroutine 'F_e_neg' FAILED: $! !\n\n";
	    return undef;
	}

	my $tmu_P_fg = 0;

	my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=0; $l<= $cutoff; $l++) { # Main for-loop (over $l). #

	    my $l1 = $n_D + $l;
	    my $l2 = $n_A + $l;

	    my $mu_P = $mu_P0;

	    $mu_P -= (1 - exp (- $lambda * $l1 * $t)) /($lambda * $l1) ;
	    ##$mu_P -= (1 - exp (- $lambda * $l1 * $t)) /$l1 ; # A BUG fixed on May 4 (Mon), 2014. #

	    $mu_P *= $freq_I->[$l1] * $freq_D->[$l2] /$l2 ;

	    $tmu_P_fg += $mu_P ;
	}			# End of the main for-loop (over $l). #

	$tmu_P_fg *= $lambda_I * $lambda_D /$lambda ;
	##$tmu_P_fg *= $lambda_I * $lambda_D /($lambda*$lambda) ; # A BUG fixed on May 4 (Mon), 2014. #

	return $tmu_P_fg ;

}	# End of the subroutine 'tld_mu_P_pwa_fg'. #


# mfac_asites_nhom_dsites_2events:
#	This subroutine calculates the the portion of the multiplicatiion factor for $n_A (= \Delta L^A) ancestral sites and non-homologous $n_D (= \Delta L^D) descendant sites between a pair of PASs, contributed by the 2-event (parsimonious) local indel histories.
#	The portion is calculated according to Eq.(SR-8B.1a) of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residues insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residues deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: ($mfac_2ev, \@components), where
#	$mfac_2ev is the portion of the multiplcation factor we desire;
#	@components stores the components of $mfac_2ev . 
#
#	The subroutine returns undef if it encounters an error.
#
sub mfac_asites_nhom_dsites_2events ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	my $tmu_P_e = tld_mu_P_pwa_e ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_e) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_2events'] The subroutine 'tld_mu_P_pwa_e' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_f = my $tmu_P_g = tld_mu_P_pwa_fg ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_f) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_2events'] The subroutine 'tld_mu_P_pwa_fg' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $mfac_2ev = $tmu_P_e + $tmu_P_f + $tmu_P_g ;
	my @components = ($tmu_P_e, $tmu_P_f, $tmu_P_g) ;

	return ($mfac_2ev, \@components) ;

}	# End of the subroutine 'mfac_asites_nhom_dsites_2events'. #



# tld_mu_P_pwa_h:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(h)], which is a sum of contributions from histories each of which consists of two successive deletions and a subsequent insertion.
#	The portion is calculated based on Eqs.(SR-8B.8a) of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_h, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_h ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_h'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_h'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > $cutoff_I) or ($n_A > 2*$cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_h = 0;

	my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

		my $l1 = - $l;
		my $l2 = -($n_A - $l);
		my $l3 = $n_D;

		(-$l2 > $cutoff_D) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_gnl ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_h'] The subroutine 'mu_P_3events_gnl' FAILED for l = $l : $! !\n\n";
		    return undef;
		}

		$tmu_P_h += ($n_A - $l + 1) * $mu_P ;

	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_h ;

}	# End of the subroutine 'tld_mu_P_pwa_h'. #


# tld_mu_P_pwa_i1_and_i2:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(i-1)] = \tilde{\mu}_P [(i-2)], which is a sum of contributions from histories each of which consists of a deletion, an insertion after that, and a deletion after that.
#	The portion is calculated based on Eqs.(SR-8B.8b,c) of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_i12, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_i1_and_i2 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_i1_and_i2'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_i1_and_i2'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > $cutoff_I) or ($n_A > 2*$cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_i12 = 0;

	my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

	    my $ub = ($cutoff_I - $n_D < $cutoff_D - $n_A + $l) ? $cutoff_I - $n_D : $cutoff_D - $n_A + $l ;

	    for (my $lp=0; $lp <= $ub; $lp++) { # Main inner for-loop (over $lp). #

		my $l1 = - $l;
		my $l2 = $n_D + $lp;
		my $l3 = -($n_A -$l + $lp);

		(($l2 > $cutoff_I) or (-$l3 > $cutoff_D)) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_gnl ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_i1_and_i2'] The subroutine 'mu_P_3events_gnl' FAILED for (l, $lp) = ($l, $lp): $! !\n\n";
		    return undef;
		}

		$tmu_P_i12 += ($n_A - $l + 1) * $mu_P ;

	    }				# End of the main inner for-loop (over $lp). #
	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_i12 ;

}	# End of the subroutine 'tld_mu_P_pwa_i1_and_i2'. #


# tld_mu_P_pwa_i3:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(i-3)], which is a sum of contributions from histories each of which consists of a deletion, a subsequent insertion, and a deletion after that.
#	The portion is calculated based on Eqs.(SR-8B.8d) of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_i3, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_i3 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_i3'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_i3'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > $cutoff_I) or ($n_A > $cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_i3 = 0;

	##my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	my $cutoff = ($cutoff_I - $n_D < $cutoff_D ) ? $cutoff_I - $n_D : $cutoff_D ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

		my $l1 = - $n_A;
		my $l2 = $n_D + $l;
		my $l3 = -$l ;

		##(-$l2 > $cutoff_D) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_gnl ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_i3'] The subroutine 'mu_P_3events_gnl' FAILED for l = $l : $! !\n\n";
		    return undef;
		}

		$tmu_P_i3 += ($n_D + 1) * $mu_P ;

	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_i3 ;

}	# End of the subroutine 'tld_mu_P_pwa_i3'. #


# tld_mu_P_pwa_j1_and_j2:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(j-1)] = \tilde{\mu}_P [(j-2)], which is a sum of contributions from histories each of which consists of an insertion and a subsequent two successive deletions.
#	The portion is calculated based on Eqs.(SR-8B.8e,f) of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_j12, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_j1_and_j2 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_j1_and_j2'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_j1_and_j2'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > $cutoff_I) and ($n_A > 2*$cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_j12 = 0;

	my $cutoff = $cutoff_I - $n_D ;
	##my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=0; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

	    my $lb = (1 > $n_A + $l - $cutoff_D) ? 1 : $n_A + $l - $cutoff_D ;
	    my $ub = ($cutoff_D < $n_A + $l -1) ? $cutoff_D : $n_A + $l -1 ;

	    for (my $lp= $lb; $lp <= $ub; $lp++) { # Main inner for-loop (over $lp). #

		my $l1 = $n_D + $l;
		my $l2 = - $lp;
		my $l3 = -($n_A + $l - $lp);

		##(($l2 > $cutoff_I) or (-$l3 > $cutoff_D)) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_gnl ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_j1_and_j2'] The subroutine 'mu_P_3events_gnl' FAILED for (l, $lp) = ($l, $lp): $! !\n\n";
		    return undef;
		}

		$tmu_P_j12 += ($n_A + $l - $lp + 1) * $mu_P ;

	    }				# End of the main inner for-loop (over $lp). #
	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_j12 ;

}	# End of the subroutine 'tld_mu_P_pwa_j1_and_j2'. #


# tld_mu_P_pwa_j3p4_and_j5p6:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(j-3)] + \tilde{\mu}_P [(j-4)] = \tilde{\mu}_P [(j-5)] + \tilde{\mu}_P [(j-6)],
#	which is a sum of two sums of contributions from histories each of which consists of an insertion and a subsequent two successive deletions.
#	The portion is calculated based on Eqs.(SR-8B.8g,h) of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_j3p4_j5p6, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_j3p4_and_j5p6 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) { # MODIFIED to make the subroutine behave better (on May 4 (Mon), 2015). #
	##unless (($n_A > 0) and ($n_D > 1)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_j3p4_and_j5p6'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_j3p4_and_j5p6'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_D == 1) {  # ADDED to make the subroutine behave better (on May 4 (Mon), 2015). #
	    return 0;
	}

	if (($n_A > $cutoff_D) or ($n_D > $cutoff_I)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_j3p4_j5p6 = 0;

	my $cutoff = $cutoff_I - $n_D ;
	##my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=0; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

	    my $lb = (0 > $l - $cutoff_D) ? 0 : $l - $cutoff_D ;
	    my $ub = ($cutoff_D - $n_A < $l -1) ? $cutoff_D - $n_A : $l -1 ;

	    for (my $lp= $lb; $lp <= $ub; $lp++) { # Main inner for-loop (over $lp). #

		my $l1 = $n_D + $l;
		my $l2 = -($n_A + $lp);
		my $l3 = - ($l - $lp);

		##(($l2 > $cutoff_I) or (-$l3 > $cutoff_D)) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_S23 ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_j3p4_and_j5p6'] The subroutine 'mu_P_3events_S23' FAILED for (l, $lp) = ($l, $lp): $! !\n\n";
		    return undef;
		}

		$tmu_P_j3p4_j5p6 += ($n_D - 1) * $mu_P ;

	    }				# End of the main inner for-loop (over $lp). #
	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_j3p4_j5p6 ;

}	# End of the subroutine 'tld_mu_P_pwa_j3p4_and_j5p6'. #


# tld_mu_P_pwa_j7p8:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(j-7)] + \tilde{\mu}_P [(j-8)],
#	which is a sum of two sums of contributions from histories each of which consists of an insertion and a subsequent two successive deletions.
#	The portion is calculated based on Eqs.(SR-8B.8i) of Supplementary Results.
#
#	[CREATED on May 1 (Fri), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_j7p8, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_j7p8 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_j7p8'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_j7p8'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > $cutoff_I) and ($n_A > 2*$cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_j7p8 = 0;

	my $cutoff = $cutoff_I - $n_D ;
	##my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=0; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

	    my $lb = (1 - $l > $n_A - $cutoff_D) ? 1 - $l : $n_A - $cutoff_D ;
	    my $ub = ($cutoff_D - $l < $n_A -1) ? $cutoff_D - $l : $n_A - 1 ;

	    for (my $lp= $lb; $lp <= $ub; $lp++) { # Main inner for-loop (over $lp). #

		my $l1 = $n_D + $l;
		my $l2 = - ($l + $lp);
		my $l3 = -($n_A - $lp);

		my $lb2 = (0 > $lp) ? 0 : $lp ;
		my $ub2 = ($l+$lp < $n_A ) ? $l+$lp : $n_A ;

		($lb2 > $ub2) and next; # This history cannot contribute a probability.

		##(($l2 > $cutoff_I) or (-$l3 > $cutoff_D)) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_S23 ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_j7p8'] The subroutine 'mu_P_3events_S23' FAILED for (l, $lp) = ($l, $lp): $! !\n\n";
		    return undef;
		}

		$tmu_P_j7p8 += ($ub2 - $lb2 + 1) * $mu_P ;

	    }				# End of the main inner for-loop (over $lp). #
	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_j7p8 ;

}	# End of the subroutine 'tld_mu_P_pwa_j7p8'. #


# tld_mu_P_pwa_k:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(k)], which is a sum of contributions from histories each of which consists of a deletion and subsequent two successive insertions.
#	The portion is calculated based on Eqs.(SR-8B.8j) of Supplementary Results.
#
#	[CREATED on May 1 (Fri), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_k, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_k ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_k'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_k'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > 2*$cutoff_I) or ($n_A > $cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_k = 0;

	my $cutoff = ($n_D -1 < $cutoff_I) ? $n_D -1 : $cutoff_I ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

		my $l1 = - $n_A;
		my $l2 = $l;
		my $l3 = $n_D - $l;

		($l3 > $cutoff_I) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_gnl ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_k'] The subroutine 'mu_P_3events_gnl' FAILED for l = $l : $! !\n\n";
		    return undef;
		}

		$tmu_P_k += ($l + 1) * $mu_P ;

	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_k ;

}	# End of the subroutine 'tld_mu_P_pwa_k'. #


# tld_mu_P_pwa_l1_and_l2:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(l-1)] = \tilde{\mu}_P [(l-2)], which is a sum of contributions from histories each of which consists of an insertion, a deletion after that, and an insertion after that.
#	The portion is calculated based on Eqs.(SR-8B.8k,l) of Supplementary Results.
#
#	[CREATED on May 1 (Fri), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_l12, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_l1_and_l2 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_l1_and_l2'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_l1_and_l2'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > 2*$cutoff_I) or ($n_A > $cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_l12 = 0;

	my $cutoff = $cutoff_D - $n_A;
	##my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=0; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

	    my $lb = (1 > $n_D + $l - $cutoff_I) ? 1 : $n_D + $l - $cutoff_I ;
	    ##my $lb = (1 > $n_D + $l - $cutoff_I) ? 1 : $n_A + $l - $cutoff_I ; # A BUG fixed on May 4 (Mon), 2015. #

	    my $ub = ($n_D - 1 < $cutoff_I) ? $n_D - 1 : $cutoff_I ;

	    for (my $lp= $lb; $lp <= $ub; $lp++) { # Main inner for-loop (over $lp). #

		my $l1 = $n_D + $l - $lp;
		my $l2 = -($n_A + $l);
		my $l3 = $lp;

		##(($l2 > $cutoff_I) or (-$l3 > $cutoff_D)) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_gnl ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_l1_and_l2'] The subroutine 'mu_P_3events_gnl' FAILED for (l, lp) = ($l, $lp): $! !\n\n";
		    return undef;
		}

		$tmu_P_l12 += ($n_D - $lp + 1) * $mu_P ;

	    }				# End of the main inner for-loop (over $lp). #
	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_l12 ;

}	# End of the subroutine 'tld_mu_P_pwa_l1_and_l2'. #


# tld_mu_P_pwa_l3:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(l-3)], which is a sum of contributions from histories each of which consists of an insertion, a subsequent deletion, and an insertion after that.
#	The portion is calculated based on Eqs.(SR-8B.8m) of Supplementary Results.
#
#	[CREATED on May 1 (Fri), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_l3, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_l3 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_l3'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_l3'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_D > $cutoff_I) or ($n_A > $cutoff_D)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_l3 = 0;

	##my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	my $cutoff = ($cutoff_I < $cutoff_D - $n_A) ? $cutoff_I : $cutoff_D - $n_A ;

	for (my $l=1; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

		my $l1 = $l;
		my $l2 = -($n_A + $l);
		my $l3 = $n_D ;

		##(-$l2 > $cutoff_D) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_gnl ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_i3'] The subroutine 'mu_P_3events_gnl' FAILED for l = $l : $! !\n\n";
		    return undef;
		}

		$tmu_P_l3 += ($n_A + 1) * $mu_P ;

	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_l3 ;

}	# End of the subroutine 'tld_mu_P_pwa_l3'. #


# tld_mu_P_pwa_m1_and_m2:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(m-1)] = \tilde{\mu}_P [(m-2)], which is a sum of contributions from histories each of which consists of two successive insertions and a subsequent deletion.
#	The portion is calculated based on Eqs.(SR-8B.8n,o) of Supplementary Results.
#
#	[CREATED on May 1 (Fri), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_m12, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_m1_and_m2 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_m1_and_m2'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_m1_and_m2'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_A > $cutoff_D) or ($n_D > 2*$cutoff_I)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_m12 = 0;

	my $cutoff = $cutoff_D - $n_A ;
	##my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=0; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

	    my $lb = (1 > $n_D + $l - $cutoff_I) ? 1 : $n_D + $l - $cutoff_I ;
	    my $ub = ($cutoff_I < $n_D + $l -1) ? $cutoff_I : $n_D + $l -1 ;

	    for (my $lp= $lb; $lp <= $ub; $lp++) { # Main inner for-loop (over $lp). #

		my $l1 = $n_D + $l - $lp;
		my $l2 = $lp;
		my $l3 = -($n_A + $l);

		##(($l2 > $cutoff_I) or (-$l3 > $cutoff_D)) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_gnl ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_m1_and_m2'] The subroutine 'mu_P_3events_gnl' FAILED for (l, $lp) = ($l, $lp): $! !\n\n";
		    return undef;
		}

		$tmu_P_m12 += ($n_D + $l - $lp + 1) * $mu_P ;

	    }				# End of the main inner for-loop (over $lp). #
	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_m12 ;

}	# End of the subroutine 'tld_mu_P_pwa_m1_and_m2'. #


# tld_mu_P_pwa_m3p4_and_m5p6:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(m-3)] + \tilde{\mu}_P [(m-4)] = \tilde{\mu}_P [(m-5)] + \tilde{\mu}_P [(m-6)],
#	which is a sum of two sums of contributions from histories each of which consists of two successive insertions and a subsequent deletion.
#	The portion is calculated based on Eqs.(SR-8B.8p,q) of Supplementary Results.
#
#	[CREATED on May 1 (Fri), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_m3p4_m5p6, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_m3p4_and_m5p6 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	##unless (($n_A > 0) and ($n_D > 1)) { # A BUG fixed on May 4 (Mon), 2015. #
	    print STDOUT "[ERROR 'tld_mu_P_pwa_m3p4_and_m5p6'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_m3p4_and_m5p6'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_A == 1) { # ADDED to make the subroutine behave better (on May 4 (Mon), 2015).
	    return 0;
	}

	if (($n_A > $cutoff_D) or ($n_D > $cutoff_I)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_m3p4_m5p6 = 0;

	my $cutoff = $cutoff_D - $n_A ;
	##my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=0; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

	    my $lb = (0 > $l - $cutoff_I) ? 0 : $l - $cutoff_I ;
	    my $ub = ($cutoff_I - $n_D < $l -1) ? $cutoff_I - $n_D : $l -1 ;

	    for (my $lp= $lb; $lp <= $ub; $lp++) { # Main inner for-loop (over $lp). #

		my $l1 = $l - $lp;
		my $l2 = $n_D + $lp;
		my $l3 = -($n_A + $l);

		##(($l2 > $cutoff_I) or (-$l3 > $cutoff_D)) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_S12 ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_m3p4_and_m3p4'] The subroutine 'mu_P_3events_S12' FAILED for (l, $lp) = ($l, $lp): $! !\n\n";
		    return undef;
		}

		$tmu_P_m3p4_m5p6 += ($n_A - 1) * $mu_P ;

	    }				# End of the main inner for-loop (over $lp). #
	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_m3p4_m5p6 ;

}	# End of the subroutine 'tld_mu_P_pwa_m3p4_and_m5p6'. #


# tld_mu_P_pwa_m7p8:
#	This satellite subroutine of 'mfac_asites_nhom_dsites_3events'
#	calculates \tilde{\mu}_P[(m-7)] + \tilde{\mu}_P [(m-8)],
#	which is a sum of two sums of contributions from histories each of which consists of two successive insertions and a subsequent deletion.
#	The portion is calculated based on Eqs.(SR-8B.8r) of Supplementary Results.
#
#	[CREATED on May 1 (Fri), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $tmu_P_m7p8, which is the portion of the multiplcation factor we desire.
#	Or undef if we encounter an error.
#
sub tld_mu_P_pwa_m7p8 ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_m7p8'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_pwa_m7p8'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_A > $cutoff_D) and ($n_D > 2*$cutoff_I)) { # In this case, there can be no histories to result in this configuration.
	   return 0;
	}

	my $lambda = $lambda_I+$lambda_D;

	my $tmu_P_m7p8 = 0;

	my $cutoff = $cutoff_D - $n_A ;
	##my $cutoff = ($n_A -1 < $cutoff_D) ? $n_A -1 : $cutoff_D ;
	##my $cutoff = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=0; $l<= $cutoff; $l++) { # Main outer for-loop (over $l). #

	    my $lb = (1 - $l > $n_D - $cutoff_I) ? 1 - $l : $n_D - $cutoff_I ;
	    my $ub = ($cutoff_I - $l < $n_D -1) ? $cutoff_I - $l : $n_D - 1 ;

	    for (my $lp= $lb; $lp <= $ub; $lp++) { # Main inner for-loop (over $lp). #

		my $l1 = $l + $lp;
		my $l2 = $n_D - $lp;
		my $l3 = -($n_A + $l);

		my $lb2 = (0 > $lp) ? 0 : $lp ;
		my $ub2 = ($l+$lp < $n_D ) ? $l+$lp : $n_D ;

		($lb2 > $ub2) and next; # This history cannot contribute a probability.

		##(($l2 > $cutoff_I) or (-$l3 > $cutoff_D)) and next; # This history cannot contribute a probability.

		my $mu_P = mu_P_3events_S12 ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D});
		unless (defined $mu_P) {
		    print STDOUT "[ERROR 'tld_mu_P_pwa_m7p8'] The subroutine 'mu_P_3events_S12' FAILED for (l, $lp) = ($l, $lp): $! !\n\n";
		    return undef;
		}

		$tmu_P_m7p8 += ($ub2 - $lb2 + 1) * $mu_P ;

	    }				# End of the main inner for-loop (over $lp). #
	}			# End of the main outer for-loop (over $l). #


	return $tmu_P_m7p8 ;

}	# End of the subroutine 'tld_mu_P_pwa_m7p8'. #


##sub mu_P_3events_gnl ($$$$$$\@\@) {
##sub mu_P_3events_S23 ($$$$$$\@\@) {
##sub mu_P_3events_S12 ($$$$$$\@\@) {
##	my ($l1, $l2, $l3, $t, $lambda_I, $lambda_D, $freq_I, $freq_D) = @_;


	### RESTART FROM HERE !! ###


# mfac_asites_nhom_dsites_3events:
#	This subroutine calculates the the portion of the multiplicatiion factor for $n_A (= \Delta L^A) ancestral sites and non-homologous $n_D (= \Delta L^D) descendant sites between a pair of PASs, contributed by the 3-event (next-parsimonious) local indel histories.
#	The portion is calculated according to Eq.(SR-8C.3) of Supplementary Results.
#
#	[CREATED on April 30 (Thr), 2015.]
#
# INPUT: ($n_A, $n_D, $t, $lambda_I, $lambda_D, \@freq_I, \@freq_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites between PASs;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites between PASs;
#	$t, $lambda_I and $lambda_D are real numbers.
#	Out of them, $lambda_I and $lambda_D are non-zero.
#	$freq_I[$l] ($l>0) is the relative frequency of $l-residues insertions;
#	$freq_D[$l] ($l>0) is the relative frequency of $l-residues deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: ($mfac_3ev, \@components), where
#	$mfac_3ev which is the portion of the multiplcation factor we desire;
#	@components stores the components of $mfac_3ev.
#
#	The subroutine returns undef if it encounters an error.
#
sub mfac_asites_nhom_dsites_3events ($$$$$\@\@$$) {

	my ($n_A, $n_D, $t, $lambda_I, $lambda_D, $freq_I, $freq_D, $cutoff_I, $cutoff_D) = @_;


	my $tmu_P_h = tld_mu_P_pwa_h ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_h) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_h' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_i1 = my $tmu_P_i2 = tld_mu_P_pwa_i1_and_i2 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_i1) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_i1_and_i2' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_i3 = tld_mu_P_pwa_i3 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_i3) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_i3' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_j1 = my $tmu_P_j2 = tld_mu_P_pwa_j1_and_j2 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_j1) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_j1_and_j2' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_j3p4 = my $tmu_P_j5p6 = tld_mu_P_pwa_j3p4_and_j5p6 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_j3p4) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_j3p4_and_j5p6' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_j7p8 = tld_mu_P_pwa_j7p8 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_j7p8) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_j7p8' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_k = tld_mu_P_pwa_k ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_k) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_k' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_l1 = my $tmu_P_l2 = tld_mu_P_pwa_l1_and_l2 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_l1) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_l1_and_l2' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_l3 = tld_mu_P_pwa_l3 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_l3) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_l3' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_m1 = my $tmu_P_m2 = tld_mu_P_pwa_m1_and_m2 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_m1) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_m1_and_m2' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_m3p4 = my $tmu_P_m5p6 = tld_mu_P_pwa_m3p4_and_m5p6 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_m3p4) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_m3p4_and_m5p6' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_m7p8 = tld_mu_P_pwa_m7p8 ($n_A, $n_D, $t, $lambda_I, $lambda_D, @{$freq_I}, @{$freq_D}, $cutoff_I, $cutoff_D);

	unless (defined $tmu_P_m7p8) {
	    print STDOUT "[ERROR 'mfac_asites_nhom_dsites_3events'] The subroutine 'tld_mu_P_pwa_m7p8' FAILED: $! !\n\n";
	    return (undef, undef);
	}


	my $mfac_3ev = $tmu_P_h + $tmu_P_i1 + $tmu_P_i2 + $tmu_P_i3
		+ $tmu_P_j1 + $tmu_P_j2 + $tmu_P_j3p4 + $tmu_P_j5p6 + $tmu_P_j7p8
		+ $tmu_P_k + $tmu_P_l1 + $tmu_P_l2 + $tmu_P_l3
		+ $tmu_P_m1 + $tmu_P_m2 + $tmu_P_m3p4 + $tmu_P_m5p6 + $tmu_P_m7p8 ;

	my @components = ($tmu_P_h, $tmu_P_i1, $tmu_P_i2, $tmu_P_i3,
		$tmu_P_j1, $tmu_P_j2, $tmu_P_j3p4, $tmu_P_j5p6, $tmu_P_j7p8,
		$tmu_P_k, $tmu_P_l1, $tmu_P_l2, $tmu_P_l3,
		$tmu_P_m1, $tmu_P_m2, $tmu_P_m3p4, $tmu_P_m5p6, $tmu_P_m7p8 ) ;

	return ($mfac_3ev, \@components);

}	# End of the subroutine 'mfac_asites_nhom_dsites_3events'. #


	## (IV) Subroutines to calculate the terms in the expansion of the probabilities in {total indel rate} * {time lapse}.

# coeff_quad_asites_vs_dgaps:
#	This subroutine calculates the coefficient of the quadratic term in {total indel rate} * {time lapse}
#	of the probability (normalized by that of case (i)) of a local PWA of case (ii),
#	that is, the case where a run of ancestral sites are aligned with a run of gaps in the descendant sequeunce.
#
#	The coefficient is calculated according to Eq.(VII-2.6b') of the main Results.
#
#	[CREATED on May 12 (Tue), 2015.]
#
# INPUT: ($n_A, $lambda_I, $lambda_D, \@freqs_I, \@freqs_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites aligned with gaps in the run;
#	$lambda_I (total insertion rate) and $lambda_D (total deletion rate) are non-zero real numbers;
#	$freqs_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freqs_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $coeff_quad, which is the coefficient we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub coeff_quad_asites_vs_dgaps ($$$\@\@$$) {

	my ($n_A, $lambda_I, $lambda_D, $freqs_I, $freqs_D, $cutoff_I, $cutoff_D) = @_;

	unless ($n_A > 0) {
	    print STDOUT "[ERROR 'coeff_quad_asites_vs_dgaps'] ILLEGAL INPUT: n_A = $n_A : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'coeff_quad_asites_vs_dgaps'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_A > 2 * $cutoff_D ) { # No history of this type can contribute a probability.
	    return 0;
	}

	my $lambda = $lambda_I+$lambda_D;


		# Contributions from the histories of type 'deletion -> deletion'. #

	my $sum_prod1 = 0;

	my $lb1 = (1 > $n_A - $cutoff_D) ? 1 : $n_A - $cutoff_D ;
	my $ub1 = ($n_A -1 < $cutoff_D) ? $n_A - 1 : $cutoff_D ;

	for (my $l=$lb1; $l<= $ub1; $l++) { # 1st main for-loop (over $l). #

	    my $l2 = $n_A - $l;

	    ## (($l > $cutoff_D) or ($l2 > $cutoff_D)) and next; # This history cannot contribute a probability.

	    my $prod = $freqs_D->[$l] * $freqs_D->[$l2] ;

	    $sum_prod1 += ($n_A - $l + 1) * $prod ;

	}			# End of the 1st main for-loop (over $l). #


		# Contributions from the histories of type 'insertion -> deletion'. #

	my $sum_prod2 = 0;

	my $lb2 = 1 ;
	my $ub2 = ($cutoff_I < $cutoff_D - $n_A) ? $cutoff_I : $cutoff_D - $n_A ;

	for (my $l=$lb2; $l <= $ub2; $l++) { # 2nd main for-loop (over $l). #

	    my $l2 = $n_A + $l; 

	    my $prod = $freqs_I->[$l] * $freqs_D->[$l2];

	    $sum_prod2 += $prod ;

	}			# End of the 2nd main for-loop (over $l). #

	$sum_prod2 *= ($n_A+1);


		# Putting a finishing touch. #

	my $coeff_quad = - $n_A * $freqs_D->[$n_A] + $lambda_D * $sum_prod1 /$lambda  + $lambda_I * $sum_prod2 /$lambda;

	return $coeff_quad ;

}	# End of the subroutine 'coeff_quad_asites_vs_dgaps'. #


# coeff_quad_agaps_vs_dsites:
#	This subroutine calculates the coefficient of the quadratic term in {total indel rate} * {time lapse}
#	of the probability (normalized by that of case (i)) of a local PWA of case (iii),
#	that is, the case where a run of descendant sites are aligned with a run of gaps in the ancestral sequeunce.
#
#	The coefficient is calculated according to Eq.(VII-2.8b) of the main Results.
#
#	[CREATED on May 12 (Tue), 2015.]
#
# INPUT: ($n_D, $lambda_I, $lambda_D, \@freqs_I, \@freqs_D, $cutoff_I, $cutoff_D), where
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites aligned with gaps in the run;
#	$lambda_I (total insertion rate) and $lambda_D (total deletion rate) are non-zero real numbers;
#	$freqs_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freqs_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $coeff_quad, which is the coefficient we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub coeff_quad_agaps_vs_dsites ($$$\@\@$$) {

	my ($n_D, $lambda_I, $lambda_D, $freqs_I, $freqs_D, $cutoff_I, $cutoff_D) = @_;

	unless ($n_D > 0) {
	    print STDOUT "[ERROR 'coeff_quad_agaps_vs_dsites'] ILLEGAL INPUT: n_D = $n_D : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'coeff_quad_agaps_vs_dsites'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if ($n_D > 2 * $cutoff_I ) { # No history of this type can contribute a probability.
	    return 0;
	}

	my $lambda = $lambda_I+$lambda_D;


		# Contributions from the histories of type 'insertion -> insertion'. #

	my $sum_prod1 = 0;

	my $lb1 = (1 > $n_D - $cutoff_I) ? 1 : $n_D - $cutoff_I ;
	my $ub1 = ($n_D -1 < $cutoff_I) ? $n_D - 1 : $cutoff_I ;

	for (my $l=$lb1; $l<= $ub1; $l++) { # 1st main for-loop (over $l). #

	    my $l2 = $n_D - $l;

	    ## (($l > $cutoff_I) or ($l2 > $cutoff_I)) and next; # This history cannot contribute a probability.

	    my $prod = $freqs_I->[$l2] * $freqs_I->[$l] ;

	    $sum_prod1 += ($n_D - $l + 1) * $prod ;

	}			# End of the 1st main for-loop (over $l). #


		# Contributions from the histories of type 'insertion -> deletion'. #

	my $sum_prod2 = 0;

	my $lb2 = 1 ;
	my $ub2 = ($cutoff_D < $cutoff_I - $n_D) ? $cutoff_D : $cutoff_I - $n_D ;

	for (my $l=$lb2; $l <= $ub2; $l++) { # 2nd main for-loop (over $l). #

	    my $l2 = $n_D + $l; 

	    my $prod = $freqs_I->[$l2] * $freqs_D->[$l];

	    $sum_prod2 += $prod ;

	}			# End of the 2nd main for-loop (over $l). #

	$sum_prod2 *= ($n_D + 1);


		# Putting a finishing touch. #

	my $coeff_quad = - $n_D * $freqs_I->[$n_D] + $lambda_I * $sum_prod1 /$lambda  + $lambda_D * $sum_prod2 /$lambda;

	return $coeff_quad ;

}	# End of the subroutine 'coeff_quad_agaps_vs_dsites'. #


# coeff_quad_asites_nhom_dsites:
#	This subroutine calculates the coefficient of the quadratic term,  {total insertion rate} * {total deletion rate} * {time lapse}^2,
#	of the probability (normalized by that of case (i)) of a local PWA of case (iv),
#	that is, the case where a run of ancestral sites and a run of descendant sites,
#	which are NOT homologous to each other, are sandwiched by a pair of PASs.
#
#	The coefficient is calculated according to Eq.(VII-2.12) of the main Results.
#
#	[CREATED on May 12 (Tue), 2015.]
#
# INPUT: ($n_A, $n_D, $lambda_I, $lambda_D, \@freqs_I, \@freqs_D, $cutoff_I, $cutoff_D), where
#	$n_A (>0) (denoted as "\Delta L^A" in the paper) is the number of ancestral sites aligned with descdendant gaps in the run;
#	$n_D (>0) (denoted as "\Delta L^D" in the paper) is the number of descendant sites aligned with ancestral gaps in the run;
#	$lambda_I (total insertion rate) and $lambda_D (total deletion rate) are non-zero real numbers;
#	$freqs_I[$l] ($l>0) is the relative frequency of $l-residue insertions;
#	$freqs_D[$l] ($l>0) is the relative frequency of $l-residue deletions;
#	$cutoff_I and $cutoff_D are the cut-off lengths for insertions and deletions, respectively.
#
# OUTPUT: $coeff_quad, which is the coefficient we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub coeff_quad_asites_nhom_dsites ($$$$\@\@$$) {

	my ($n_A, $n_D, $lambda_I, $lambda_D, $freqs_I, $freqs_D, $cutoff_I, $cutoff_D) = @_;

	unless (($n_A > 0) and ($n_D > 0)) {
	    print STDOUT "[ERROR 'coeff_quad_asites_nhom_dsites'] ILLEGAL INPUT: (n_A, n_D) = ($n_A, $n_D) : $! !\n\n";
	    return undef;
	}

	unless (($lambda_I != 0) and ($lambda_D != 0)) {
	    print STDOUT "[ERROR 'coeff_quad_asites_nhom_dsites'] ILLEGAL INPUT: (lambda_I, lambda_D) = ($lambda_I, $lambda_D): $! !\n\n";
	    return undef;
	}

	if (($n_A > $cutoff_D) or ($n_D > $cutoff_I )) { # No history of this type can contribute a probability.
	    return 0;
	}

	##my $lambda = $lambda_I+$lambda_D;

	my $coeff_quad = 1/2 * $freqs_D->[$n_A] * $freqs_I->[$n_D] ;

	my $ub = ($cutoff_I - $n_D < $cutoff_D - $n_A) ? $cutoff_I - $n_D : $cutoff_D - $n_A ;

	for (my $l=0; $l <= $ub; $l++) { # The main for-loop (over $l). #

	    my $l1 = $n_D + $l;
	    my $l2 = $n_A + $l;

	    my $prod = $freqs_I->[$l1] * $freqs_D->[$l2];

	    $coeff_quad += $prod ;

	}			# End of the main for-loop (over $l). #

	return $coeff_quad ;

}	# End of the subroutine 'coeff_quad_asites_nhom_dsites'. #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


