package MyPhyl_Partitions;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyPhyl_Partitions.pm:
#	This module contains subroutines that perform various
#	partition manipulations related to phylogenetic reconstruction.
#
#	This module mainly uses the representation of a partition
#	by  a (#{sequences}-1)-bit binary string
#	with '0' and '1' on the $k th bit ($k starting with 0)
#	representing that the $k th sequence has the same/different state
#	as/than the (#{sequences}-1)-th sequence.
#
#	Very frequently, a binary string is represented
#	with a numerical array, whose element each represents
#	a piece of the string with a specified unit length (usually 32 bits).
#

use strict;
use warnings;

use MyBin_Oper qw(
	bitwise_or_numarrays bitwise_and_numarrays
	bitwise_xor_numarrays
	select_by_bin_value count_binary_states
	sbset2numrep_binstring
	make_monobit_pattern match_bin_arrays
	);
use MyArrayManipu qw(
	binsearch_array
	);
	# diff_array
 	## union_array nr_array compare_arrays
	## union_array

use MyArrayManipu_num qw( rescale_confining_vector );

use MyLinearAlgebra qw(
	solve_lin_eq_lu
	);

use MySpecFunc qw( betai );
use MyInverseFunc qw( inv_betai_nwt inv_betai_lvr2 );

use MyTaxonomy qw( fetch_ancestry );
use MyTreeManipu_II qw();
	## br_draw_subtree
use MyTreeManipu_III qw( make_node2abs_depth );

=set aside for potential use
use MyFile_manipu qw(
	open_input_file
	);
use MyProbDistri qw(
	cum_gaussian
	);
use MyTaxonomy qw(
	fetch_ext_offsprings_nr
	fetch_common_ancestors
	);
	## fetch_ancestry
	## fetch_last_common_ancestor fetch_last_common_ancestor_ov2
	## fetch_all_offsprings
use MyTreeMap qw(
	fractionate_by_ndepth
	);
	## assign_gnode2sp_lca
	## fetch_splca_comps assign_gnode2sp_lca_comp
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	check_partition_compatibility
	bch_check_partition_compatibility
	search_partial_complements_in
	search_partial_complements_out
	cal_trio_cladewise_prob
	trio_cladewise_prob_nr
	upper_bound_eff_count
	partition_hierarchy
	partition_hierarchy2tree
	make_branch2partition
	) ;

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;

sub trio_cladewise_prob_nr ($$$$$\@);
sub upper_bound_eff_count ($$$$$$$);
sub fetch_solitary_node (\@$$\@\%);
sub child_nodes_of_partition (\@\@\@$$$\@\%\%);


# check_partition_compatibility:
#	This subroutine checks if the two partitions are compatible or not.
#	If the partitions are compatible,
#	it also checks whether they equal each other and,
#	if not, which includes the other.
#
#	The subroutine regards '1' as 'present'.
#	The subroutine REQUIRES that the two input partitions 
#	are represented by the binary strings of the same size.
#
# INPUT: (\@numrep1, \@numrep2, \@numrep_null), where
#	@numrep1/2 is the numerical array representation of the 1st/2nd binary string;
#	@numrel_null is the 'null partition' (binary string consisting only of '0's)
#	in the same numerical array representation;
#
# OUTPUT: ($fl_cmpt, $sgn_incl), where
#	$fl_compt = 1 if the two partitions are compatible,
#		  = 0 if they are incompatible;
#	$sgn_incl = -1 if @numrep1 is included in @numrep2,
#		 = +1 if @numrep1 includes @numrep2,
#		 = 0 if @numrep1 equals @numrep2,
#		 = undef if the two partitions are
#			either non-overlapping ($fl_compt = 1)
#			or incompatible ($fl_compt = 0).
#
#	The subroutine returns (undef, undef) if the process failed.
#
sub check_partition_compatibility (\@\@\@) {

	my ($numrep1, $numrep2, $numrep_null) = @_;

	my $if_match = match_bin_arrays (@{$numrep1}, @{$numrep2});
	unless (defined $if_match) {
	    print STDERR "[ERROR 'check_partition_compatibility'] The subroutine 'match_bin_arrays' FAILED on the arrays:\n";
	     print STDERR '@{$numrep1} = (', join (', ', @{$numrep1}),
		')  and  @{$numrep2} = (', join (', ', @{$numrep2}), "): $! !\n\n";
	    return (undef, undef);
	}

	($if_match) and (return (1, 0)); # The two partitions equal each other. #

	my $numrep_and = bitwise_and_numarrays (@{$numrep1}, @{$numrep2});
	unless (defined $numrep_and) {
	    print STDERR "[ERROR 'check_partition_compatibility'] The subroutine 'bitwise_and_numarrays' FAILED on the arrays:\n";
	     print STDERR '@{$numrep1} = (', join (', ', @{$numrep1}),
		')  and  @{$numrep2} = (', join (', ', @{$numrep2}), "): $! !\n\n";
	    return (undef, undef);
	}

	my $if_match0 = match_bin_arrays (@{$numrep_null}, @{$numrep_and});
	unless (defined $if_match0) {
	    print STDERR "[ERROR 'check_partition_compatibility'] The subroutine 'match_bin_arrays' FAILED on the arrays:\n";
	     print STDERR '@{$numrep_null} = (', join (', ', @{$numrep_null}),
		')  and  @{$numrep_and} = (', join (', ', @{$numrep_and}), "): $! !\n\n";
	}
	($if_match0) and (return (1, undef)); # @{$numrep1} does NOT overlap with @{$numrep2}.

	my $if_match1 = match_bin_arrays (@{$numrep1}, @{$numrep_and});
	unless (defined $if_match1) {
	    print STDERR "[ERROR 'check_partition_compatibility'] The subroutine 'match_bin_arrays' FAILED on the arrays:\n";
	     print STDERR '@{$numrep1} = (', join (', ', @{$numrep1}),
		')  and  @{$numrep_and} = (', join (', ', @{$numrep_and}), "): $! !\n\n";
	    return (undef, undef);
	}
	($if_match1) and (return (1, -1)); # @{$numrep1} is included in @{$numrep2}.

	my $if_match2 = match_bin_arrays (@{$numrep2}, @{$numrep_and});
	unless (defined $if_match1) {
	    print STDERR "[ERROR 'check_partition_compatibility'] The subroutine 'match_bin_arrays' FAILED on the arrays:\n";
	     print STDERR '@{$numrep2} = (', join (', ', @{$numrep2}),
		')  and  @{$numrep_and} = (', join (', ', @{$numrep_and}), "): $! !\n\n";
	    return (undef, undef);
	}
	($if_match2) and (return (1, 1)); # @{$numrep2} is included in @{$numrep1}.

	return (0, undef); # The two partitions are incompatible with each other.

}	# End of the subroutine 'check_partition_compatibility' #


# bch_check_partition_compatibility:
#	This subroutine checks the compatibility of a bunch of partitions,
#	by performing 'check_partition_compatibility' many times.
#
# INPUT: (\@set_partitions, $indx_numrep, $size_bin, $denom, \@indices_rlv), where
#	@{$set_partitions[$k]->[$indx_numrep]} when $indx_numrep is defined,
#	or @{$set_partitions[$k]} otherwise,
#	gives the numerical representation of the $k th partition;
#	$size_bin is the size of the binary strings that represent the partitions;
#	$denom is the unit length of pieces into which binary strings are divided;
#	@indices_rlv lists only indices of relevant entries;
#
# OUTPUT: (\@non_ovlp, \@smaller, \@larger, \@identical, \@incompatible), where 
#	@{$non_ovlp[$k]} lists the indices of the partitions that do NOT overlap with the $indices_rlv[$k] th partition;
#	@{$smaller[$k]} lists the indices of the partitions included in the $indices_rlv[$k] th partition (excluding the identical ones);
#	@{$larger[$k]} lists the indices of the partitions including the $indices_rlv[$k] th partition (excluding the identical ones);
#	@{$identical[$k]} lists the indices of the partitions identical to the $indices_rlv[$k] th partition;
#	@{$incompatible[$k]} lists the indices of the partitions incompatible with the $indices_rlv[$k] th partition.
#
#	The subroutine returns (undef, undef, undef, undef) if the process failed.
#
sub bch_check_partition_compatibility (\@$$$\@) {

	my ($set_partitions, $indx, $size_bin, $denom, $indices_rlv) = @_;

	my $ct_ptits = @{$set_partitions};
	my $ct_rlv = @{$indices_rlv};
	unless ($ct_ptits >= $ct_rlv) {
	    print STDERR "[ERROR 'bch_check_partition_compatibility'] ILLEGAL INPUT: The number of relevant entries ($ct_rlv) EXCEEDS the number of partitions ($ct_ptits): $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
	($ct_rlv>0) or (return ([], [], [], [], []));

		# Create a null partition (corresponding to  monomorphic sites) #
	my $numrep_null = make_monobit_pattern (0, $size_bin, $denom);
	unless (defined $numrep_null) {
	    print STDERR "[ERROR 'bch_check_partition_compatibility'] 'make_monobit_pattern' FAILED to create a numerical array representing $size_bin '0's in the unit of $denom bits: $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

		# Initialize the output #
	my @non_ovlp = my @smaller = my @larger = my @identical = my @incompatible = ();
	foreach my $set (\@non_ovlp, \@smaller, \@larger, \@identical, \@incompatible) {
	     for (1 .. $ct_rlv) { push @{$set}, []; }
	}

		# Check compatibility and inclusion between the partitions #

	for (my $i=0; $i<$ct_rlv-1; $i++) { # Outer for-loop #
	    my $indx1 = $indices_rlv->[$i];
	    my $partition1 = $set_partitions->[$indx1];
	    my $numrep1 = (defined $indx) ? $partition1->[$indx] : $partition1;

	    for (my $j=$i+1; $j<$ct_rlv; $j++) { # Inner for-loop # 
		my $indx2 = $indices_rlv->[$j];
		my $partition2 = $set_partitions->[$indx2];
		my $numrep2 = (defined $indx) ? $partition2->[$indx] : $partition2;

		my ($fl_cmpt, $sgn_incl) = check_partition_compatibility (@{$numrep1}, @{$numrep2}, @{$numrep_null});
		unless (defined $fl_cmpt) {
		    print STDERR "[ERROR 'bch_check_partition_compatibility'] The subroutine 'check_partition_compatibility' FAILED on the $i th and $j th partitions: $! !\n\n";
		    return (undef, undef, undef, undef);
		}
		if ($fl_cmpt) {
		    if (!(defined $sgn_incl)) {
			# The $i th partition does NOT overlap with the $j th one.
			push @{$non_ovlp[$i]}, $indx2;
			push @{$non_ovlp[$j]}, $indx1;

		    } elsif ($sgn_incl < 0) {
			# The $i th partition is included in the $j th one.
			push @{$larger[$i]}, $indx2;
			push @{$smaller[$j]}, $indx1;

		    } elsif ($sgn_incl > 0) {
			# The $j th partition is included in the $i th one.
			push @{$larger[$j]}, $indx1;
			push @{$smaller[$i]}, $indx2;

		    } else {
			# The $i th partition equals to the $j th one.
			push @{$identical[$i]}, $indx2;
			push @{$identical[$j]}, $indx1;
		    }

		} else {
			# The $i th and $j th partitions are incompatible. #
		    push @{$incompatible[$i]}, $indx2;
		    push @{$incompatible[$j]}, $indx1;
		}
	    }				# End of the inner for-loop #

	}		# End of the outer for-loop #


		# Finish the output #

	foreach my $set (\@non_ovlp, \@smaller, \@larger, \@identical, \@incompatible) { # Outer foreach-loop #
	    for (my $i=0; $i<$ct_rlv; $i++) { # Inner for-loop #
		my @std_entries = sort {$a <=> $b} @{$set->[$i]};
		$set->[$i] = \@std_entries;
	    }			# End of the inner for-loop #
	}				# End of the outer foreach-loop # 

	return (\@non_ovlp, \@smaller, \@larger, \@identical, \@incompatible) ;

}	# End of the subroutine 'bch_check_partition_compatibility' #


# search_partial_complements_in:
#	This subroutine looks for the partial complement (A - B) of each partition (B)
#	that are included in the reference partition (A).
#	The searches are conducted for the references listed in @indices_rlv.
#
#	[Algorithm]
#	(A - B) is given by a bitwise operation A ^ B.
#	So, all we have to do is to conduct a binary search for A ^ B in the set of partitions inclluded in the reference partition A.
#
#	REQIREMENTS:
#	To efficiently conduct binary searchs,
#	the subroutine require @set_partitions and @indices_rlv
#	already sorted in ascending numerical order.
#
# INPUT: (\@smaller, \@set_partitions, $indx_numrep, $size_bin, $denom, \@indices_rlv), where
#	@{$smaller[$k]} lists the indices of the partitions included in the $indices_rlv[$k] th partition (excluding the identical ones);
#	@{$set_partitions[$k]->[$indx_numrep]} when $indx_numrep is defined,
#	or @{$set_partitions[$k]} otherwise,
#	gives the numerical representation of the $k th partition;
#	$size_bin is the size of the binary strings that represent the partitions;
#	$denom is the unit length of pieces into which binary strings are divided;
#	@indices_rlv lists only indices of relevant entries;
#
# OUTPUT: \@part_cmpl_smaller, where
#	$part_cmpl_smaller[$k]->[$l] is the index of the partial complement of $smaller[$k]->[$l] within the $indices_rlv[$k] th partition.
#
#	The subroutine returns undef if the process failed.
#
sub search_partial_complements_in (\@\@$$$\@) {

	my ($smaller, $set_partitions, $indx_numrep, $size_bin, $denom, $indices_rlv) = @_;

		## Handle exceptional cases ##
	my $ct_ptits = @{$set_partitions};
	my $ct_rlv = @{$indices_rlv};
	unless ($ct_ptits >= $ct_rlv) {
	    print STDERR "[ERROR 'search_partial_complements_in'] ILLEGAL INPUT: The number of relevant entries ($ct_rlv) EXCEEDS the number of partitions ($ct_ptits): $! !\n";
	    return undef;
	}
	unless ($ct_rlv == @{$smaller}) {
	    print STDERR "[ERROR 'search_partial_complements_in'] ILLEGAL INPUT: The size (", scalar (@{$smaller}), ") of the input array \@\{\$smaller\} CLASHES with the number ($ct_rlv) of relevant partitions: $! !\n";
	    return undef;
	}
	($ct_rlv>0) or (return []);

		## Prepare arrays necessary for binary searches ##
	my $first_partition = $set_partitions->[0];
	my $first_numrep = (defined $indx_numrep) ? $first_partition->[$indx_numrep] : $first_partition;
	my $size_numrep = @{$first_numrep};
	my @indices_sch = (0 .. $size_numrep -1);
	my @methods_sch = ();
	for (1 .. $size_numrep) { push @methods_sch, '+'; }

	my @part_cmpl_smaller = ();

		## Main Process: Search for partial complements ##

	for (my $k=0; $k<$ct_rlv; $k++) { # Outer for-loop #
	    my $indx_ref = $indices_rlv->[$k];
	    my $partition_ref = $set_partitions->[$indx_ref];
	    my $numrep_ref = (defined $indx_numrep) ? $partition_ref->[$indx_numrep] : $partition_ref;

	    my $indices_smaller = $smaller->[$k];
	    my $ct_smaller = @{$indices_smaller};
	    my @indices_cmpl = ();
	    my $ct_potential = $ct_smaller - 1;
	    for (my $j=0; $j<$ct_smaller; $j++) { # Inner for-loop #
		($ct_potential > 0) or last; # We ran out of potential complements. 
		(defined $indices_cmpl[$j]) and next; # Skip if the complement is already known.

		my $indx_sml = $indices_smaller->[$j];
		my $partition_sml = $set_partitions->[$indx_sml];
		my $numrep_sml = (defined $indx_numrep) ? $partition_sml->[$indx_numrep] : $partition_sml;
		my $numrep_cmpl = bitwise_xor_numarrays (@{$numrep_ref}, @{$numrep_sml});
		unless (defined $numrep_cmpl) {
		    print STDERR "[ERROR 'search_partial_complements_in'] The subroutine 'bitwise_xor_numarrays' FAILED for the $j th partition included in the $indx_ref th partition : $! !\n";
		    return undef;
		}
		my ($fl_exist, $indx2_cmpl, $indx_cmpl) = binsearch_array (@{$numrep_cmpl}, @{$set_partitions}, $indx_numrep, @{$indices_smaller}, @indices_sch, @methods_sch);
		unless (defined $fl_exist) {
		    print STDERR "[ERROR 'search_partial_complements_in'] The subroutine 'binsearch_arrays' FAILED for the $j th partition included in the $indx_ref th partition: $! !\n";
		    return undef;
		}

		if ($fl_exist) {
		    $indices_cmpl[$j] = $indx_cmpl;
		    $indices_cmpl[$indx2_cmpl] = $indx_sml;
		    $ct_potential -= 2;
		}

	    }			# End of the inner for-loop #

	    push  @part_cmpl_smaller, \@indices_cmpl;

	}		# End of the outer for-loop #

	return \@part_cmpl_smaller;

}	# End of the subroutine 'search_partial_complements_in' #


# search_partial_complement_out:
#	This subroutine looks for the partial complement (C - A) of each partition (C)
#	that includes the reference partition (A),
#	as well as for the partial complement (D + A) of each partition (D)
#	that does NOT overlap with the reference partition (A).
#	The searches are conducted for the references listed in @indices_rlv.
#
#	[Algorithm]
#	(C - A) is given by C ^ A, and (D + A) is given by D | A.
#	All we have to do is to conduct a binary search for C ^ A in the set of partitions NOT overlapping with A,
#	as well as a binary search for D | A in the set of partitions that include A.
#
#	REQIREMENTS:
#	To efficiently conduct binary searchs,
#	the subroutine require @set_partitions and @indices_rlv
#	already sorted in ascending numerical order.
#
# INPUT: (\@non_ovlp, \@larger, \@set_partitions, $indx_numrep, $size_bin, $denom, \@indices_rlv), where
#	@{$non_ovlp[$k]} lists the indices of the partitions that do NOT overlap with the $indices_rlv[$k] th partition;
#	@{$larger[$k]} lists the indices of the partitions including the $indices_rlv[$k] th partition (excluding the identical ones);
#	@{$set_partitions[$k]->[$indx_numrep]} when $indx_numrep is defined,
#	or @{$set_partitions[$k]} otherwise,
#	gives the numerical representation of the $k th partition;
#	$size_bin is the size of the binary strings that represent the partitions;
#	$denom is the unit length of pieces into which binary strings are divided;
#	@indices_rlv lists only indices of relevant entries;
#
# OUTPUT:  (\@part_cmpl_novlp, \@part_cmpl_larger), where
#	$part_cmpl_novlp[$k]->[$l] is the index of the partial complement of $non_ovlp[$k]->[$l] within (the complement of) the $indices_rlv[$k] th partition;
#	$part_cmpl_larger[$k]->[$l] is the index of the partial complement of $larger[$k]->[$l] within (the complement of) the $indices_rlv[$k] th partition. 
#
#	The subroutine returns (undef, undef) if the process failed.
#
sub search_partial_complements_out (\@\@\@$$$\@) {

	my ($non_ovlp, $larger, $set_partitions, $indx_numrep, $size_bin, $denom, $indices_rlv) = @_;

	my $ct_ptits = @{$set_partitions};
	my $ct_rlv = @{$indices_rlv};
	unless ($ct_ptits >= $ct_rlv) {
	    print STDERR "[ERROR 'search_partial_complements_out'] ILLEGAL INPUT: The number of relevant entries ($ct_rlv) EXCEEDS the number of partitions ($ct_ptits): $! !\n";
	    return (undef, undef);
	}
	unless ($ct_rlv == @{$non_ovlp}) {
	    print STDERR "[ERROR 'search_partial_complements_out'] ILLEGAL INPUT: The size (", scalar (@{$non_ovlp}), ") of the input array \@\{\$smaller\} CLASHES with the number ($ct_rlv) of relevant partitions: $! !\n";
	    return undef;
	}
	unless ($ct_rlv == @{$larger}) {
	    print STDERR "[ERROR 'search_partial_complements_out'] ILLEGAL INPUT: The size (", scalar (@{$larger}), ") of the input array \@\{\$smaller\} CLASHES with the number ($ct_rlv) of relevant partitions: $! !\n";
	    return undef;
	}
	($ct_rlv>0) or (return ([], []));

		## Prepare arrays necessary for binary searches ##
	my $first_partition = $set_partitions->[0];
	my $first_numrep = (defined $indx_numrep) ? $first_partition->[$indx_numrep] : $first_partition;
	my $size_numrep = @{$first_numrep};
	my @indices_sch = (0 .. $size_numrep -1);
	my @methods_sch = ();
	for (1 .. $size_numrep) { push @methods_sch, '+'; }


		## Main Process: Search for Partial Complements ##

	my @part_cmpl_novlp = my @part_cmpl_larger = ();

	for (my $k = 0; $k<$ct_rlv; $k++) { # Outer for-loop # 

	    my $indx_ref = $indices_rlv->[$k];
	    my $partition_ref = $set_partitions->[$indx_ref];
	    my $numrep_ref = (defined $indx_numrep) ? $partition_ref->[$indx_numrep] : $partition_ref;

	    my $indices_larger = $larger->[$k];
	    my $ct_larger = @{$indices_larger};
	    my $indices_novlp = $non_ovlp->[$k];
	    my $ct_novlp = @{$indices_novlp};
	    my @indices_cmpl_novlp = my @indices_cmpl_larger = ();
	    my $ct_potential = $ct_novlp;
	    for (my $j=0; $j<$ct_larger; $j++) { # Inner for-loop #
		($ct_potential > 0) or last; # We ran out of potential complements. 

		my $indx_lrg = $indices_larger->[$j];
		my $partition_lrg = $set_partitions->[$indx_lrg];
		my $numrep_lrg = (defined $indx_numrep) ? $partition_lrg->[$indx_numrep] : $partition_lrg;
		my $numrep_cmpl = bitwise_xor_numarrays (@{$numrep_ref}, @{$numrep_lrg});
		unless (defined $numrep_cmpl) {
		    print STDERR "[ERROR 'search_partial_complements_out'] The subroutine 'bitwise_xor_numarrays' FAILED for the $j th partition that include the $indx_ref th partition : $! !\n";
		    return undef;
		}
		my ($fl_exist, $indx2_cmpl, $indx_cmpl) = binsearch_array (@{$numrep_cmpl}, @{$set_partitions}, $indx_numrep, @{$indices_novlp}, @indices_sch, @methods_sch);
		unless (defined $fl_exist) {
		    print STDERR "[ERROR 'search_partial_complements_out'] The subroutine 'binsearch_arrays' FAILED for the $j th partition that include the $indx_ref th partition: $! !\n";
		    return undef;
		}

		if ($fl_exist) {
		    $indices_cmpl_larger[$j] = $indx_cmpl;
		    $indices_cmpl_novlp[$indx2_cmpl] = $indx_lrg;
		    $ct_potential--;
		}

	    }			# End of the inner for-loop #

	    push  @part_cmpl_larger, \@indices_cmpl_larger;
	    push @part_cmpl_novlp, \@indices_cmpl_novlp;

	}		# End of the outer for-loop #

	return (\@part_cmpl_novlp, \@part_cmpl_larger);

}	# End of the subroutine 'search_partial_complements_out'


# cal_trio_cladewise_prob:
#	This subroutine calculates the probability Lambda (C) that the members in a clade (C)
#	have the same character that is different from the character of their LCA.
#	(More precisely, the ratio of such a probability to the probability Theta (C) that all the members have the same character as their LCA.)
#	The subroutine uses a Newton-Raphson algorithm to solve the
#	ML-like equations for the trios of clades, each of which consists of
#	3 non-overlapping clades that combine to form the entire set of OTUs.
#
#	REVISED on April 23 (Fri), 2010:
#	Now the subroutine correctly deals with exterior nodes that are NOT expected by any means
#	to have a pair of partially complementary partitions (that would combine to give the partition in question).
#
# INPUT: ($totct_effsites, $ct_mono_sites, \@set_partitions, $indx_sites, \@indices_rlv, \@list_compatible, \@list_part_cmpl, \%exterior, \@parameters), where
#	@{$set_partitions->[$j][$indx_sites]} if $indx_sites is  defined,
#	@{$set_partitions->[$j]} otherwise, is the list of positions of sites that show the $j th partition;
#	@indices_rlv lists the indices of the partitions relevant for this session;
#	$list_compatible[$k]->[$l] is the index of the $l th partition that is compatible with the $indices_rlv[$k] th partition in @set_partitions;  
#	$list_part_cmpl[$k]->[$l] is the index of the partition that is the partial complement of $list_compatlble[$k]->[$l] regarding the $indices_rlv[$k] th partition.
#	It is undef if there is no such partition in the MSA;
#	%exterior = ($indx => 1, ...) for the partitions that are exterior and 
#	are NOT by any means expected to have a pair of partially complementary partitions;
#	@parameters = ($thrsh_prob, $ct_stat, $eps1, $eps2, $max_ite, $eps_betai, $maxite_betai)
#	is a set of parameters necessary for the calculation.
#	Especially, $ct_stat gives the (effective) number of possible states.
#	$eps1 gives the accuracy of the numerical solution we want.
#	$eps2 is the threshold value of the difference in F_i (R_1, R_2, R_3)
#	under which the iteration should stop.
#	$max_ite is the maximum number of iteration steps you specify;
#	$eps_betai is the resolution of both $x and $y required for the inverse incomplete beta function;
#	$maxite_betai is the maximum number of iteration steps for the inverse incomplete beta function;
#
# OUTPUT: \@cladewise_prob, where
#	@{$cladewise_prob[$k]} = @list_tri_clwise_prb lists the information on
#	the cladewise probabilities on all the existing pairs of partial complementar partitions, as well as the most frequent lone partition, both regarding the $indices_rlv[$k] th partition;
#	@{$list_tri_clwise_prb[$i]} = ($indx_ptit1, $indx_ptit2 (undef for a lone partition), $effct_sb, $effct_ptit1, $effct_ptit2, $ratio_sb, $ratio_ptit1, $ratio_ptit2, $prb_siml, $del1, $del2, $ct_ite)
#	for the pair giving the $i th largest $prb_siml, which is the probability that
#	the simultaneous changes in 2 clades result in the $indx_rlv[$k] th partition.
#	$indx_ptit1/2 is the index of the partition 1/2, which is partially complement with each other regarding the subject partition (the $indx_rlv[$k] th partiton);
#	$effct_sb/ptit1/ptit2 is the 'effective count' of the subject partition/partition1/partition2;
#	$ratio_sb/ptit1/ptit2 is the ratio Lambda (C) / Theda (C) for the subject partition/partition1/partition2.
#	$prb_siml is the probability of simultaneous changes in the two clades giving partitions 1 and 2.
#	$del1 is the absolute magnitude of the increment,
#	$del2 is the absolute magnitude of the equation functions,
#	$ct_ite is the number of iterations.
#
#	REVISED on April 23 (Fri), 2010:
#	@{$cladewise_prob[$k]} = ([undef, undef, $effct_sb, 0, 0, $effct_sb/$ct_mono_sites/($ct_stat-1), 0, 0, 0, 0, 0, 0]) if the $indx_rlv[$k] th partition is the exterior partition theoretically incapable of having a pair of partially complementary partitions.
#
#	The subroutine returns undef if it failed.
#
sub cal_trio_cladewise_prob ($$\@$\@\@\@\%\@) {

		## REVISED on April 23 (Fri), 2010 ##
	my ($totct_effsites, $ct_mono_sites, $set_ptit, $indx_sites,
	$indices_rlv, $list_cmpt, $list_pt_cmpl, $exterior, $parameters) = @_;
	##my ($totct_effsites, $ct_mono_sites, $set_ptit, $indx_sites,
	##$indices_rlv, $list_cmpt, $list_pt_cmpl, $parameters) = @_;

	unless (($totct_effsites>0) and ($ct_mono_sites>0) and ($totct_effsites>$ct_mono_sites)) {
	    print STDERR "[ERROR 'cal_trio_cladewise_prob'] ILLEGAL INPUT: (\$totct_effsites, \$ct_mono_sites) = ($totct_effsites, $ct_mono_sites) : $! !\n";
	    return undef;
	}

	my $ct_rlv = @{$indices_rlv};

	unless ($ct_rlv <= @{$set_ptit}) {
	    print STDERR "[ERROR 'cal_trio_cladewise_prob'] ILLEGAL INPUT: The number of relevant partitions ($ct_rlv) EXCEEDS that of partitions (", scalar (@{$set_ptit}), "): $! !\n";
	    return undef;
	}
	unless ($ct_rlv == @{$list_cmpt}) {
	    print STDERR "[ERROR 'cal_trio_cladewise_prob'] ILLEGAL INPUT: The size of the list of compatible  partitions (", scalar (@{$list_cmpt}), ") is NOT EQUAL TO the number of the relevant partitions ($ct_rlv): $! !\n";
	    return undef;
	}
	unless ($ct_rlv == @{$list_pt_cmpl}) {
	    print STDERR "[ERROR 'cal_trio_cladewise_prob'] ILLEGAL INPUT: The size of the list of partially complementary  partitions (", scalar (@{$list_pt_cmpl}), ") is NOT EQUAL TO the number of the relevant partitions ($ct_rlv): $! !\n";
	    return undef;
	}

	my ($thrsh_prob, $ct_stat, $eps1, $eps2, $max_ite, $eps_betai, $maxite_betai) = @{$parameters};
	my $prb_mono = $ct_mono_sites/$totct_effsites;

	my ($p_lw, $p_up) = (0, 1 - 1/$ct_stat);

		## REVISED on Apr 15 (Thr), 2010 ##
	my ($effct_0, $error_x, $error_y, $itect) = upper_bound_eff_count ($totct_effsites, 0, $thrsh_prob, $p_lw, $p_up, $eps_betai, $maxite_betai);
	## upper_bound_eff_count ($totct_sites, $ct_obs, $tgt_prb, $p_lw, $p_up, $resol, $maxite) 

	my @cladewise_prob = (); 
	for (my $k=0; $k<$ct_rlv; $k++) { # Outer for-loop #
	    my $indx_sb = $indices_rlv->[$k];
	    my $ptit_sb = $set_ptit->[$indx_sb];
	    my $sites_sb = (defined $indx_sites) ? $ptit_sb->[$indx_sites] : $ptit_sb;
	    my $ct_sites_sb = @{$sites_sb};

	    ##my ($effct_sb, $error_x1, $error_y1, $itect1) = upper_bound_eff_count ($totct_effsites, $ct_sites_sb, $thrsh_prob, $p_lw, $p_up, $eps_betai, $maxite_betai); ## Pending a revision ???
	    my $effct_sb = $ct_sites_sb; ## Subject for future revision ?? ##

		    ## The subject canNOT by any means have
		    ## a pair of partially complementary partitions.
		    ## (ADDED on April 23 (Fri), 2010)
	    if (defined $exterior->{$indx_sb}) {
		push @cladewise_prob, [[undef, undef, $effct_sb, 0, 0, $effct_sb/$ct_mono_sites/($ct_stat -1), 0, 0, 0, 0, 0, 0]];
		next;
	    }

	    my $indices_cmpt = $list_cmpt->[$k];
	    my $ct_cmpt = @{$indices_cmpt};

	    my @largest_lone = (); # element = [$indx_cmpt, undef, $effct_sb, $effct_cmpt, $effct_0], where $effct_0 is the effective count when the MSA has no such site.

	    unless ($ct_cmpt>0) {
		   ## REVISED on Apr 15 (Thr), 2010 ##
		my @curr_trio = (undef, undef, $effct_sb, $effct_0, $effct_0);
		push @largest_lone, \@curr_trio;
		##push @cladewise_prob, [];
		##next;
	    }
	    my $indices_pt_cmpl = $list_pt_cmpl->[$k];

			# Pair up the compatible partitions #
	    my %already; # = ($indx => 1, ...) for the entries already appeared
	    my @pairs = (); # element = [$indx_cmpt, $indx_pt_cmpl, $effct_sb, $effct_cmpt, $effct_pt_cmpl, $ratio_sb, $ratio_cmpt, $ratio_pt_cmpl]

	    for (my $j=0; $j<$ct_cmpt; $j++) { # The 1st inner for-loop #

		my $indx_cmpt = $indices_cmpt->[$j];
		(defined $already{$indx_cmpt}) and next; # Skip the pair already examined
		$already{$indx_cmpt} = 1;

		my $ptit_cmpt = $set_ptit->[$indx_cmpt];
		my $sites_cmpt = (defined $indx_sites) ?
			$ptit_cmpt->[$indx_sites] : $ptit_cmpt;
		my $ct_sites_cmpt = @{$sites_cmpt};

		    ## REVISED on Apr 15 (Thr), 2010 ##
	        my ($effct_cmpt, $error_x1, $error_y1, $itect1) = upper_bound_eff_count ($totct_effsites, $ct_sites_cmpt, $thrsh_prob, $p_lw, $p_up, $eps_betai, $maxite_betai);
		unless (defined $effct_cmpt) {
		    print STDERR "[CAUTION 'cal_trio_cladewise_prob'] 'upper_bound_eff_count' FAILED on the $j th partition compatible with the $k th partition: $! !\n";
		}
		((defined $effct_cmpt) and ($effct_cmpt>$ct_sites_cmpt)) or ($effct_cmpt = $ct_sites_cmpt);
		##my $effct_cmpt = $ct_sites_cmpt; ## Subject for future revision ?? ##

		my $indx_pt_cmpl = $indices_pt_cmpl->[$j];

		if (defined $indx_pt_cmpl) {
			# MSA has the partially complementary partition #

		    $already{$indx_pt_cmpl} = 1;
		    my $ptit_pt_cmpl = $set_ptit->[$indx_pt_cmpl];
		    my $sites_pt_cmpl = (defined $indx_sites) ?
			$ptit_pt_cmpl->[$indx_sites] : $ptit_pt_cmpl;
		    my $ct_sites_pt_cmpl = @{$sites_pt_cmpl};

			## REVISED on Apr 15 (Thr), 2010 ##
		    my ($effct_pt_cmpl, $error_x1, $error_y1, $itect1) = upper_bound_eff_count ($totct_effsites, $ct_sites_pt_cmpl, $thrsh_prob, $p_lw, $p_up, $eps_betai, $maxite_betai);
		    unless (defined $effct_pt_cmpl) {
			print STDERR "[CAUTION 'cal_trio_cladewise_prob'] 'upper_bound_eff_count' FAILED on the partial complement of the $j th partition compatible with the $k th partition: $! !\n";
		    }
		    ((defined $effct_pt_cmpl) and ($effct_pt_cmpl>$ct_sites_pt_cmpl)) or ($effct_pt_cmpl = $ct_sites_pt_cmpl);
		    ##my $effct_pt_cmpl = $ct_sites_pt_cmpl; ## Subject for future revision ?? ##
		    my ($ratio_sb, $ratio_cmpt, $ratio_pt_cmpl, $delta1, $delta2, $ct_ite)
		     = trio_cladewise_prob_nr ($totct_effsites, $ct_mono_sites, $effct_sb, $effct_cmpt, $effct_pt_cmpl, @{$parameters});
		    unless (defined $ratio_sb) {
			print STDERR "[ERROR 'cal_trio_cladewise_prob'] 'trio_cladewise_prob_nr' FAILED on the $j th partition trio ($indx_sb, $indx_cmpt, $indx_pt_cmpl), which is compatible with the $k th partition : $! !\n";
			return undef;
		    }
		    my $prb_simul = ($ct_stat - 1)*$prb_mono * $ratio_cmpt * $ratio_pt_cmpl;
		    push @pairs, [$indx_cmpt, $indx_pt_cmpl, $effct_sb, $effct_cmpt, $effct_pt_cmpl, $ratio_sb, $ratio_cmpt, $ratio_pt_cmpl, $prb_simul, $delta1, $delta2, $ct_ite];
		    next;
		}

		my @curr_trio = ($indx_cmpt, undef, $effct_sb, $effct_cmpt, $effct_0);
		unless (@largest_lone > 0) {
		    push @largest_lone, \@curr_trio;
		    next;
		}
		my $largest_effct_cmpt = $largest_lone[0]->[3];
		if ($effct_cmpt <  $largest_effct_cmpt) {
		   next; # Just discard the current one
		} elsif ($effct_cmpt > $largest_effct_cmpt) {
		    @largest_lone = (\@curr_trio); # Update the list of largest lone partitions #
		} else {
		    push @largest_lone, \@curr_trio; # Add the current one to the list of largest lone partitions #
		}
	    }				# End of the 1st inner for-loop #

	    if (@largest_lone>0) {
		my ($effct_cmpt, $effct_pt_cmpl) = @{$largest_lone[0]}[3, 4];
		my ($ratio_sb, $ratio_cmpt, $ratio_pt_cmpl, $delta1, $delta2, $ct_ite) = 
		  trio_cladewise_prob_nr ($totct_effsites, $ct_mono_sites, $effct_sb, $effct_cmpt, $effct_pt_cmpl, @{$parameters});
		unless (defined $ratio_sb) {
		    print  STDERR "[ERROR 'cal_trio_cladewise_prob'] The subroutine 'trio_cladewise_prob_nr' FAILED on the largest 'lone' partitions compatible with the $k th partition ($indx_sb): $! !\n";
		    return undef;
		}
		my $prb_simul = ($ct_stat - 1) * $prb_mono * $ratio_cmpt * $ratio_pt_cmpl;
		foreach my $element (@largest_lone) { # 2nd inner for-loop #
		    @{$element}[5..11] = ($ratio_sb, $ratio_cmpt, $ratio_pt_cmpl, $prb_simul, $delta1, $delta2, $ct_ite);
		    push @pairs, $element;
		}			# End of the 2nd inner for-loop #
	    }

	    my @std_pairs = sort { $b->[8] <=> $a->[8] } @pairs;
	    push @cladewise_prob, \@std_pairs;
	    
	}		# End of the outer for-loop #

	return \@cladewise_prob;

}	# End of the subroutine 'cal_trio_cladewise_prob' #


# trio_cladewise_prob_nr:
#	This subroutine inputs information on a trio of partitions
#	represented by 3 non-overlapping clades C_1, C_2, and C_3
#	that are combined to form the entire set of OTUs.
#	Then, using the Neweton-Raphson method,
#	it numerically solves 3 simultaneous non-linear equations
#	for the ratios R_i := Lambda (C_i) / Theta (C_i) for the 3 clades C_i (i=1,2,3).
#	Here Lambda (C) is the probability that all members of the clade C have the same state
#	that is different from the state of their LCA,
#	and Theta (C) is the probability that all members of the clade C have the same state as their LCA.
#
#	EQUATIONS:
#	R_1 + R_2 * R_3 + (ct_stat - 2) * R_1 * R_2 * R_3 
#		= L_1 / ((ct_stat - 1) *L_0 ) * { 1 + (ct_stat -1)* R_1 * R_2 * R_3 },
#	and its two cyclic couterparts,
#	where R_i (i=1,2,3) are the ratio as explained above,
#	ct_stat is the (effective) number of possible states,
#	L_i (i=1,2,3) are the likelihood of observing the partition i,
#	and L_0 is the likelihood of observing a mono-character site.
#	L_i (or L_0) could be represented as effct_i / totct_sites,
#	where effct_i is the effective count of the partition i,
#	and totct_sites is the total number of available sites.
#
#	These equations can be rewritten as:
#	0 = F_1 (R_1, R_2, R_3) := R_1 + R_2*R_3 + (ct_stat - 2)*R_1*R_2*R_3
#	     - effct_1/((ct_stat-1)*effct_0) * {1 + (ct_stat -1)* R_1*R_2*R_3}
#	and its two cyclic counterparts.
#
# INPUT: ($totct_sites, $effct_mono, $effct_ptit1, $effct_ptit2, $effct_ptit3, \@parameters), where
#	$effct_mono is the effective number of mono-character sites in the MSA;
#	$effct_ptit1/2/3 is the 'effective number' of the partition 1/2/3 in the MSA.
#	(The 'effective number' may be either the actual number or the upper-limit or
#	any other kind of number that represents the frequency of the partition);
#	@parameters = ($thrsh_prob, $ct_stat, $eps1, $eps2, $max_ite)
#	is a set of parameters necessary for the calculation.
#	Especially, $ct_stat gives the (effective) number of possible states.
#	$eps1 gives the accuracy of the numerical solution we want.
#	$eps2 is the threshold for the value of the finction  F_i (R_1, R_2, R_3)
#	under which the iteration should stop.
#	$max_ite is the maximum number of iteration steps you specify.
#
# OUTPUT: ($ratio_1, $ratio_2, $ratio_3, $delta1, $delta2, $ite), where
#	$ratio_i = Lambda (C_i) / Theta (C_i) as explained above;
#	$delta1 is the sum of the absolute differences in the solution ($ratio_1/2/3)
#	between the current and the last steps;
#	$delta2 is the sum of the absolute values of the equation functions (F_1/2/3)
#	in the current step;
#	$ite is the number of iterations.
#
sub trio_cladewise_prob_nr ($$$$$\@) {

	my ($totct_sites, $effct_mono, $effct_1, $effct_2, $effct_3, $parameters) = @_;
	my ($thrsh_prob, $ct_stat, $eps1, $eps2, $max_ite) = @{$parameters};

	unless (($totct_sites>0) and ($effct_mono>0) and ($totct_sites > $effct_mono)) {
	    print STDERR "[ERROR 'trio_cladewise_prob_nr'] ILLEGAL INPUT: (\$totct_sites, \$ct_mono_sites) = ($totct_sites, $effct_mono): $! !\n";
	    return (undef, undef, undef, undef, undef, undef); 
	}
	unless ($ct_stat>1) {
	    print STDERR "[ERROR 'trio_cladewise_prob_nr'] ILLEGAL INPUT: \$ct_stat = $ct_stat <= 1: $! !\n";
	    return (undef, undef, undef, undef, undef, undef); 
	}

	my $prb_mono = $effct_mono / $totct_sites;

		# Boundaries for the probabilities #
	my @lower = (0,0,0); my @upper = (1,1,1);

		## Initialize the unknown variables
		## and pre-calculate some coefficients 
	my $r1 = my $j1 = $effct_1/$effct_mono/($ct_stat -1);
	my $r2 = my $j2 = $effct_2/$effct_mono/($ct_stat -1);
	my $r3 = my $j3 = $effct_3/$effct_mono/($ct_stat -1);

	my $s1 = $ct_stat -2 - $effct_1/$effct_mono; # = $ct_stat -2 - $j1 *($ct_stat -1)
	my $s2 = $ct_stat -2 - $effct_2/$effct_mono; # = $ct_stat -2 - $j2 *($ct_stat -1)
	my $s3 = $ct_stat -2 - $effct_3/$effct_mono; # = $ct_stat -2 - $j3 *($ct_stat -1)

	my ($delta1, $delta2);

	my $ite = 0;
	while ($ite <= $max_ite) { # The main while-loop #
	    my $f1 = $r1 + $r2 *$r3 + $s1 *$r1 *$r2 *$r3 - $j1;
	    my $f2 = $r2 + $r3 *$r1 + $s2 *$r1 *$r2 *$r3 - $j2;
	    my $f3 = $r3 + $r1 *$r2 + $s3 *$r1 *$r2 *$r3 - $j3;

	    $delta2 = 0;
	    foreach my $f ($f1, $f2, $f3) { $delta2 += ($f > 0) ?  $f : -$f; }
	    if ($delta2 < $eps2) {
		return ($r1, $r2, $r3, $delta1, $delta2, $ite +1/2);
	    }
	 
	    $ite++;

		# Make coefficient matrices
	    my @grad = ();
	    $grad[0] = [1 + $s1 *$r2 *$r3, $r3 + $s1 *$r1 *$r3, $r2 + $s1 *$r1 *$r2];
	    $grad[1] = [$r3 + $s2 *$r2 *$r3, 1 + $s2 *$r1 *$r3, $r1 + $s2 *$r1 *$r2];
	    $grad[2] = [$r2 + $s3 *$r2 *$r3, $r1 + $s3 *$r1 *$r3, 1 + $s3 *$r1 *$r2];
	    my @intcep = (-$f1, -$f2, -$f3);

		# Solve for the increments in the unknown variables #
	    my $delta_r = solve_lin_eq_lu (@grad, @intcep);
	    unless (defined $delta_r) {
		print STDERR "[ERROR 'trio_cladewise_prob_nr'] The subroutine 'solve_lin_eq_lu' FAILED at the $ite th step : $! !\n";
		return (undef, undef, undef, undef, undef, undef); 
	    }

		# If necessary,  rescale the increment vector
		# in order to confine the probabilities within [0, 1].
	    my @base = ($prb_mono * $r1, $prb_mono * $r2, $prb_mono * $r3);
	    my @vec = ();
	    foreach my $x (@{$delta_r}) { push @vec, $prb_mono * $x; }
	    my $scale = rescale_confining_vector (@vec, @base, @lower, @upper);
	    unless (defined $scale) {
		print STDERR "[ERROR 'trio_cladewise_prob_nr'] The subroutine 'rescale_confining_vector' FAILED at the $ite th step: $! !\n";
		return (undef, undef, undef, undef, undef, undef); 
	    }
	    unless ($scale == 1) {
		for (my $i=0; $i<3; $i++) { $delta_r->[$i] *= $scale; } 
	    }

		# Update the unkown variables #
	    $r1 += $delta_r->[0];
	    $r2 += $delta_r->[1];
	    $r3 += $delta_r->[2];

	    $delta1 = 0;
	    foreach my $x (@{$delta_r}) { $delta1 += ($x > 0) ? $x : - $x; }
	    if ($delta1 < $eps1) {
		   # Re-evaluate the total magnitude of the equation functions #
	        my $f1 = $r1 + $r2 *$r3 + $s1 *$r1 *$r2 *$r3 - $j1;
	        my $f2 = $r2 + $r3 *$r1 + $s2 *$r1 *$r2 *$r3 - $j2;
	        my $f3 = $r3 + $r1 *$r2 + $s3 *$r1 *$r2 *$r3 - $j3;
	        $delta2 = 0;
	        foreach my $f ($f1, $f2, $f3) { $delta2 += ($f > 0) ?  $f : -$f; }

		return ($r1, $r2, $r3, $delta1, $delta2, $ite);
	    }

	}		# End of the main while-loop #

	print STDERR "[CAUTION 'trio_cladewise_prob_nr'] The iteration reached the maximum number $max_ite, whereas the numerical solution has not converged yet: $! !\n";

	return ($r1, $r2, $r3, $delta1, $delta2, $ite);

}	# End of the subroutine 'trio_cladewise_prob_nr' #


# upper_bound_eff_count:
#	This is a satellite subroutine of the subroutine 'cal_trio_cladewise_prob'.
#	It calculates, under a specified condition,
#	the upper-bound of the effective count of patterns that actually were observed $ct_obs times.
#
#	THEORY:
#	Under a binomial distribution with the Bernoulli probability of p and the total sample size of N,
#	the probability that we observe r sites with the pattern in question is:
#	 Prob (X = r | p, N)_bin
#	      = C(N, r) * p ** r * (1 - p)**(N - r),
#	 where  C(N, r) := N!/(N - r)!r!) .
#
#	Assuming a uniform prior for $p: 
#	Prob ( p < Y < p + dp) = dp/(B-A)   for 0<= A < p < B <= 1,
#
#	The joint probability, given the sample size N, is given as:
#
#	Prob (p < Y < p + dp, X = r | N)
#		= dp/(B-A) * C(N, r) * p ** r * (1 - p) ** (N-r). -- (1)
#
#	The posterior probability that we observe r sites with the pattern is:
#	Prob (X = r | N) = C(N, r)/(B-A) * Int_A^B dp p ** r * (1-p)**(N-r). -- (2)
#
#	Using Bayes' theorem, the posterior probability for the Bernoulli probablity parameter,
#	provided that we observe r sites in question, is:
#	Prob ( p < Y < p + dp | X = r, N) = Prob (p < Y < p+dp, X = r | N) /Prob (X = r| N)
#	= dp * p**r * (1-p)**(N-r) / [ Int_A^B dp p**r * (1-p)**(N-r) ] . --- (3)
#
#	When we want to achieve a target probability P0 by integrating the above posterior probability
#	over the interval A < p < pM, we get the following equation for pM:
#
#	P0 = Int_A^pM dp * p ** r * (1-p)**(N-r) / [ Int_A^B dp p**r * (1-p)**(N-r) ]
#	    = [I_pM (r+1, N-r+1) - I_A (r+1, N-r+1) / [I_B (r+1, N-r+1) - I_A (r+1, N-r+1)],
#	where I_x (a, b) is the incomplete beta function.
#	This can be rearranged as:
#
#	I_pM (r+1, N-r+1) = P0 * I_B (r+1, N-r+1) + (1 - P0)* I_A (r+1, N-r+1). -- (4)
#
#	This equation (4) can be numerically solved for pM using 'inv_betai_nwt' or 'inv_betai_lvr2'.
#
# INPUT: ($totct_sites, $ct_obs, $tgt_prb, $p_lw, $p_up, $resol, $maxite), where
#	$totct_sites, corresponding to N above, is the total number of sites that can be used in the alignment;
#	$ct_obs, corresponding to r above, is the number of sites that are observed to have a particular pattern;
#	$tgt_prb, corresponding to P0 above, is the probability that
#	the Bernoulli probability p resides in the interval [$p_lw, $p_M];
#	$p_lw and $p_up, respectively corresponding to A and B above, are
#	the lower bound and upper bound of the Bernoulli probability you want to impose;
#	$resol is the resolution you desire for the numerical solution for the inverse incomplete beta;
#	$maxite is the maximum iteration count for the inverse incomplete beta.
#
# OUTPUT: ($effct, $error_x, $error_y, $itect),
#	$effct is the 'effective count', which equals $totct_sites * {upper bound of Bernoulli probability given by the solution of equation (4) above};
#	$error_x is the magnitude of the error in the solution for x;
#	$error_y is the magnitude of the error in the beta function;
#	$itect is the count of iterations, it is $maxite + {iteration count}
#	when 'inv_betai_lvr2' is used.
#
#	The subroutine returns (undef, undef, undef, undef) if it failed.
#
sub upper_bound_eff_count ($$$$$$$) {

	my ($totct_sites, $ct_obs, $tgt_prb, $p_lw, $p_up, $resol, $maxite) = @_;

	unless ((0 <= $p_lw) and ($p_lw < $p_up) and ($p_up <=1)) {
	    print STDERR "[ERROR 'upper_bound_eff_count'] ILLEGAL INTERVAL: (\$p_lw, \$p_up) = ($p_lw,  $p_up): $! !\n";
	    return undef;
	}
	unless ((0 <= $tgt_prb) and ($tgt_prb <=1)) {
	    print STDERR "[ERROR 'upper_bound_eff_count'] ILLEGAL TARGET PROBABILITY: \$tgt_prb = $tgt_prb is NOT in the interval \[0, 1\]: $! !\n";
	    return undef;
	}

	my $a = $ct_obs +1;
	my $b = $totct_sites - $ct_obs +1;

	my $betai_lw = ($p_lw == 0) ? 0 : betai ($a, $b, $p_lw);
	unless (defined $betai_lw) {
	    print STDERR "[ERROR 'upper_bound_eff_count'] 'betai' FAILED for the lower bound ($p_lw) of the Bernoulli probability interval \[$p_lw, $p_up\]: $! !\n";
	    return (undef, undef, undef, undef);
	}
	my $betai_up = ($p_up == 1) ? 1 : betai ($a, $b, $p_up);
	unless (defined $betai_up) {
	    print STDERR "[ERROR 'upper_bound_eff_count'] 'betai' FAILED for the upper bound ($p_up) of the Bernoulli probability interval \[$p_lw, $p_up\]: $! !\n";
	    return (undef, undef, undef, undef);
	}

	my $y = $tgt_prb * $betai_up + (1 - $tgt_prb) * $betai_lw;

	my ($px, $itect, $error_x, $error_y)
	    = inv_betai_nwt ($a, $b, $y, $maxite, $resol, $resol);
	unless (defined $px) {
	    print STDERR "[ERROR 'upper_bound_eff_count'] 'inv_betai_nwt' FAILED for (\$a, \$b, \$y) = ($a, $b, $y): $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless ((defined $error_x) and ($error_x <= $resol) and (defined $error_y) and ($error_y <= $resol)) {
	     # Try a binary search as a substitute #
	    my ($px2, $itect2, $error_x2, $error_y2)
		= inv_betai_lvr2 ($a, $b, $y, $maxite, $maxite, $resol, $resol);
	    unless (defined $px2) {
		print STDERR "[ERROR 'upper_bound_eff_count'] 'inv_betai_lvr2' FAILED for (\$a, \$b, \$y) = ($a, $b, $y): $! !\n";
	    }
	    if ($error_y2 < $error_y) {
		($px, $itect, $error_x, $error_y)
		   = ($px2, $maxite + $itect2, $error_x2, $error_y2);
	    }
	}

	my $effct = $totct_sites * $px;

	return ($effct, $error_x, $error_y, $itect);

}	# End of the subroutine 'upper_bound_eff_count' #


# partition_hierarchy:
#	This subroutine constructs a hierarchical structure
#	in a set of partitions,
#	with the aid of inclusion relaionships among them.
#
#	THEORY:
#	Because the number of included partitions will
#	decrease as you go down the hierarchy,
#	you can establish the hierarchical relationship easily
#	if you process the partitions in the decreasing order
#	of the number of included partitions.
#
# INPUT: (\%indx2pres, \@smaller, \@indices_rlv), where
#	%indx2pres = ($indx_ptit => 1, ...)
#	for the partitions contained in	the set in question;
#	@{$smaller[$k]} lists the indices of the partitions included in the $indices_rlv[$k] th partition (excluding the identical ones);
#	@indices_rlv lists the indices of the partitions relevant for the current session.
#
# OUTPUT: (\@indices_top, \%ptit2children), where
#	@indices_top lists the indices of the partitions NOT included in any other partitions;
#	%ptit2children = ($indx_ptit => \@children, ...)
#	 for the partitions contained in the set in question.
#	@children lists the indices of the 'children',
#	or immediately included partitions,
#	which are the partitions included in the subject partition
#	but NOT in any other partitions included in the subject.
#
sub partition_hierarchy (\%\@\@) {

	my ($indx2pres, $smaller, $indices_rlv) = @_;

	my $ct_rlv = @{$indices_rlv};
	unless ($ct_rlv == @{$smaller}) {
	    print STDERR "[ERROR 'partition_hierarchy'] ILLEGAL INPUT: The size of \@\$smaller (", scalar (@{$smaller}), ") CLASHES with the number of relevant partitions ($ct_rlv): $! !\n";
	    return (undef, undef);
	}

	my %ptit2included; # For only partitions contained in the set in question #
	my %ptit2ct_incl;
	for (my $k=0; $k<$ct_rlv; $k++) { # 1st for-loop #
	    my $indx_sb = $indices_rlv->[$k];
	    (defined $indx2pres->{$indx_sb}) or next; # Skip partitions NOT in the set 
	    my @included = ();
	    foreach my $indx_sml (@{$smaller->[$k]}) { (defined $indx2pres->{$indx_sml}) and (push @included, $indx_sml); }
	    $ptit2included{$indx_sb} = \@included;
	    $ptit2ct_incl{$indx_sb}  = scalar (@included);
	}			# End of the 1st for-loop #

	my %ptit2children;
	my %included;
	foreach my $indx_sb (keys %ptit2included) { # 2nd foreach-loop #

	    my @std_included =sort { $ptit2ct_incl{$b} <=> $ptit2ct_incl{$a} }
			@{$ptit2included{$indx_sb}} ;
	    my %indirect; # ($indx => 1, ...) for indirectly included partitions
	    my @children = ();
	    foreach my $indx (@std_included) { # 2nd inner foreach-loop #
		(defined $indirect{$indx}) and next; # Skip indirect offsprings
		$included{$indx} = 1;
		push @children, $indx;
		my $indirect_ptits = $ptit2included{$indx};
		foreach my $indx_ind (@{$indirect_ptits}) { $indirect{$indx_ind} = 1; }
	    }				# End of the 2nd inner foreach-loop #

	    $ptit2children{$indx_sb} = \@children;

	}			# End of the 2nd foreach-loop #

	my @indices_top = ();
	foreach my $indx (sort { $ptit2ct_incl{$b} <=> $ptit2ct_incl{$a} } keys %ptit2ct_incl) {
	    (defined $included{$indx}) or (push @indices_top, $indx);
	}

	return (\@indices_top, \%ptit2children);

}	# End of the subroutine 'partition_hierarchy' #


# fetch_solitary_node:
#	This satellite subroutine of 'partition_hierarchy2tree'
#	examines if a partition splits a single exterior node from the rest or not,
#	and, if it does, returns the ID of the exterior node separated.
#
# INPUT: (\@numrep, $size_bin, $denom, \@seqnames, \%seqname2nodeid), where
#	@numrep is the numerical array representation of the input partition;
#	$size_bin is the length of the binary string representing the partition;
#	$denom is the unit length (in bits) which the binary string is cut into;
#	@seqnames lists the sequence names;
#	%seqname2nodeid = ($seqname => $node_id, ...) for  exterior nodes.
#
# OUTPUT: $node_id if the input partition separates a single sequence,
#	  -1 otherwise.
#
#	The subroutine returns undef if failed.
#
sub fetch_solitary_node (\@$$\@\%) {

	my ($numrep, $size_bin, $denom, $seqnames, $seqname2nodeid) = @_;

	my ($ct_0, $ct_1) = count_binary_states (@{$numrep}, $size_bin, $denom) ;

	if ($ct_1 == 1) { # The partition separates only one sequence #

	    my $separated = select_by_bin_value (1, @{$seqnames}, @{$numrep}, $denom);

	    unless ((defined $separated) and (@{$separated} == 1)) {
		my @separated = (defined $separated) ? @{$separated} : ('undef') ;
		print STDERR  "[ERROR 'fetch_solitary_node'] 'select_by_bin_value' FAILED: The partition, which should separate one and only one sequence, separates the following: ", join (', ', @separated), " : $! !\n";
		return undef;
	    }
	    my $name_solo = $separated->[0];
	    my $id_solo = $seqname2nodeid->{$name_solo};

	    return $id_solo;

	} else {	# The partition separates more than one sequence #
	    return -1 ;
	}

}	# End of the subroutine 'fetch_solitary_node' #


# child_nodes_of_partition:
#	This satellite subroutine of 'partition_hierarchy2tree'
#	enumerates child nodes of a partition, both interior and exterior.
#	It also marks the partitions corresponding to (the upstream of) exterior nodes.
#
# INPUT: (\@numrep_ptit, \@ch_ptits, \@set_ptits, $indx_numrep, $size_bin, $denom, \@seqnames, \%seqname2nodeid, \%ext_ptits), where
#	@numrep_ptit is the numerical representation of the subject partition;
#	@ch_ptits lists the indices of the child partitions of the subject;
#	@{$set_partitions[$k]->[$indx_numrep]} when $indx_numrep is defined,
#	or @{$set_partitions[$k]} otherwise,
#	gives the numerical representation of the $k th partition;
#	$size_bin is the length of the binary string representing the partition;
#	$denom is the unit length of pieces into which binary strings are divided;
#	@seqnames lists the names of the sequences except the last one;
#	%seqname2nodeid = ($seq_name => $node_id, ...) for the exterior nodes;
#	%ext_ptits = ($indx_ptit => $nodeid, ...) for the partitions corresponding to
#	(the upstream of) the exterior nodes;
#
# OUTPUT: \@child_nodes,
#	which lists the IDs of the child nodes.
#	The exterior nodes will be assigned pre-specified node IDs,
#	and the interior nodes will be assigned the node ID equaling 1 + {index of the partition}.
#
sub child_nodes_of_partition (\@\@\@$$$\@\%\%) {

	my ($numrep_ptit, $ch_ptits, $set_ptits, $indx_numrep, $size_bin,
		$denom, $seqnames, $seqname2nodeid, $ext_ptits) = @_;

	my $numrep_rest_ch = $numrep_ptit;

	my @child_nodes = ();
	foreach my $indx_ch (@{$ch_ptits}) { # 2nd foreach-loop #
	    my $info_ch = $set_ptits->[$indx_ch];
	    my $numrep_ch = (defined $indx_numrep) ? $info_ch->[$indx_numrep] : $info_ch ;

		# Subtract the exterior nodes separated by the child partition
		# from those separated by the parent
	    $numrep_rest_ch = bitwise_xor_numarrays (@{$numrep_rest_ch}, @{$numrep_ch}) ;

	    if (defined $ext_ptits->{$indx_ch}) {
		    # Already known to be (the upstream of) an exterior node #
		push @child_nodes, $ext_ptits->{$indx_ch};
		next;
	    }

	    my $id_solo = fetch_solitary_node (@{$numrep_ch}, $size_bin, $denom, @{$seqnames}, %{$seqname2nodeid}) ;
	    unless (defined $id_solo) {
		print STDERR "[ERROR 'child_nodes_of_partition'] 'fetch_solitary_node' FAILED on the child partition $indx_ch : $! !\n";
		return undef;
	    }

	    if ($id_solo < 0) {
		    # The partition corresponds to (the upstream of) an interior node
		push @child_nodes, 1 + $indx_ch ;

	    } else {
		    # The partition corresponds to (the upstream of) an exterior node
		push @child_nodes, $id_solo ;
		$ext_ptits->{$indx_ch} = $id_solo;
	    }

	}			# End of the 2nd foreach-loop #

	my $seqs_rest_ch = select_by_bin_value (1, @{$seqnames}, @{$numrep_rest_ch}, $denom);
	foreach my $seqname (@{$seqs_rest_ch}) {
	    my $id_extnode = $seqname2nodeid->{$seqname};
	    push @child_nodes, $id_extnode;
	}

	return \@child_nodes;

}	# End of the subroutine 'child_nodes_of_partition' #


# partition_hierarchy2tree:
#	This subroutine constructs a tree
#	based on a  hierarchical relationship
#	among partitions that are compatible with each other.
#	This subroutine uses the numerical array representation of
#	a binary string representing a partition, where
#	'1' denotes a different character than the last sequence,
#	and '0' denotes the same character as the last sequence.
#
#	ALGORITHM:
#	The tree is constructed in the following way:
#	(1) First deal with the top node.
#	The top node here is defined as the node on top of all the partitions and the sequence entries.
#	So, consider the top partitions as children of the top node.
#	And, those sequences, including $last_seq, that do NOT belong to the top partitions
#	are also regarded as children of the top node;
#	(2) Then, examine each of other partitions.
#	The set of child nodes consists of child  partitions and
#	those sequences included in the subject partition but not in the child partitions;
#	(3) When conducting (1) or (2), we have to assign node IDs to partitions and sequences.
#	IDs for the sequences, namely the exterior nodes, are assigned uniquely, and easily,
#	according to the scheme explained below.
#	However, we have to be careful when assigning a node ID to a partition.
#	(a) If the partition splits 2 or more sequences (from $last_seq), it is judged as (the upstream of) an interior node;
#	(b) If the partition splits only one sequence (from $last_seq), it is judged as (the upstream of) an exterior node.
#
# INPUT: (\@top_ptits, \%ptit2children, \@set_partitions, $indx_numrep, $denom, \@seqnames, $last_seq), where
#	@top_ptits lists the indices of the top partitions;
#	%ptit2children = ($indx_partition => \@children,...)
#	for the partitions in a set of compatible partitions, where
#	@children lists the indices of 'children', which are directly included in $indx_partition;
#	@{$set_partitions[$k]->[$indx_numrep]} when $indx_numrep is defined,
#	or @{$set_partitions[$k]} otherwise,
#	gives the numerical representation of the $k th partition;
#	$denom is the unit length of pieces into which binary strings are divided;
#	@seqnames lists the names of the sequences except the last one;
#	$last_seq is the name of the last sequence, which gives the reference for the other sequences.
#
# OUTPUT: (\%node2children, \%nodeid2seqname, \%seqname2nodeid, \%ext_ptits), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes,
#	where {ID of an interior node} = 1 + {index of the corresponding partition};
#	%nodeid2seqname = ($node_id => $sequence_name, ...)  for the exterior nodes,
#	where {ID of an exterior  node} = $topnode_id + 1 + {index of the corresponding sequence entry}.
#	%seqname2nodeid = ($sequence_name => $node_id, ...) for the exterior nodes and the top node,
#	with $topnode_id = 1 + scalar (@set_partitions) being used here;
#	%ext_ptits = ($indx_partition => $node_id, ...) for the partitions in %ptit2children
#	each of which splits only exterior nodes.
#
sub partition_hierarchy2tree (\@\%\@$$\@$) {

	my ($top_ptits, $ptit2children, $set_ptits, $indx_numrep, $denom, $seqnames, $last_seq) = @_;

		# Preliminary Processes #
	my $ct_ptits = @{$set_ptits};
	my $topnode_id = 1 + $ct_ptits;
	my $size_bin = @{$seqnames}; # This equals #{sequences} - 1
	my $last_ext_id = $topnode_id + 1 + $size_bin;

	my %nodeid2seqname = ($last_ext_id => $last_seq);
	my %seqname2nodeid = ($TOP_NODE => $topnode_id, $last_seq => $last_ext_id);
	for (my $i = 0; $i< $size_bin; $i++) { # 1st for-loop #
	    my $seqname = $seqnames->[$i];
	    my $nodeid = $topnode_id + 1 + $i;
	    $nodeid2seqname{$nodeid} = $seqname;
	    $seqname2nodeid{$seqname} = $nodeid;
	}			# End of the 1st for-loop #

	my %node2children;
	my %ext_ptits; # ($indx_ptit => $node_id, ...) for partitions immediately upstream of the exterior nodes #

		# Enumerate the children of the top node #

	my $ptit_top_node = make_monobit_pattern (1, $size_bin, $denom);

	my $children_top = child_nodes_of_partition (@{$ptit_top_node}, @{$top_ptits}, @{$set_ptits}, $indx_numrep, $size_bin, $denom, @{$seqnames}, %seqname2nodeid, %ext_ptits);
	unless (defined $children_top) {
	    print STDERR "[ERROR 'partition_hierarchy2tree'] 'child_nodes_of_partition' FAILED on the top node: $! !\n";
	    return (undef, undef, undef, undef);
	}
	push @{$children_top}, $last_ext_id; # Include the node corresponding to $last_seq
	$node2children{$topnode_id} = $children_top;

	foreach my $indx_ptit (keys %{$ptit2children}) { # Outer foreach-loop #

	    my $info_ptit = $set_ptits->[$indx_ptit];
	    my $numrep_ptit = (defined $indx_numrep) ? $info_ptit->[$indx_numrep] : $info_ptit ; 
	    my $ch_ptits = $ptit2children->{$indx_ptit};
	    unless (@{$ch_ptits}>0) {
		(defined $ext_ptits{$indx_ptit}) and next; # Already known to be (the upstream of) an exterior node --> There should be no child nodes #
	    }
	    my $ch_nodes = child_nodes_of_partition (@{$numrep_ptit}, @{$ch_ptits}, @{$set_ptits}, $indx_numrep, $size_bin, $denom, @{$seqnames}, %seqname2nodeid, %ext_ptits) ;
	    unless (defined $ch_nodes) {
		print STDERR "[ERROR 'partition_hierarchy2tree'] 'child_nodes_of_partition' FAILED on the partition '$indx_ptit': $! !\n";
		return (undef, undef, undef, undef);
	    }
	    if (@{$ch_nodes} ==1) {
		my $ch_node = $ch_nodes->[0];
		if (defined $nodeid2seqname{$ch_node}) {
			# The partition is (immediately upstream of) an exterior node #
		    $ext_ptits{$indx_ptit} = $ch_node;
		    next;
		} else {
		    print STDERR "[ERROR 'partition_hierarchy2tree'] It's IMPOSSIBLE !! The partition '$indx_ptit' is immediately UPSTREAM OF A SINGLE INTEROIR NODE '$ch_node' : $! !\n\n";
		    return (undef, undef, undef, undef);
		}
	    }

	    my $id_node = 1 + $indx_ptit;
	    $node2children{$id_node} = $ch_nodes;

	}					# End of the outer foreach-loop #

	return (\%node2children, \%nodeid2seqname, \%seqname2nodeid, \%ext_ptits);

}	# End of the  subroutine 'partition_hierarchy2tree' #


# make_branch2partition:
#	This subroutine is a gateway of the conversion from the tree
#	(using the parent-child relationship of the nodes)
#	to the set of partitions.
#	It constructs a hash table associating each branch with 
#	the numerical representation of the correponding partition.
#	As the above subroutines, '1'/'0' represents that the entry has a character
#	different from/same as that in the last sequence entry.
#	Accordingly, a partition is represented as a binary string of length (#{sequences} -1),
#	which in turn is represented by an array of non-negative integers.
#
#	ALGORITHM:
#	It assigns partitions to nodes in a bottom-up manner.
#	First, exterior nodes are assigned a partition,
#	represented by a binary string with '1' at only the bit corresponding to the sequence,
#	except the last sequence, which is represented by a binary string consiting only of '1'.
#	Then, while climing up the tree,
#	(i) if a node is NOT the ancestor of the last sequence, it is represented by
#	the OR operations over the binary strings of its children;
#	(ii) an ancestor of the last sequence is represented by
#	{the binary string of a children which also is an ancestor of the last sequence}
#	XOR {the OR operations over the binary strings of the other children}.
#	This will continue until it reaches the top node,
#	which will be assigned nothing,
#	because the top node does NOT correspond to any branch.
#	
#
# INPUT: (\%node2pa, \%node2ch, \%name2id, \@seqnames, $denom_bin), where
#	%node2pa = ($node_id => $parent_id, ...) for the nodes except the top node;
#	%node2ch = ($node_id => \@child_ids, ...) for the interior nodes;
#	%name2id = ($node_name => $node_id, ...) for the exterior nodes and the top node;
#	@seqnames lists the sequence names;
#	$denom_bin is the unit length of pieces into which binary strings are divided;
#
# OUTPUT: \%br2partition = ($node_id => \@numrep_partition, ...)
#	  for the nodes each immediately below the branch in question.
#
sub make_branch2partition (\%\%\%\@$) {

	my ($node2pa, $node2ch, $name2id, $seqnames, $denom_bin) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>1) {
	    print STDERR "[ERROR 'make_branch2partitions'] ILLEGAL INPUT: The number of sequences ($ct_seqs) is TOO SMALL to define non-trivial partitions: $! !\n";
	    return undef;
	}

		## Prepare auxiliary variables ##

	my $bin_size = $ct_seqs - 1;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});

	my %nodeid2indx; # ($node_id => $index, ...) for the exterior nodes.
	for (my $k=0; $k<$ct_seqs; $k++) { # 1st preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    $nodeid2indx{$node_id} = $k;
	}			# End of the 1st preliminary for-loop #

=for DEBUGGING the subroutine
	print STDERR "< 'make_branch2partition' >\n\n";
	print STDERR '%nodeid2indx = (', "\n";
	foreach my $id (sort {$nodeid2indx{$a} <=> $nodeid2indx{$b}} keys %nodeid2indx) {
		my $indx = $nodeid2indx{$id};
		my $name = $seqnames->[$indx];
		print STDERR "\t$id  => $indx ($name)\,\n";
	}
	print STDERR "\t).\n\n";
=cut

	my $lastseq_name = $seqnames->[$#{$seqnames}];
	my $lastseq_nodeid = $name2id->{$lastseq_name};
	unless (defined $lastseq_nodeid) {
	    print STDERR "[ERROR 'make_branch2partitions'] ILLEGAL INPUT: The last sequence (name: $lastseq_name) does NOT HAVE a NODE ID: $! !\n";
	    return undef;
	}
	my $ancestry_lastseq = fetch_ancestry ($lastseq_nodeid, %{$node2pa});
	my %if_anc_lastseq;
	foreach my $nodeid (@{$ancestry_lastseq}) { $if_anc_lastseq{$nodeid} = 1; }

		## Main Process ##
	my %br2partition;

	my @std_nodeids = sort { $node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2depth};
	foreach my $nodeid (@std_nodeids) { # Main foreach-loop #
	    ($nodeid == $topnode_id) and last;

	    my $indx = $nodeid2indx{$nodeid};
	    if (defined $indx) {
		# The node is an exterior node #
		if ($indx == $bin_size) {
			# The last sequence #
		   my $numrep_ptit = make_monobit_pattern (1, $bin_size, $denom_bin);
		   $br2partition{$nodeid} = $numrep_ptit;
		} else {
			# Otherwise #
		   my @sbset = ($indx);
		   my $numrep_ptit = sbset2numrep_binstring (@sbset, $bin_size, $denom_bin);
		   $br2partition{$nodeid} = $numrep_ptit;
		}
		next;
	    }

		# The node must be an interior node #

	    my $children = $node2ch->{$nodeid};
	    unless ((defined $children) and (@{$children}>0)) {
		print STDERR "[ERROR 'make_branch2partitions'] The 'interior' node $nodeid does NOT have any CHILD NODES: $! !\n";
		return undef;
	    }

	    my $numrep_ptit_last_ch;
	    my $numrep_ptit_sbjct = make_monobit_pattern (0, $bin_size, $denom_bin);
	    foreach my $ch_id (@{$children}) { # Inner foreach-loop #
		my $numrep_ptit_ch = $br2partition{$ch_id};
		unless (defined $numrep_ptit_ch) {
		    print STDERR "[ERROR 'make_branch2partitions'] The child node '$ch_id' of the node '$nodeid' is NOT ASSIGNED a PARTITION: $! !\n";
		    return undef;
		}

		if (defined $if_anc_lastseq{$ch_id}) {
			# The child is an ancestor of the last sequence #
		    $numrep_ptit_last_ch = $numrep_ptit_ch;
		} else {
			# Otherwise #
		    $numrep_ptit_sbjct = bitwise_or_numarrays (@{$numrep_ptit_sbjct}, @{$numrep_ptit_ch});
		}
	    }			# End of the inner foreach-loop #

	    if (defined $numrep_ptit_last_ch) {
		# 'Merge' the partition for the children
		# which is an ancestor of the last sequence
		# and the composite partition of the other chldren.

		$numrep_ptit_sbjct = bitwise_xor_numarrays (@{$numrep_ptit_last_ch}, @{$numrep_ptit_sbjct});
	    }

	    $br2partition{$nodeid} = $numrep_ptit_sbjct;

	}			# End of the main foreach-loop #

=for DEBUGGING the subroutine
	print STDERR " --- 'make_branch2partition' ENDED --- \n\n";
=cut

	return \%br2partition;

}	# End of the subroutine 'make_branch2partition' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


