package MySpecFunc;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MySpecFunc.pm : a module for numericaly computing various special functions.
#
# REVISED on June 14 (Tue), 2005:  The calculation has became more rapid but less accurate.
#
#	All subroutines in this module are basically Perl-translation versions of the programs in:
#	"Numerical Recipes in C: The Art of Scientific Computing", 2nd edition,
#	written by Press WH, Teukolsky SA, Vetterling WT, Flannery BP (Cambridge Univ. Press; Cambridge, UK; 1992).
#
#


use strict;
use warnings;
use Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(gammln gammp gammq erff erffc betai);
#our @EXPORT_OK = qw(gammln);

##sub nrerror(\$); # reference to a string??? #
sub betacf ($$$); 

our $ITMAX = 200;		# Maximun allowed number of iterations #
#our $ITMAX = 1000;		# Maximun allowed number of iterations #
#our $EPS = 1.0e-8;		# Relative accuracy #
#our $EPS = 3.0e-7;		# Relative accuracy #
our $EPS = 1.0e-15;		# Relative accuracy #
##our $EPS = 1.0e-10;		# Relative accuracy #   --- The only revision on June 14 (Tue), 2005
our $FPMIN = 1.0e-30;		# Number near the smallest rpresentable floating-point number #


# gammln(float $xx): This subroutine returns the value ln(Gamma($xx)) for $xx >0.
#
sub gammln ($) {

	my ($xx) = @_;

	if ( ($xx == 1) or ($xx == 2) ) { return 0; }

	my @cof = (76.18009172947146, -86.50532032941677, 24.01409824083091,
		 -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5);

	my $x = my $y = $xx;
	my $tmp = $x + 5.5;
	$tmp -= ($x + 0.5) * log ($tmp);
	my $ser = 1.000000000190015;

	for (my $j=0; $j < 6; $j++) {

		$y++;
		$ser += $cof[$j]/$y;
	}

	return -$tmp + log (2.5066282746310005 * $ser/ $x);

}				# end of the subroutine gammln #


# gser (float* $rf_gamser,float $a,float $x, float* $rf_gln):
# Returns the incomplete gamma function P($a,$x) evaluated by its series representation as ${$rf_gamser}.
# Also returns ln (Gamma($a)) as ${$rf_gln}. 
#
sub gser (\$$$\$) {

	my ($rf_gamser, $a, $x, $rf_gln) = @_;

	${$rf_gln} = gammln($a);
	if ($x <= 0.0) {		# the outer if-else #

		if ($x < 0.0) {

			print STDERR "x=$x less than 0 in routine gser\n";
			return 0;
		}

		${$rf_gamser} = 0.0;
		return 1;
	} else {

		my $ap = $a;
		my $del = my $sum = 1.0/$a;

		for (my $n=1; $n <= $ITMAX; $n++) {

			$ap++;
			$del *= $x/$ap;
			$sum += $del;
			my $abdel = ($del < 0) ? -$del : $del;
			my $absum = ($sum < 0) ? -$sum : $sum;

			if ($abdel < $absum * $EPS) {

				${$rf_gamser} = $sum*exp(-$x+$a*log($x)- ${$rf_gln} );
				return 1;
			}
		}

		print STDERR "a=$a too large, iteration times $ITMAX too small in routine gser\n";
		return 0;

	}			# end of the outer if-else #

}				# end of the subroutine gser #

# gcf (float* $rf_gammcf, float $a, float $x, float* $rf_gln) :
# Returns the incomplete gamma function Q($a, $x) evaluated by its continued fraction representation as ${$rf_gammcf}. Also returns ln (Gamma($a) ) as ${$rf_gln}.
#
sub gcf (\$$$\$) {

	my ($rf_gammcf, $a, $x, $rf_gln) = @_;

	${$rf_gln} = gammln($a);

	if ($x < 0.0) {

		print STDERR "x=$x less than 0 in routine gcf\n";
		return 0;
	} elsif ($x == 0.0) {

		${$rf_gammcf} = 1.0;
		return 1;
	}

		# Set up for evaluating continued fraction
		# by modified Lentz's method (sec. 5.2) with b_0 = 0.
	my $b = $x +1.0 - $a;
	my $c = 1.0/$FPMIN;
	my $d = 1.0/$b;
	my $h = $d;

	my $i;
	for($i=1; $i <= $ITMAX; $i++) {	# iterate to converge. 

		my $an = -$i*($i - $a);
		$b += 2.0;
		$d = $an*$d + $b;
		my $abd = ($d < 0) ? -$d : $d;
		($abd < $FPMIN) and ($d = $FPMIN);
		$c = $b + $an/$c;
		my $abc = ($c < 0) ? -$c : $c;
		($abc < $FPMIN) and ($c = $FPMIN);
		$d = 1.0/$d;
		my $del = $d*$c;
		$h *= $del;
		my $abdel = ($del < 1.0) ? -$del+ 1.0 : $del - 1.0;
		( $abdel < $EPS)  and last;
	}

	if ($i > $ITMAX) {

		print STDERR "a= $a too large, iterative times of $ITMAX too small in gcf\n";
		return 0;
	}

	${$rf_gammcf} = exp(-$x + $a * log($x) - ${$rf_gln} ) * $h; # Put factors in front. 
	return 1;

} 			# end of the subroutine gcf #


# float gammp (float $a, float $x) :
# Returns the incomplete gamma function P($a, $x) = 1/Gamma(a)* int_0^x dt exp(-t) * t**(a-1) 
#
sub gammp ($$) {

	my ($a, $x) = @_; 
	my ($gamser, $gammcf, $gln);

	if ($x < 0.0 or $a <= 0.0) {

		print STDERR "Invalid arguments (x,a) = ($x, $a) in routine gammp\n";
		return undef;
	}

	if ( $x < ($a + 1.0) ) { 	# Use the series representation. #

		gser($gamser, $a, $x, $gln) or (return undef);
		return $gamser;
	} else {			# Use the continued fraction representation #

		gcf ($gammcf, $a, $x, $gln) or (return undef);
		return 1.0 - $gammcf; 	# and take its complement. #
	}

}			# end of the subroutine gammp #


# float gammq (float $a, float $x) :
# Returns the incomplete gamma function Q($a, $x) = 1/Gamma(a)* int_x^infty dt exp(-t) * t**(a-1) 
#
sub gammq ($$) {

	my ($a, $x) = @_;
	my ($gamser, $gammcf, $gln);

	if ($x < 0.0 or $a <= 0.0 ) {

		print STDERR "Invalid arguments (x,a) = ($x, $a) in routine gammq\n";
		return undef;
	}

	if ($x < ($a + 1.0) ) {		# Use the series representation #

		gser($gamser, $a, $x, $gln) or (return undef);
		return 1.0 - $gamser;	# and take its complement. #

	} else {			#  Use the continued fraction representation #

		gcf ($gammcf, $a, $x, $gln) or (return undef);
		return $gammcf;
	}

}			# end of the subroutine gammq #


# float erff (float $x)
# Returns the error function erf(x) = 2/sqrt(pi)* int_0^x dt exp(-t*t) .
#
sub erff ($) {

	my ($x) = @_;

	return ($x < 0) ? -gammp (1/2, $x * $x) : gammp (1/2, $x * $x) ;

}		# end of the  subroutine erff #

# float erffc (float $x)
# Returns the complementary error function erfc(x) = 2/sqrt(pi) * int_x^infty dt exp (-t*t) .
#
sub erffc ($) {

	my ($x) = @_;

	return ($x < 0) ? 1 + gammp (1/2, $x * $x) : gammq (1/2, $x * $x) ;

}		# end of the subroutine erffc #


# float betacf (float $a, float $b, float $x) 
#
# Used by betai: Evaluates continued fraction for incomplete betafunction
# by modified Lentz's method (sec. 5.2).
#
# (Reference: "Numerical Recipes in C", 2nd ed., pp. 227-228. )
#
sub betacf ($$$) {

	my ($a, $b, $x) = @_;

	   # These q's will be used in factors that occur in the coefficients (6.4.6).
	my $qab = $a + $b;
	my $qap = $a + 1;
	my $qam = $a -1;

	  # The 1st step of Lenz's method #
	my $c =1;
	my $d = 1 - $qab * $x /$qap;
	( ($d < $FPMIN)  and (- $d  < $FPMIN) )  and ($d = $FPMIN);
	$d = 1 /$d;
	my $h = $d;	# The variable which is to be returned #

	my $m;
	for ($m =1; $m <= $ITMAX; $m++) {

		my $m2 = 2 * $m;
		my $aa = $m * ($b - $m) * $x /(($qam+$m2)*($a+$m2));

			# One step (the even one) of the recurrence #
		$d = 1 + $aa *$d;
		( ($d < $FPMIN) and (-$d < $FPMIN) ) and ($d = $FPMIN);
		$c = 1 + $aa /$c;
		( ($c < $FPMIN) and (-$c < $FPMIN) ) and ($c = $FPMIN);
		$d = 1/$d;
		$h *= $d * $c;
		$aa = - ($a+ $m)*($qab +$m)*$x/ (($a+$m2)*($qap+$m2));

			# Next step of the recurrence (the odd one) #
		$d = 1 + $aa *$d;
		( ($d < $FPMIN) and (-$d < $FPMIN) ) and ($d = $FPMIN);
		$c = 1 + $aa/$c;
		( ($c < $FPMIN) and (-$c < $FPMIN) ) and ($c = $FPMIN);
		$d = 1 /$d; 
		my $del = $d *$c;
		$h *= $del;
		( ($del - 1 < $EPS) and (-$del +1  <$EPS) ) and last; # Are we done ?
	}

	if ($m > $ITMAX) {
		print STDERR "a (=$a) or b (=$b) too big, or MAXIT (=$ITMAX) too small in \'betacf\'.\n";
	}

	return $h;

}		# end of the subroutine betacf #


# float betai (float $a, float $b, float $x)
# Returns the incomplete  beta function:
#	I_x (a,b) = 1/B(a,b) * int_0^x dt t**(a-1) * (1 - t)**(b-1)
#
sub betai ($$$) {

	my ($a, $b, $x) = @_;

	if ( ($x < 0 ) or ($x > 1) or ($a <=0) or ($b <= 0) ) {
		print STDERR "Illegal input for 'betai': (a, b, x) = ($a, $b, $x).\n";	
		return undef;
	}

	if ($x ==0 ) {
		return 0;
	} elsif ($x == 1) {
		return 1;
	}

		# Factors in front of the continued fraction # 
	#my $bt = ( ($x == 0) or ($x == 1) ) ? 0 :
	my $bt = exp ( gammln($a+ $b) - gammln ($a) - gammln ($b)
			+ $a * log ($x) + $b * log (1 - $x) ) ;

	if ( ($a + $b + 2) * $x < ($a + 1) ) { # Use continued fraction directly #

		return $bt * betacf ($a, $b, $x) / $a;

	} else { # Use continued fraction after making the symmetry transformation #

		return (1 - $bt * betacf ($b, $a, 1 - $x) / $b ); 
	}
	

}	# end of the subroutine  betai #


1;	# a "spell" to make the module to be "use"-able and "require"-able #


