package MyTreeManipu_III;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeManipu_III.pm: This package conducts advanced tree manipulation & analyses,
#	mainly based on the modules 'MyTaxonomy.pm' and 'MyTreeManipu_II.pm'.
#
# MyTreeManipu_II.pm: This package can process more general types of trees,
#	(i.e. NOT limited to binary trees with trichotomous top nodes).
#	It is also consistent with the format of the taxonomy tree which are handled by
#	the module 'MyTaxonomy.pm'.
#
# MyTreeManipu.pm: This package performs basic tree manipulations.
#
#  Subroutines whose substitutes are available in this module ('MyTreeManipu_II.pm')
#	or in 'MyTaxonomy.pm':
#   our @EXPORT = qw(retrieve_top_node confine_outgroup print_unrooted);
#   our @EXPORT_OK = qw(switch_topnode  print_subtree topo_unrooted topo_subtree
#		list_sequences fetch_node_series dist_nodes
#		    );
#
# Subroutines whose substitutes are not available yet:
#   our @EXPORT_OK = ( retrieve_bootstrap  prod_bootstrap );
#

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array multiple_sort );
use MyTaxonomy qw(
		 fetch_ext_offsprings fetch_ext_offsprings_nr
		fetch_last_common_ancestor fetch_common_ancestors
		);
		# fetch_ancestry fetch_last_common_ancestor_ov2
		# fetch_all_offsprings

use  MyTreeManipu_II qw(
		);
		# shift_topnode move_topnode segregate_subset_nodes
		# sum_branch_lengths inter_node_distance

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
		make_node2abs_depth make_node2ct_exoffs make_node2discrete_level
		make_node2totdist_sbtree group_close_extnodes
		make_taxon2depth  max_min_depths_on_subtree
		fetch_taxon_nodes fetch_extnodes_by_taxon_attr
		search_orthocands search_orthocands2 search_closest_homologs
		search_closest_outgroups search_closest_outgroups2
		search_closest_outgroups3
		ascend_search_closest_ogs traceback_closest_outgroups
		);

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

	# the symbol marking the start of the grouping of nodes.
#our $START_GROUP = $MyTreeManipu_II::START_GROUP;
	# the symbol marking the end of the grouping of nodes.
#our $END_GROUP = $MyTreeManipu_II::END_GROUP;


### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### definition of subroutines ###

# make_node2abs_depth:
#	This subroutine counts the ABSOLUTE depths of ALL nodes.
#	The absolute depth of a node is defined as the number of its ancestral nodes,
#	including itself and the top node.
#
# INPUT: ($topnode_id, \%node2children), whre
#	%node2children = ($node_id => \@child_node_ids, ...) for interior nodes.
#
# OUTPUT: \%node2depth, where
#	%node2depth = ($node_id => $depth, ...) for ALL nodes UNDER the node $topnode_id.
#
sub make_node2abs_depth ($\%) {

	my ($topnode_id, $node2chi) = @_;

	my %node2depth;
	my @sbjcts = ($topnode_id);
	my $depth = 0;

	while (@sbjcts > 0) {
	    $depth++;
	    my @new_sbjcts = ();

	    foreach my $nodeid (@sbjcts) {
		$node2depth{$nodeid} = $depth;
		my $children = $node2chi->{$nodeid};
		(defined $children) and (push @new_sbjcts, @{$children});
	    }
		# Update the set of subject nodes #
	    @sbjcts = @new_sbjcts;
	}

	return \%node2depth;

}	# End of the subroutine 'make_node2abs_depth' #


# make_node2ct_exoffs:
#	This subroutine assigns the nuber of exterior offsprings to each node
#	in  a bottom-up manner.
#
# INPUT: (\%node2ch, \%node2depth), where
#	%node2ch = ($node_id => \@children, ...) for all interior nodes;
#	%node2depth = ($node_id => $absolute_depth, ...) for all nodes;
#
# OUTPUT: \%node2ct_exoffs, where
#	%node2ct_exoffs = ($node_id => $ct_ext_offsprings, ...) for all nodes.
#
sub make_node2ct_exoffs (\%\%) {

	my ($node2ch, $node2depth) = @_;

	my %node2ct_exoffs;
	foreach my $node (sort {$node2depth->{$b} <=> $node2depth->{$a}} keys %{$node2depth}) { # Main foreach-loop #
	    my $children = $node2ch->{$node};
	    unless (defined $children) { # An exterior node #
		$node2ct_exoffs{$node} = 1;
		next;
	    }
	    my $ct_exoffs = 0;
	    foreach my $ch (@{$children}) { $ct_exoffs += $node2ct_exoffs{$ch}; }
	    $node2ct_exoffs{$node} = $ct_exoffs;
	}		# End of the main foreach-loop #

	return \%node2ct_exoffs;

}	# End of the subroutine 'make_node2ct_exoffs' #


# make_node2discrete_level:
#	This subroutine assigns the DISCRETE LEVELs to ALL nodes.
#	Here the discrete level of a node is defined as
#	the smallest number of intervening nodes
#	between itself and the exterior node (including the boundaries).
#	(CREATED on June 17 (Wed), 2010)
#
# INPUT: (\%node2ch, \%node2pa, \%node2depth), where
#	%node2ch = ($node_id => \@children, ...) for all interior nodes;
#	%node2pa = ($node_id => $parent_id, ...) for all non-top nodes;
#	%node2depth = ($node_id => $absolute_depth, ...) for all nodes;
#
# OUTPUT: \%node2level, where
#	%node2level = ($node_id => $discrete_level, ...) for all nodes.
#
sub make_node2discrete_level (\%\%\%) {

	my ($node2ch, $node2pa, $node2depth) = @_;

	my %node2level;

		# First assign temporary levels in a bottom-up manner #
	foreach my $node (sort {$node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2depth}) { # The 1st main foreach-loop #
	    my $children  = $node2ch->{$node};
	    unless (defined $children) { # Exterior nodes #
		$node2level{$node} = 1;
		next;
	    }

	    my $min_level_ch;
	    foreach my $ch (@{$children}) {
		my $level_ch = $node2level{$ch};
		((defined $min_level_ch) and ($min_level_ch <= $level_ch)) or ($min_level_ch = $level_ch);
	    }
	    $node2level{$node} = 1 + $min_level_ch;

	}		# End of the 1st main foreach-loop #

		# Then refine the levels by comparing the parent nodes (in a top-down manner) #
	foreach my $node (sort {$node2depth->{$a} <=> $node2depth->{$b} } keys %{$node2depth}) { # The 2nd main foreach-loop #
	    my $pa = $node2pa->{$node};
	    (defined $pa) or next; # Skip the top node #

	    my $level_pa = $node2level{$pa};
	    my $pot_level = $level_pa + 1;
	    my $curr_level = $node2level{$node};
	    ($pot_level < $curr_level) and ($node2level{$node} = $pot_level);
	}		# End of the 2nd main foreach-loop #

	return \%node2level;

}	# End of the subroutine 'make_node2discrete_level' #


# make_node2totdist_sbtree:
#	This subroutine calculates the total length of
#	the sub-trees in a bottom-up manner,
#	and stores the results into a hash table.
#
# INPUT: (\%node2ch, \%br_attr, $indx_length, \%node2depth), where
#	%node2ch = ($node_id => \@child_node_ids, ...) for interior nodes;
#	%br_attr = ($node_id => $attr, ...) for the nodes each immediately below a branch,
#	with the branch length given by $attr->[$indx_length] if $indx_length is defined,
#				or by $attr if $indx_length is undefined;
#	%node2depth = ($node_id => $absolute_depth, ...) for ALL nodes.
#
# OUTPUT: \%node2totdist, where
#	%node2totdist = ($node_id => $totdist, ...), with
#	$totdist being the summation of the branch lengths over the branches under $node_id,
#	and therefore being 0 for the exterior nodes.
#
#	The subroutine returns undef if it failed.
#
sub make_node2totdist_sbtree (\%\%$\%) {

	my ($node2ch, $br_attr, $indx_length, $node2depth) = @_;

	my %node2totdist;
	foreach my $node (sort { $node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2depth}) { ## Main foreach-loop #
	    my $children = $node2ch->{$node};
	    unless ((defined $children) and (@{$children}>0)) {
		$node2totdist{$node} = 0;
	    }

	    my $totdist = 0;
	    foreach my $ch (@{$children}) { # Inner foreach-loop #

		my $totdist_ch = $node2totdist{$ch};
		unless (defined $totdist_ch) {
		    print STDERR "[ERROR 'make_node2totdist_sbtree'] NO TOTAL DISTANCE of the subtree UNDER the child '$ch' of the node '$node': $! !\n";
		    return undef;
		}

		$totdist += $totdist_ch;

		my $attr = $br_attr->{$ch};
		unless (defined $attr) {
		    print STDERR "[ERROR 'make_node2totdist_sbtree'] NO ATTRIBUTE for the BRANCH between the node '$node' and its child '$ch': $! !\n";
		    return undef;
		}
		my $brlen = (defined $indx_length) ? $attr->[$indx_length] : $attr;
		unless (defined $brlen) {
		    print STDERR "[ERROR 'make_node2totdist_sbtree'] NO BRANCH LENGTH INFORMATION between the node '$node' and its child '$ch': $! !\n";
		    return undef;
		}

		$totdist += $brlen;

	    }			# End of the inner foreach-loop #

	    $node2totdist{$node} = $totdist;

	}		## End of the main foreach-loop ##

	return \%node2totdist;

}	# End of the subroutine 'make_node2totdist_sbtree' #


# group_close_extnodes:
#	This subroutine groups together the exterior nodes
#	whose sub-tree starting from their LCA has a total branch length
#	less than a specified threshould value.
#
# INPUT: ($topnode_id, \%node2ch, \%node2totdist, $indx_totdist, $thrsh_totdist), where
#	%node2ch = ($node_id => \@child_ids, ...) for the interior nodes;
#	%node2totdist = ($node_id => $totdist, ...)
#	with $totdist->[$indx_totdist], or $totdist if $indx_totdist = undef,
#	being the total branch length of the subtree under $node_id;
#	$thrsh_totdist is the threshold value of the total branch length
#	to determine whether the sequences are grouped together or not.
#
# OUTPUT: \@set_grp_topnodes, where
#	$set_grp_topnodes[$g] = $grp_topnode_id
#	is the top node ID of the $g th group (or sub-tree) of 'close' sequences,
#
sub group_close_extnodes ($\%\%$$) {

	my ($topnode_id, $node2ch, $node2totdist, $indx_totdist, $thrsh_totdist) = @_;

	my @set_grp_topnodes = ();
	my @subjects = ($topnode_id);

	while (@subjects>0) { # Outer while-loop #
	    my @new_sbjcts = ();

	    foreach my $node (@subjects) { # Inner while-loop #
		my $totdist = $node2totdist->{$node};
		unless (defined $totdist) {
		    print STDERR "[ERROR 'group_close_extnodes'] NO TOTAL BRANCH LENGTH defined for the subtree under the node '$node' : $! !\n";
		    return undef;
		}
		(defined $indx_totdist) and ($totdist = $totdist->[$indx_totdist]);

		if ($totdist < $thrsh_totdist) {
		    push @set_grp_topnodes, $node;
		    next;
		}

		my $ch = $node2ch->{$node};
		(defined $ch) and (push @new_sbjcts, @{$ch});

	    }			# End of the inner while-loop #

	    @subjects = @new_sbjcts;

	}		# End of the outer while-loop # 

	return \@set_grp_topnodes;

}	# End of the subroutine 'group_close_extnodes'


# make_taxon2depth:
#	This subroutine constructs a hash table:
#	%taxon2depth = (taxon ID => depth of the common ancestor with the query taxon, ...)
#		for a specified set of taxa.
#
# REVISED on Sunday, December 31, 2006:
#	The subroutine now returns a hash %taxon2lca = ($taxon_ID => $lca, ...)
#	as well.
#	In order to keep backward compatibility, I return \%taxon2depth
#	at the tail of the output list.
#
# INPUT: ($qu_taxon, \@sb_taxa, \%taxon_name2id, %taxon_id2parent).
#	If %taxon_name2id is EMPTY, we regard $qu_taxon and @sb_taxa are given
#	in terms of taxon IDs, not taxon names.
#
sub make_taxon2depth ($\@\%\%) {

	my ($qu_taxon, $sb_taxa, $taxon_name2id, $taxon_id2parent)  = @_;

	my $fl_id = (scalar keys %{$taxon_name2id}) ? 0 : 1;
		# =1 if the 0th and 1st arguments are given in terms of IDs,
		# =0 if they are given in terms of names.

	my $qu_id = $fl_id ? $qu_taxon : $taxon_name2id->{$qu_taxon};
	unless (defined $qu_id) {
		print STDERR "UNDEFINED taxon ID for the query taxon $qu_taxon.\n";
		return (undef, undef);
	}

	my $ct_sb = @{$sb_taxa};

	my %taxon2depth; # (taxon_ID => depth of the common ancestor with the query taxon, ...)
	my %taxon2lca; # ($taxon_ID => $lca, ...)

	for (my $i=0; $i <$ct_sb; $i++) {

	    my $sb_taxon = $sb_taxa->[$i];
	    my $sb_id = $fl_id ? $sb_taxon : $taxon_name2id->{$sb_taxon};
	    unless (defined $sb_id) {
		print STDERR "UNDEFINED taxon ID for the subject taxon $sb_taxon.\n";
		return (undef, undef);
	    }

	    my ($lca, $depth) = fetch_last_common_ancestor ($qu_id, $sb_id, %{$taxon_id2parent});
	    unless (defined $depth) {
		print STDERR "'fetch_last_common_ancestor' failed for the query taxon $qu_taxon and subject taxon $sb_taxon.\n";
		return (undef, undef);
	    }

	    $taxon2depth{$sb_id} = $depth;
	    $taxon2lca{$sb_id} = $lca;
	}

	return (\%taxon2lca, \%taxon2depth);

}		# end of the subroutine make_taxon2depth #


# max_min_depths_on_subtree:
#	This subroutine examines a subtree with a specified top node
#	and collects external nodes with the maximum and minimum depths.
#
# INPUT: ($sb_topnode, \%nodeid2depth, \%node2children), where
#	$sb_topnode is the top node of the subtree under consideration.
#
# OUTPUT: ($min_depth, $max_depth, \@min_seqs, \@max_seqs), where
#	@min/max_seqs stores the IDs of the external nodes (= sequences)
#	with the minimum/maximum depth.  
#
sub max_min_depths_on_subtree ($\%\%) {

	my ($sb_topnode, $nodeid2depth, $node2children) = @_;

	my $ext_nodes = fetch_ext_offsprings_nr ($sb_topnode, %{$node2children});

	my @max_seqs = my @min_seqs = ();
	my ($max_depth, $min_depth);

	foreach my $nodeid (@{$ext_nodes}) { # the outer foreach #
	    my $depth = $nodeid2depth->{$nodeid};
	    (defined $depth) or next;

	    unless (defined $max_depth) {
		$max_depth = $min_depth = $depth;
		@max_seqs = ($nodeid);
		@min_seqs = ($nodeid);
		next;
	    }

	    if ($depth < $min_depth) {
		$min_depth = $depth;
		@min_seqs = ($nodeid);
		next;
	    }
	    if ($depth > $max_depth) {
		$max_depth = $depth;
		@max_seqs = ($nodeid);
		next;
	    }

	    ($depth == $min_depth) and (push @min_seqs, $nodeid) ;
	    ($depth == $max_depth) and (push @max_seqs, $nodeid) ;

	}				# end of the outer foreach #

	return ($min_depth, $max_depth, \@min_seqs, \@max_seqs);

}		# end of the subroutine 'max_min_depths_on_subtree' #


# fetch_taxon_nodes: This subroutine fetches all the nodes associated with
#		a specified taxon ID (or name) under a specified top-node.
#
sub fetch_taxon_nodes ($$\%\%) {

	my ($top_node, $taxon, $node2children, $node2taxon) = @_;

	my @taxon_nodes = ();

	my $ext_offs = fetch_ext_offsprings ($top_node, %{$node2children});
	for my $node (@{$ext_offs}) {
		($node2taxon->{$node} eq $taxon) and (push @taxon_nodes, $node);
	}

	return \@taxon_nodes;

}		# end of the subroutine fetch_taxon_nodes #


# fetch_extnodes_by_taxon_attr:
#	This subroutine searches a subtree for
#	external nodes of a specified taxon attribute.
#
# INPUT: ($top_node, $sel_attr, \%node2children, \%node2taxon, \%taxon2attr, $method), where
#	$top_node is the ID of the top node of the subtree you want to search;
#	$sel_attr is the attribute of the taxa you want to extract;
#	%node2children = ($node_id => \@children, ...);
#	%node2taxon = ($node_id => $taxon_id, ...);
#	%taxon2attr = ($taxon_id => $attribute, ...);
#	$method = 'a' (or 'd') if you want to deal with the attributes as strings,
#		= '+' (or '-') if the attributes are regarded as numbers.
#
# OUTPUT: \@relv_extnodes, where
#	@relv_extnodes = ( a list of IDs of nodes  ).
#
sub fetch_extnodes_by_taxon_attr ($$\%\%\%$) {

	my ($top_node, $sel_attr, $node2children, $node2taxon,
		$taxon2attr, $method) = @_;

	my $fl_num = (($method eq '+') or ($method eq '-')) ? 1 : 0;
	my @relv_extnodes = ();

	my $extnodes = fetch_ext_offsprings_nr ($top_node, %{$node2children});

	foreach my $node (@{$extnodes}) {
	    my $taxon = $node2taxon->{$node};
	    (defined $taxon) or next;
	    my $attr = $taxon2attr->{$taxon};
	    (defined $attr) or next;
	    ( ($fl_num and ($attr == $sel_attr))
		or (!($fl_num) and ($attr eq $sel_attr)) )
		and (push @relv_extnodes, $node);
	}

	return \@relv_extnodes;

}		# end of the subroutine 'fetch_extnodes_by_taxon_attr' #


# search_orthocands:
#	This subroutine looks for ortholog candidates of a query sequence
#	from  a set of sequences of specified species.
#
# INPUT: ($qu_node, $target_taxon, \%node2parent, \%node2children, \%node2taxon, \%taxon2depth), where
#	%node2taxon = (external node ID => source taxon ID, ...),
#	%taxon2depth = (taxonID => depth of the common ancestor with the query taxon, ...).
#
# OUTPUT: (#{target sequences}, \@orthocands), where 
#	@orthocands  = (a list of IDs).
#
# PREREQUISITE: The input tree should have outgroup sequences properly segregated,
#	e.g. via  'segregate_subset_nodes' in the module 'MyTreemanipu_II.pm'.
#
sub search_orthocands ($$\%\%\%\%) {

	my ($qu_node, $target_taxon,
	$node2parent, $node2children, $node2taxon, $taxon2depth) = @_;

	my $target_depth = $taxon2depth->{$target_taxon};
	unless (defined $target_depth) {
	    print STDERR "NO DEPTH found for the target taxon ID $target_taxon.\n";
	    return (undef, undef);
	}

	my $ct_target = 0;
	foreach my $node (keys %{$node2taxon}) {
		($node2taxon->{$node} eq $target_taxon) and $ct_target++;
	}

	$ct_target or (return  (0, []) ) ;

	my $pres_node = $qu_node;

	while (defined (my $parent = $node2parent->{$pres_node}) ) { # the outermost while

	    my $siblings = $node2children->{$parent};
	    unless (defined $siblings) { # This cannot happen #
		print STDERR  "UNDEFINED siblings for the node ID $pres_node .\n";
		return ($ct_target, undef);
	    }

	    my $min_depth;
	    my @orthocands = ();

	    foreach my $sibl (@{$siblings}) {      # the middle foreach #

		($sibl  eq $pres_node) and next;  # the node was already inspected #
		my $sbjcts = fetch_ext_offsprings ($sibl, %{$node2children});
		my $ct_sb = @{$sbjcts};
		my @reservoir = ();  # a temporary storage of ortholog candidates
		my $max_depth;

		for (my $i=0; $i<$ct_sb; $i++) {   # the inner for #
		    my $sb_node = $sbjcts->[$i];
		    my $sb_taxon = $node2taxon->{$sb_node}; 
		    unless (defined $sb_taxon) {
			print STDERR "NO TAXON ID is associated with the node ID $sb_node.\n";
			return ($ct_target, undef);
		    }
		    my $sb_depth = $taxon2depth->{$sb_taxon};
		    unless (defined $sb_depth) {
			print STDERR "NO DEPTH found for the subject taxon ID $sb_taxon.\n";
			return ($ct_target, undef);
		    }


		    if (defined $min_depth) {
			($min_depth > $sb_depth) and ($min_depth = $sb_depth) ;
		    } else {
			$min_depth = $sb_depth;
		    }

		    if (defined $max_depth) {
			($max_depth < $sb_depth) and ($max_depth = $sb_depth);
		    } else {
			$max_depth = $sb_depth;
		    }

		    if ($sb_taxon eq $target_taxon) {
			push @reservoir, $sb_node;
		    } elsif ($sb_depth < $target_depth) {
			 # an outgroup appeared prior to
			 # (or at the same time with) target sequences
			@reservoir = ();
			last;
		    }

		}		# end of the innder for #

		( ($min_depth <= $target_depth) and
		   ($max_depth == $target_depth) and @reservoir)
			and  (push @orthocands, @reservoir);

	    }		# end of the middle foreach #

		# There appeared species phylogenetically as far as or farther than the target.
	    ($min_depth <= $target_depth) and ( return ($ct_target, \@orthocands) ) ;
	    ## elsif ($min_depth < $target_depth) { return ($ct_target, []) ; }

	    $pres_node = $parent;  # climb 1 step further

	}			# end of the outermost while #

		# Normally, this cannot happen .
	print STDERR "'search_orthcands' FAILED in a manner which cannot happen normally.\n";
	return ($ct_target, undef); # no ortholog candidates were found #

}		# end of the subroutine search_orthocands #


# search_orthocands2:
#	This subroutine looks for ortholog candidates of a query sequence
#	from  a set of sequences of specified species.
#	This is a revised version of 'search_orthocands'.
#	The major difference is that the subroutine now examines the
#	phylogenetic relationship more rigorously.
#
# INPUT: ($qu_node, $target_taxon, \%node2parent, \%node2children, \%node2taxon, \%taxon2depth), where
#	%node2taxon = (external node ID => source taxon ID, ...),
#	%taxon2depth = (taxonID => depth of the common ancestor with the query taxon, ...).
#
# OUTPUT: (#{target sequences}, \@orthocands), where 
#	@orthocands  = (a list of IDs).
#
# PREREQUISITE: The input tree should have outgroup sequences properly segregated,
#	e.g. via  'segregate_subset_nodes' in the module 'MyTreemanipu_II.pm'.
#
sub search_orthocands2 ($$\%\%\%\%) {

	my ($qu_node, $target_taxon,
	$node2parent, $node2children, $node2taxon, $taxon2depth) = @_;

	my $target_depth = $taxon2depth->{$target_taxon};
	unless (defined $target_depth) {
	    print STDERR "NO DEPTH found for the target taxon ID $target_taxon.\n";
	    return (undef, undef);
	}

	my $ct_target = 0;
	foreach my $node (keys %{$node2taxon}) {
		($node2taxon->{$node} eq $target_taxon) and $ct_target++;
	}

	$ct_target or (return  (0, []) ) ;

	my $pres_node = $qu_node;

	while (defined (my $parent = $node2parent->{$pres_node}) ) { # the outermost while

	    my $siblings = $node2children->{$parent};
	    unless (defined $siblings) { # This cannot happen #
		print STDERR  "UNDEFINED siblings for the node ID $pres_node .\n";
		return ($ct_target, undef);
	    }

	    ##my $min_depth;
	    my $minmin_depth; # Revised on Monday, April 17, 2006
	    my @orthocands = ();

	    foreach my $sibl (@{$siblings}) {      # the middle foreach #

		($sibl  eq $pres_node) and next;  # the node was already inspected #
		my $sbjcts = fetch_ext_offsprings ($sibl, %{$node2children});
		my $ct_sb = @{$sbjcts};
		my @reservoir = ();  # a temporary storage of ortholog candidates
		my $max_depth;
		my $min_depth; # Revised on Monday, April 17, 2006

		for (my $i=0; $i<$ct_sb; $i++) {   # the inner for #
		    my $sb_node = $sbjcts->[$i];
		    my $sb_taxon = $node2taxon->{$sb_node}; 
		    unless (defined $sb_taxon) {
			print STDERR "NO TAXON ID is associated with the node ID $sb_node.\n";
			return ($ct_target, undef);
		    }
		    my $sb_depth = $taxon2depth->{$sb_taxon};
		    unless (defined $sb_depth) {
			print STDERR "NO DEPTH found for the subject taxon ID $sb_taxon.\n";
			return ($ct_target, undef);
		    }


		    if (defined $min_depth) {
			($min_depth > $sb_depth) and ($min_depth = $sb_depth) ;
		    } else {
			$min_depth = $sb_depth;
		    }

		    if (defined $max_depth) {
			($max_depth < $sb_depth) and ($max_depth = $sb_depth);
		    } else {
			$max_depth = $sb_depth;
		    }

		    if ($sb_taxon eq $target_taxon) {
			push @reservoir, $sb_node;
		    } elsif ($sb_depth < $target_depth) {
			 # an outgroup appeared prior to
			 # (or at the same time with) target sequences
			@reservoir = ();
			last;
		    }

		}		# end of the innder for #

			## Revised on Monday, April 17, 2006 ##
		##( ($min_depth <= $target_depth) and
		( ($min_depth == $target_depth) and
		   ($max_depth == $target_depth) and @reservoir)
			and  (push @orthocands, @reservoir);

			## Added on Monday, April 17, 2006 ##
		((defined $minmin_depth) and ($minmin_depth <= $min_depth))
			or ($minmin_depth = $min_depth);

	    }		# end of the middle foreach #

		# There appeared species phylogenetically as far as
		# or farther than the target.
		##  Revised on Monday, April 17, 2006 ##
	    ##($min_depth <= $target_depth) and ( return ($ct_target, \@orthocands) ) ;
	    ($minmin_depth <= $target_depth) and ( return ($ct_target, \@orthocands) ) ;
	    ## elsif ($min_depth < $target_depth) { return ($ct_target, []) ; }

	    $pres_node = $parent;  # climb 1 step further

	}			# end of the outermost while #

		# Normally, this cannot happen .
	print STDERR "'search_orthcands' FAILED in a manner which cannot happen normally.\n";
	return ($ct_target, undef); # no ortholog candidates were found #

}		# end of the subroutine search_orthocands2 #


# search_closest_homologs:
#	This subroutine looks for homologs which are the closest to
#	the query sequence(s) among sequences of specified species.
#	The subroutine does NOT care if the homologs are ortholog or not.
#
# INPUT: ($qu_node, $target_taxon, \%node2parent, \%node2children, \%node2taxon, \%taxon2depth), where
#	%node2taxon = (external node ID => source taxon ID, ...),
#	%taxon2depth = (taxonID => depth of the common ancestor with the query taxon, ...).
#
# OUTPUT: (\@closest_homologs, LCA, LCA's child on the query side,
#		the minimum depth in the branch on the query side), 
#	where @closest_homologs  = (\@homologs_br1, \@homologs_br2, ...),
#	and @homologs_brm = 
#	    (the minimum depth in the branch,
#		the maximum depth in the branch, the top node on the branch,
#		[a list of closest homologs])
#
# RECOMMENDED FEATURE: The input tree should have outgroup sequences properly segregated,
#	e.g. via  'segregate_subset_nodes' in the module 'MyTreemanipu_II.pm'.
#
sub search_closest_homologs ($$\%\%\%\%) {

	my ($qu_node, $target_taxon,
	$node2parent, $node2children, $node2taxon, $taxon2depth) = @_;

	my @null = (undef, undef, undef, undef);

	my $qu_taxon = $node2taxon->{$qu_node};
	my $qu_depth = $taxon2depth->{$qu_taxon};
	my $target_depth = $taxon2depth->{$target_taxon};
	unless (defined $target_depth) {
	    print STDERR "[CAUTION 'search_closest_homologs'] NO DEPTH found for the target taxon ID $target_taxon.\n";
	    return @null
	}

	my $ct_target = 0;
	foreach my $node (keys %{$node2taxon}) {
		($node2taxon->{$node} eq $target_taxon) and $ct_target++;
	}

	$ct_target or (return  ([], undef, undef, undef) ) ;

	my $pres_node = $qu_node;
	my $min_depth = $qu_depth;

	while (defined (my $parent = $node2parent->{$pres_node}) ) { # the outermost while

	    my $siblings = $node2children->{$parent};
	    unless (defined $siblings) { # This cannot happen #
		print STDERR  "[CAUTION 'search_closest_homologs'] UNDEFINED siblings for the node ID $pres_node .\n";
		##return ($ct_target, undef);
		return @null; ## REVISED on Friday, June 2, 2006
	    }

	    my @candidates = ();
	    # element = [local_min_depth, local_max_depth,
	    #		branch_top_node (= $sibl), \@local_candidates]
	    my @dismissed = (); # for branchs with no target homologs
				# element = local_min_depth

	    foreach my $sibl (@{$siblings}) {      # the middle foreach #

		($sibl  eq $pres_node) and next;  # the node was already inspected #
		my $sbjcts = fetch_ext_offsprings ($sibl, %{$node2children});
		my $ct_sb = @{$sbjcts};
		my @reservoir = ();  # a temporary storage of closest candidates
		my $lmax_depth;
		my $lmin_depth;


		for (my $i=0; $i<$ct_sb; $i++) {   # the inner for #
		    my $sb_node = $sbjcts->[$i];
		    my $sb_taxon = $node2taxon->{$sb_node}; 
		    unless (defined $sb_taxon) {
			print STDERR "[CAUTION 'search_closest_homologs'] NO TAXON ID is associated with the node ID $sb_node.\n";
			return @null;
		    }
		    my $sb_depth = $taxon2depth->{$sb_taxon};
		    unless (defined $sb_depth) {
			print STDERR "[CAUTION 'search_closest_homologs'] NO DEPTH found for the subject taxon ID $sb_taxon.\n";
			return @null;
		    }

		    if (defined $lmin_depth) {
			($lmin_depth > $sb_depth) and ($lmin_depth = $sb_depth) ;
		    } else {
			$lmin_depth = $sb_depth;
		    }

		    if (defined $lmax_depth) {
			($lmax_depth < $sb_depth) and ($lmax_depth = $sb_depth);
		    } else {
			$lmax_depth = $sb_depth;
		    }

		    ($sb_taxon eq $target_taxon) and (push @reservoir, $sb_node) ;

		}		# end of the innder for #

		if (@reservoir) {
		    push @candidates, [$lmin_depth, $lmax_depth, $sibl, \@reservoir];
		} else {
		    push @dismissed, $lmin_depth;
		}

	    }		# end of the middle foreach #

	    if (@candidates) {
		return (\@candidates, $parent, $pres_node, $min_depth);

	    } else { # target homologs have not appeared yet #
		foreach my $lmin_depth (@dismissed) {
		    if (defined $min_depth) {
			($min_depth > $lmin_depth) and ($min_depth = $lmin_depth);
		    } else {
			$min_depth = $lmin_depth;
		    }
		}
	    }

	    $pres_node = $parent;  # climb 1 step further

	}			# end of the outermost while #

		# Normally, this cannot happen .
	print STDERR "'search_orthcands' FAILED in a manner which cannot happen normally.\n";
	return @null; # no ortholog candidates were found #

}		# end of the subroutine search_closest_homologs #


# search_closest_outgroups:
#	This subroutine looks for outgroups which are the shallowest
#	among those which are the closest to the query sequence(s).
#
# INPUT: ($init_node, $target_depth, \%node2parent, \%node2children, \%node2taxon, \%taxon2depth), where
#	$init_node = query node, LCA, or LCA's offspring on the query lineage.
#	%node2taxon = (external node ID => source taxon ID, ...),
#	%taxon2depth = (taxonID => depth of the common ancestor with the query taxon, ...).
#
# OUTPUT: \@closest_ooutgroup, where
#	@{$closest_outgroup[l-1]} = (node_ID, taxon, depth).
#
# PREREQUISITE: The input tree should have outgroup sequences properly segregated,
#	e.g. via  'segregate_subset_nodes' in the module 'MyTreemanipu_II.pm'.
#
sub search_closest_outgroups ($$\%\%\%\%) {

	my ($init_node, $target_depth,
	$node2parent, $node2children, $node2taxon, $taxon2depth) = @_;

	my $pres_node = $init_node;
	my $min_depth;

	while (defined (my $parent = $node2parent->{$pres_node}) ) { # the outermost while

	    my $siblings = $node2children->{$parent};
	    unless (defined $siblings) { # This cannot happen #
		print STDERR  "UNDEFINED siblings for the node ID $pres_node .\n";
		return undef;
	    }

	    my @closest_outgroups = ();
	    # element = [node_ID, taxon, depth]

	    foreach my $sibl (@{$siblings}) {      # the middle foreach #

		($sibl  eq $pres_node) and next;  # the node was already inspected #
		my $sbjcts = fetch_ext_offsprings ($sibl, %{$node2children});
		my $ct_sb = @{$sbjcts};
		my @reservoir = ();  # a temporary storage of closest outgroup candidates
		my $lmin_depth;

		for (my $i=0; $i<$ct_sb; $i++) {   # the inner for #
		    my $sb_node = $sbjcts->[$i];
		    my $sb_taxon = $node2taxon->{$sb_node}; 
		    unless (defined $sb_taxon) {
			print STDERR "NO TAXON ID is associated with the node ID $sb_node.\n";
			return undef;
		    }
		    my $sb_depth = $taxon2depth->{$sb_taxon};
		    unless (defined $sb_depth) {
			print STDERR "NO DEPTH found for the subject taxon ID $sb_taxon.\n";
			return undef;
		    }

		   ($sb_depth < $target_depth) or next; # dismiss 'in-group's

		   my @sb_info = ($sb_node,  $sb_taxon, $sb_depth);
		   if (!(defined $lmin_depth) or ($lmin_depth > $sb_depth) ) {
			$lmin_depth = $sb_depth;
			@reservoir = (\@sb_info);

		   } elsif ($lmin_depth == $sb_depth) {
			push @reservoir, \@sb_info;
		   }

		}		# end of the innder for #

		(@reservoir) or next; # we have no interest in a no-og branch

		   # NOTE that  we take 'deeper' branches #
		if ( !(@closest_outgroups) or
			($lmin_depth > $closest_outgroups[0]->[2]) ) {
		    @closest_outgroups = @reservoir;
		} elsif ($lmin_depth == $closest_outgroups[0]->[2]) {
		    push @closest_outgroups, @reservoir;
		}

	    }		# end of the middle foreach #

	    (@closest_outgroups) and  (return \@closest_outgroups);

	    $pres_node = $parent;  # climb 1 step further

	}			# end of the outermost while #

		# This can happen when the input does NOT contain
		# any sequences less deep than the target.
	return []; # no outgroup candidates were found #

}		# end of the subroutine search_closest_outgroups #


# search_closest_outgroups2:
#	This subroutine looks for outgroups which are the shallowest
#	among those which are the closest to the query sequence(s).
#	This is a slightly revised version of 'search_closest_outgroups'
#	where only the output format has changed.
#
# INPUT: ($init_node, $target_depth, \%node2parent, \%node2children, \%node2taxon, \%taxon2depth), where
#	$init_node = query node, LCA, or LCA's offspring on the query lineage.
#	%node2taxon = (external node ID => source taxon ID, ...),
#	%taxon2depth = (taxonID => depth of the common ancestor with the query taxon, ...).
#
# OUTPUT: \@closest_outgroup, where
#	@closest_outgroup = ( \@closest_outgroup_br1, \@closest_outgroup_br2, ...), where
#	@{$closest_outgroup_brm[l-1]} = (node_ID, taxon, depth).
#
# PREREQUISITE: The input tree should have outgroup sequences properly segregated,
#	e.g. via  'segregate_subset_nodes' in the module 'MyTreemanipu_II.pm'.
#
sub search_closest_outgroups2 ($$\%\%\%\%) {

	my ($init_node, $target_depth,
	$node2parent, $node2children, $node2taxon, $taxon2depth) = @_;

	my $pres_node = $init_node;
	my $min_depth;

	while (defined (my $parent = $node2parent->{$pres_node}) ) { # the outermost while

	    my $siblings = $node2children->{$parent};
	    unless (defined $siblings) { # This cannot happen #
		print STDERR  "UNDEFINED siblings for the node ID $pres_node .\n";
		return undef;
	    }

	    my @closest_outgroups = ();
	    # element = [node_ID, taxon, depth]

	    foreach my $sibl (@{$siblings}) {      # the middle foreach #

		($sibl  eq $pres_node) and next;  # the node was already inspected #
		my $sbjcts = fetch_ext_offsprings ($sibl, %{$node2children});
		my $ct_sb = @{$sbjcts};
		my @reservoir = ();  # a temporary storage of closest outgroup candidates
		my $lmin_depth;

		for (my $i=0; $i<$ct_sb; $i++) {   # the inner for #
		    my $sb_node = $sbjcts->[$i];
		    my $sb_taxon = $node2taxon->{$sb_node}; 
		    unless (defined $sb_taxon) {
			print STDERR "NO TAXON ID is associated with the node ID $sb_node.\n";
			return undef;
		    }
		    my $sb_depth = $taxon2depth->{$sb_taxon};
		    unless (defined $sb_depth) {
			print STDERR "NO DEPTH found for the subject taxon ID $sb_taxon.\n";
			return undef;
		    }

		   ($sb_depth < $target_depth) or next; # dismiss 'in-group's

		   my @sb_info = ($sb_node,  $sb_taxon, $sb_depth);
		   if (!(defined $lmin_depth) or ($lmin_depth > $sb_depth) ) {
			$lmin_depth = $sb_depth;
			@reservoir = (\@sb_info);

		   } elsif ($lmin_depth == $sb_depth) {
			push @reservoir, \@sb_info;
		   }

		}		# end of the innder for #

		(@reservoir) or next; # we have no interest in a no-og branch

		   # NOTE that  we take 'deeper' branches #
		if ( !(@closest_outgroups) or
			($lmin_depth > $closest_outgroups[0][0]->[2]) ) {
		    @closest_outgroups = (\@reservoir);
		} elsif ($lmin_depth == $closest_outgroups[0][0]->[2]) {
		    push @closest_outgroups, \@reservoir;
		}

	    }		# end of the middle foreach #

	    (@closest_outgroups) and  (return \@closest_outgroups);

	    $pres_node = $parent;  # climb 1 step further

	}			# end of the outermost while #

		# This can happen when the input does NOT contain
		# any sequences less deep than the target.
	return []; # no outgroup candidates were found #

}		# end of the subroutine search_closest_outgroups2 #


# search_closest_outgroups3:
#	This subroutine looks for outgroups which are the shallowest
#	among those which are the closest to the query sequence(s).
#	This is a slightly revised version of 'search_closest_outgroups2',
#	which in turs is a slightly revised version of 'search_closest_outgroups'.
#	This one ('3') differs from '2' only in the input format,
#	while '2' difffers from the original one only in the output format.
#	
#
# INPUT: ($init_node, $target_depth, \%node2parent, \%node2children, \%node2taxon), where
#	$init_node = query node, LCA, or LCA's offspring on the query lineage.
#	%node2taxon = (external node ID => [depth, source taxon ID], ...),
#	where 'depth' is the depth of the common ancestor with the query taxon.
#
# OUTPUT: \@closest_outgroup, where
#	@closest_outgroup = ( \@closest_outgroup_br1, \@closest_outgroup_br2, ...), where
#	@{$closest_outgroup_brm[l-1]} = (node_ID, taxon, depth).
#
# PREREQUISITE: The input tree should have outgroup sequences properly segregated,
#	e.g. via  'segregate_subset_nodes' in the module 'MyTreemanipu_II.pm'.
#
sub search_closest_outgroups3 ($$\%\%\%) {

	my ($init_node, $target_depth,
	$node2parent, $node2children, $node2taxon) = @_;
	##$node2parent, $node2children, $node2taxon, $taxon2depth) = @_;

	my $pres_node = $init_node;
	my $min_depth;

	while (defined (my $parent = $node2parent->{$pres_node}) ) { # the outermost while

	    my $siblings = $node2children->{$parent};
	    unless (defined $siblings) { # This cannot happen #
		print STDERR  "UNDEFINED siblings for the node ID $pres_node .\n";
		return undef;
	    }

	    my @closest_outgroups = ();
	    # element = [node_ID, taxon, depth]

	    foreach my $sibl (@{$siblings}) {      # the middle foreach #

		($sibl  eq $pres_node) and next;  # the node was already inspected #
		my $sbjcts = fetch_ext_offsprings ($sibl, %{$node2children});
		my $ct_sb = @{$sbjcts};
		my @reservoir = ();  # a temporary storage of closest outgroup candidates
		my $lmin_depth;

		for (my $i=0; $i<$ct_sb; $i++) {   # the inner for #
		    my $sb_node = $sbjcts->[$i];
		    ##my $sb_taxon = $node2taxon->{$sb_node}; 
		    ##unless (defined $sb_taxon) {
		    my $sb_taxon_info = $node2taxon->{$sb_node}; 
		    unless (defined $sb_taxon_info) {
			print STDERR "NO TAXON ID is associated with the node ID $sb_node.\n";
			return undef;
		    }
		    ##my $sb_depth = $taxon2depth->{$sb_taxon};
		    my ($sb_depth, $sb_taxon) = @{$sb_taxon_info};
		    unless (defined $sb_depth) {
			print STDERR "NO DEPTH found for the subject taxon ID $sb_taxon.\n";
			return undef;
		    }

		   ($sb_depth < $target_depth) or next; # dismiss 'in-group's

		   my @sb_info = ($sb_node,  $sb_taxon, $sb_depth);
		   if (!(defined $lmin_depth) or ($lmin_depth > $sb_depth) ) {
			$lmin_depth = $sb_depth;
			@reservoir = (\@sb_info);

		   } elsif ($lmin_depth == $sb_depth) {
			push @reservoir, \@sb_info;
		   }

		}		# end of the innder for #

		(@reservoir) or next; # we have no interest in a no-og branch

		   # NOTE that  we take 'deeper' branches #
		if ( !(@closest_outgroups) or
			($lmin_depth > $closest_outgroups[0][0]->[2]) ) {
		    @closest_outgroups = (\@reservoir);
		} elsif ($lmin_depth == $closest_outgroups[0][0]->[2]) {
		    push @closest_outgroups, \@reservoir;
		}

	    }		# end of the middle foreach #

	    (@closest_outgroups) and  (return \@closest_outgroups);

	    $pres_node = $parent;  # climb 1 step further

	}			# end of the outermost while #

		# This can happen when the input does NOT contain
		# any sequences less deep than the target.
	return []; # no outgroup candidates were found #

}		# end of the subroutine search_closest_outgroups3 #


# ascend_search_closest_ogs:
#	This satellite subroutine of 'traceback_closest_outgroups'
#	'traces back' the putative evolutionary history by CLIMBING THE TREE,
#	while collecting information on the branches it climbed 
#	and on the subtrees sticking out of the pass,
#	untill it finds a sequence whose 'depth' is less than
#	a specified one.
#
# INPUT: ($tgt_depth, $start_node, \%nodeid2depth, \%node2parent,
#	\%node2children, \%br_attr, $indx_boot, $indx_length), where
#	$tgt_depth is the lower boundary (excluded) of the depths of 'outgroup's;
#	%nodeid2depth, %node2parent, %node2children are as indicated by the name;
#	 ($boot, $length) = @{$br_attr{$nodeid}}[$indx_boot, $indx_length] 
#	are for the branch directly upstream of the node.
#
# OUTPUT: (\@info_closest_ogs, \@info_paths, \@info_side_branches), where
#	@{$info_closest_ogs[$i]} = ($sb_topnode, $boot, $brlen, $min_depth, $max_depth, \@min_seqs, \@max_seqs) is for the $i-th 'sibling' lineage of $start_node (at the relevant node);
#	@{$info_paths[$n]} = ($node_id, $boot, $brlen)
#	for the $n-th node along the main path;
#	@{$info_side_branches[$n][$j]} = (the same format as the above)
#	is for the $j-th branches that does NOT contain outgroup sequences
#	and that is off the main path at the ($n+1)-th node along the main path;
#	Here $start_node is regarded as the 0 th node.
#
sub ascend_search_closest_ogs ($$\%\%\%\%$$) {

	my ($tgt_depth, $start_node, $nodeid2depth, $node2parent,
		$node2children, $br_attr, $indx_boot, $indx_length) = @_;

	my @info_closest_ogs = my @info_side_branches = my @info_paths = ();

	my $pres_node = $start_node;

	while (defined (my $parent = $node2parent->{$pres_node}) ) { # the outermost while-loop #

	    my ($path_brlen, $path_boot) = @{$br_attr->{$pres_node}}[$indx_length, $indx_boot];
	    push @info_paths, [$start_node, $path_boot, $path_brlen];
	    my @side_branches = ();

	    my $siblings = $node2children->{$parent};
	    foreach my $sibl (@{$siblings}) { # the inner foreach #
		($sibl == $pres_node) and next;

		my ($min_depth, $max_depth, $min_seqs, $max_seqs)
		= max_min_depths_on_subtree ($sibl, %{$nodeid2depth}, %{$node2children});
		my ($brlen, $boot) = @{$br_attr->{$sibl}}[$indx_length, $indx_boot];
		if ($min_depth < $tgt_depth) {
		    push @info_closest_ogs,
			[$sibl, $boot, $brlen, $min_depth, $max_depth, $min_seqs, $max_seqs];
		} else {
		    push @side_branches,
			[$sibl, $boot, $brlen, $min_depth, $max_depth, $min_seqs, $max_seqs];
		}
	
	    }				# end of the inner foreach #

	     push @info_side_branches, \@side_branches;
	    (@info_closest_ogs>0) and last;
	    $pres_node = $parent;

	}				# end of the outermost while-loop # 

	my @indices_sort = (3,4); my @methods_sort = ('-', '+');
	my $std_info_closest_ogs = (@info_closest_ogs<2) ? \@info_closest_ogs :
		multiple_sort (@info_closest_ogs, @indices_sort, @methods_sort) ;

	##return (\@info_closest_ogs, \@info_paths, \@info_side_branches);
	return ($std_info_closest_ogs, \@info_paths, \@info_side_branches);

}		# end of the subroutine 'ascend_search_closest_ogs' #


# traceback_closest_outgroups:
#	This subroutine starts from a specified node,
#	then 'traces back' the putative evolutionary history,
#	while collecting information on the branches it passed
#	and on the subtrees sticking out of the pass,
#	untill it finds a sequence whose 'depth' is less than
#	a specified one.
#
# INPUT: ($tgt_depth, $start_node, \@excl_kids, \%nodeid2depth, \%node2parent,
#	\%node2children, \%br_attr, $indx_boot, $indx_length), where
#	$tgt_depth is the lower boundary (excluded) of the depths of 'outgroup's;
#	@excl_kids  = (a list of IDs of children of the $start_node that are
#		to be excluded from the search);
#	%nodeid2depth, %node2parent, %node2children are as indicated by the name;
#	 ($boot, $length) = @{$br_attr{$nodeid}}[$indx_boot, $indx_length] 
#	are for the branch directly upstream of the node.
#
# OUTPUT: (\@info_closest_ogs, \@info_paths, \@info_side_branches), where
#	@{$info_closest_ogs[$i]} = ($sb_topnode, $boot, $brlen, $min_depth, $max_depth, \@min_seqs, \@max_seqs) is for the $i-th 'sibling' lineage of $start_node (at the relevant node);
#	@{$info_paths[$n]} = ($node_id, $boot, $brlen)
#	for the $n-th node along the main path;
#	@{$info_side_branches[$n][$j]} = (the same format as the above)
#	is for the $j-th branches that does NOT contain outgroup sequences
#	and that is off the main path at the $n-th node along the main path;
#	Here $start_node is regarded as the 0 th node.
#
sub traceback_closest_outgroups ($$\@\%\%\%\%$$) {

	my ($tgt_depth, $start_node, $excl_kids, $nodeid2depth, $node2parent,
		$node2children, $br_attr, $indx_boot, $indx_length) = @_;

	my $lb_depth ;
	foreach my $nodeid (keys %{$nodeid2depth}) {
	    my $depth = $nodeid2depth->{$nodeid};
	    ((defined $lb_depth) and ($lb_depth <= $depth)) or
		($lb_depth = $depth);
	}

	my $kids = $node2children->{$start_node};
	(undef, undef, my $to_be_explored) = diff_array (@{$excl_kids}, @{$kids});

=probably meaningless
	 unless (defined $node2parent->{$start_node}) { # Start from the top node #
	    my @info_closest_ogs = ();
	        = descend_search_closest_ogs ($tgt_depth, $lb_depth, @{$to_be_explored}, %{$nodeid2depth},
		%{$node2parent}, %{$node2children}, %{$br_attr}, $indx_boot, $indx_length);
	    return ([], \@info_closest_ogs);
	}
=cut

		## First look for the remaining children nodes ##
	my @info_closest_ogs = my @info_side_branches = ();

	foreach my $kid (@{$to_be_explored}) {
	    my ($min_depth, $max_depth, $min_seqs, $max_seqs)
		= max_min_depths_on_subtree ($kid, %{$nodeid2depth}, %{$node2children});
	    my ($brlen, $boot) = @{$br_attr->{$kid}}[$indx_length, $indx_boot];

	    if ($min_depth < $tgt_depth) {
		push @info_closest_ogs,
		 [$kid, $boot, $brlen, $min_depth, $max_depth, $min_seqs, $max_seqs];
	    } else {
		push @info_side_branches,
		 [$kid, $boot, $brlen, $min_depth, $max_depth, $min_seqs, $max_seqs];
	    }
	}


	my @indices_sort = (3,4); my @methods_sort = ('-', '+');

		# We end the subroutine unconditionally
		# if it starts at the top node, 
		#
	if ((@info_closest_ogs>0) or !(defined $node2parent->{$start_node})) {
	    my $std_info_closest_ogs = (@info_closest_ogs<2) ? \@info_closest_ogs :
		multiple_sort (@info_closest_ogs, @indices_sort, @methods_sort);
	    return ($std_info_closest_ogs, [], [\@info_side_branches]);
	}

		## Then explore upper branches ##

	my ($info_closest_ogs, $info_paths, $info_side_branches)
	     = ascend_search_closest_ogs ($tgt_depth, $start_node,
		%{$nodeid2depth}, %{$node2parent}, %{$node2children}, %{$br_attr},
			$indx_boot, $indx_length);

		# supply information for the initial branches #
	unshift @{$info_side_branches}, \@info_side_branches;

	return ($info_closest_ogs, $info_paths, $info_side_branches);

}		# end of the subroutine 'traceback_closest_outgroups' #


1;		# a "spell" which makes the module "use"-able and/or "require"-able #



